﻿//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "alignment.h"

#include <QString>
#include <QDebug>
#include <QLineF>
#include <QPainter>
#include <QObject>
#include "profileutilities.h"

Alignment::Alignment()
{
    mDescription = "";
    mId = -9;
	//below is example
	//    std::unique_ptr<std::vector<std::unique_ptr<Point2D>>> list(new std::vector<std::unique_ptr<Point2D>>());
    std::unique_ptr<std::vector<std::unique_ptr<AlignmentPoint>>> list(new std::vector<std::unique_ptr<AlignmentPoint>>());

	mPointVector.swap(list);
    calculateBoundingRect();
}


Alignment::Alignment(const QString &description)
{
    mDescription = description;
    mId = -9;
    std::unique_ptr<std::vector<std::unique_ptr<AlignmentPoint>>> list(new std::vector<std::unique_ptr<AlignmentPoint>>());
    mPointVector.swap(list);
    calculateBoundingRect();
}

Alignment::Alignment(const Alignment & al)
{
    mDescription = al.getDescription();
    mId = al.getId();
    //Make sure the vector is empty
    //qDebug() << "Alignment::CopyAlignment About to clear point vector";
    std::unique_ptr<std::vector<std::unique_ptr<AlignmentPoint>>> list(new std::vector<std::unique_ptr<AlignmentPoint>>());
    mPointVector.swap(list);
    if(al.getPointCount()>0){
        qDebug() << "Alignment::CopyAlignment point count " << al.getPointCount();
        for(int j = 0; j < al.getPointCount(); j++){
            std::unique_ptr<AlignmentPoint> appendPoint(new AlignmentPoint(al.getPoint(j)));
            mPointVector->push_back(move(appendPoint));
        }
    }
    this->calculateStationing();
    calculateBoundingRect();
}

void Alignment::copyAlignment(const Alignment & al)
{
    mDescription = al.getDescription();
    mId = al.getId();
    //don't want to share the same flag.


    //Make sure the vector is empty
    qDebug() << "Alignment::CopyAlignment About to clear point vector";
    //mPointVector->clear();
    if(al.getPointCount()>0){
        qDebug() << "Alignment::CopyAlignment point count " << al.getPointCount();
        for(int j = 0; j < al.getPointCount(); j++){
            std::unique_ptr<AlignmentPoint> appendPoint(new AlignmentPoint(al.getPoint(j)));
            mPointVector->push_back(move(appendPoint));
        }
    }
    this->calculateStationing();
    calculateBoundingRect();
}

QPointF Alignment::getPointF(const int pn)
{
    return mPointVector->at(static_cast<size_t>(pn))->getPointF();
}


void Alignment::setId(const int algId){
    mId = algId;
}

//void Alignment::setPoint(const int pn, Point2D & point){
//    setEasting(pn, point.getEasting());
//    setNorthing(pn, point.getNorthing());

//    if(getPointCount()>1){
//        if(pn == 0){
//            setLine(pn);
//            return;
//        }
//        if(pn == (getPointCount() - 1)){
//            setLine(pn-1);
//            return;
//        }
//        setLine(pn);
//        setLine(pn-1);
//    }
//}

void Alignment::setNorthing(const int pn, double value)
{
    if(!isPointValid(pn)){
        return;
    }
    mPointVector->at(static_cast<size_t>(pn))->setNorthing(value);
    if(value > mMaxNorthing){
        mMaxNorthing = value;
    }
    if(value < mMinNorthing){
        mMinNorthing = value;
    }

}

void Alignment::setEasting(const int pn, const double value)
{
    qDebug() << "Alignment::setEasting";
    if(!isPointValid(pn)){
        return;
    }
    mPointVector->at(static_cast<size_t>(pn))->setEasting(value);
    if(value > mMaxEasting){
        mMaxEasting = value;
    }
    if(value < mMinEasting){
        mMinEasting = value;
    }
}

void Alignment::setStationing(const int pn, const double value)
{
    if(!isPointValid(pn)){
        return;
    }
    mPointVector->at(static_cast<size_t>(pn))->setStationing(value);
    return;
}

void Alignment::setPointType(const int pn, const AlignmentPoint::PointType type)
{
    if(!isPointValid(pn)){
        return;
    }
    mPointVector->at(static_cast<size_t>(pn))->setPointType(type);
    return;
}

bool Alignment::setCenterPoint(const int pn, const double easting, const double northing)
{
    if(!isPointValid(pn)){
        return false;
    }
    mPointVector->at(static_cast<size_t>(pn))->setCenter(easting, northing);
    return true;
}

bool Alignment::setCenterNorthing(const int pn, const double northing)
{
    if(!isPointValid(pn)){
        return false;
    }
    mPointVector->at(static_cast<size_t>(pn))->setCenterNorthing(northing);
    return true;
}

bool Alignment::setCenterEasting(const int pn, const double easting)
{
    if(!isPointValid(pn)){
        return false;
    }
    mPointVector->at(static_cast<size_t>(pn))->setCenterEasting(easting);
    return true;
}

double Alignment::getStationing(const int pn) const
{
    if(!isPointValid(pn)){
        return -9;
    }
    return mPointVector->at(static_cast<size_t>(pn))->getStationing();
}

double Alignment::getEasting(const int pn) const
{
    if(!isPointValid(pn)){
        return -9;
    }
    else {
        return mPointVector->at(static_cast<size_t>(pn))->getEasting();
    }
}

double Alignment::getNorthing(const int pn) const
{
    if(!isPointValid(pn)){
        return -9;
    }
    else {
        return mPointVector->at(static_cast<size_t>(pn))->getNorthing();
    }
}

AlignmentPoint::PointType Alignment::getPointType(const int pn)
{
    if(!isPointValid(pn)){
        return AlignmentPoint::Line;
    }
    else {
        return mPointVector->at(static_cast<size_t>(pn))->getPointType();
    }
}

QPointF Alignment::getCenterPoint(const int pn)
{
    if(!isPointValid(pn)){
        return QPointF();
    }
    else {
        return mPointVector->at(static_cast<size_t>(pn))->getCenter();
    }
}

void Alignment::addPoint(AlignmentPoint & point)
{
    std::unique_ptr<AlignmentPoint> appendPoint(new AlignmentPoint(point));
    //appendPoint->setUpdateFlag(mUpdateAlignment);
    mPointVector->push_back(move(appendPoint));
    //QObject::connect(mPointVector->back(), &AlignmentPoint::pointMoved, this, &Alignment::pointMoved);
    //QObject::connect(mPointVector->back(), &AlignmentPoint::deletePoint, this, &Alignment::deletePoint);
    //QObject::connect(this, &Alignment::setPointMode, mPointVector->back(), &AlignmentPoint::setMode);
    this->calculateStationing();
    this->calculateBoundingRect();
}

void Alignment::insertPoint(const int pn)
{
    std::unique_ptr<AlignmentPoint> point(new AlignmentPoint());

    mPointVector->insert(mPointVector->begin()+pn,move(point));
}

bool Alignment::removePoints(const int pnStart, const int count)
{
    if(pnStart < 0 || (pnStart+count) > getPointCount()){
        return false;
    }
    int pnEnd = pnStart + count -1;
    qDebug() << "Alignment::removePoints pnStart and pnEnd " << pnStart << " " << pnEnd;
    //take care of the points
    mPointVector->erase(mPointVector->begin()+pnStart, mPointVector->begin()+pnStart+count);
    calculateBoundingRect();
    return true;
}

bool Alignment::isPointValid(const int pn) const
{
    return (pn >= 0 && pn < static_cast<int>(mPointVector->size()));

}

void Alignment::setDescription(const QString & nm){
    mDescription = nm;
}

int Alignment::getId() const
{
    return mId;
}

QString Alignment::getDescription() const
{
    return mDescription;
}

Alignment& Alignment::operator=(const Alignment& other) {

    this->mDescription = other.mDescription;
    this->mId = other.mId;
    this->mPointVector->clear();
    if (other.getPointCount()>0) {
        qDebug() << "Alignment::operator= point count " << other.getPointCount();
        for (int j = 0; j < other.getPointCount(); j++) {
            std::unique_ptr<AlignmentPoint> appendPoint(new AlignmentPoint(other.getPoint(j)));
            //appendPoint->setUpdateFlag(mUpdateAlignment);
            this->mPointVector->push_back(move(appendPoint));
        }
    }

    return *this;
}

qreal Alignment::getSegmentLength(const int pn)const
{

        if(pn > (getPointCount()-2) || pn < 0){
            return -9;
        }
        if(mPointVector->at(pn)->getPointType() == AlignmentPoint::Line){
            QLineF line(mPointVector->at(pn)->getPointF(),mPointVector->at(pn+1)->getPointF());
            return line.length();
        }
        if(mPointVector->at(pn)->getPointType() == AlignmentPoint::CurveLeft){
            qreal span = GAT::spanAngle(mPointVector->at(pn)->getPointF(), mPointVector->at(pn+1)->getPointF(), mPointVector->at(pn)->getCenter());
            QLineF line(mPointVector->at(pn)->getPointF(),mPointVector->at(pn)->getCenter());
            return span*line.length();
        }
        if(mPointVector->at(pn)->getPointType() == AlignmentPoint::CurveRight){
            //flip flop the span so that it is positive
            qreal span = GAT::spanAngle(mPointVector->at(pn+1)->getPointF(), mPointVector->at(pn)->getPointF(), mPointVector->at(pn)->getCenter());
            QLineF line(mPointVector->at(pn)->getPointF(),mPointVector->at(pn)->getCenter());
            return span*line.length();
        }
        return -9;
}

bool Alignment::isMeasurementPointValid(const int pn) const
{
    if(pn >= 0 && pn < getMeasurementCount()){
        return true;
    }
    return false;
}

void Alignment::calculateStationing()
{
    //qDebug() << "Alignment::calculateStationing enter";
    if(mPointVector == nullptr){
        //qDebug() << "Alignment::calculateStationing nullptr";
        return;
    }
    //qDebug() << "Alignment::calculateStationing passed nullptr check";
    if(mPointVector->empty()){
        //qDebug() << "Alignment::calculateStationing size is 0";
        return;
    }
    //qDebug() << "Alignment::calculateStationing setting the first point";
    mPointVector->at(0)->setStationing(0);
   // qDebug() << "Alignment::calculateStationing set the first point";
    for(unsigned int i = 1; i < static_cast<size_t>(this->getPointCount()); i++){
        //qDebug() << "Alignment::calculateStationing itterator" << i;
        mPointVector->at(i)->setStationing(mPointVector->at(i-1)->getStationing() + getSegmentLength(i-1));
        //qDebug() << "Alignment::calculateStationing " << mPointVector->at(i)->getStationing();
    }
    //qDebug() << "Alignment::calculateStationing exiting";
}

void Alignment::setPointVector(std::unique_ptr<std::vector<std::unique_ptr<AlignmentPoint>>> pointVector)
{
    mPointVector.swap(pointVector);
    calculateStationing();
    calculateBoundingRect();
}

void Alignment::setMeasurementList(std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement> > > newMeasurementVector)
{
    mMeasurements.swap(newMeasurementVector);
}

int Alignment::getMeasurementCount() const
{
    if(mMeasurements == nullptr){
        qDebug() << "Prfile::getPointCount nullptr";
        return 0;
    }
    return static_cast<int>(mMeasurements->size());
}

qreal Alignment::getPointOneX(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mMeasurements->at(static_cast<unsigned int>(pn))->getPointOneX();
    }
    return -9;
}

qreal Alignment::getPointOneY(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mMeasurements->at(static_cast<unsigned int>(pn))->getPointOneY();
    }
    return -9;
}

qreal Alignment::getPointTwoX(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mMeasurements->at(static_cast<unsigned int>(pn))->getPointTwoX();
    }
    return -9;
}

qreal Alignment::getPointTwoY(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mMeasurements->at(static_cast<unsigned int>(pn))->getPointTwoY();
    }
    return -9;
}

int Alignment::getPointType(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mMeasurements->at(static_cast<unsigned int>(pn))->getType();
    }
    //TODO: Make invalid value
    return 0;
}

qreal Alignment::getMeasurementValue(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mMeasurements->at(static_cast<unsigned int>(pn))->getMeasurement();
    }
    return -9;
}

int Alignment::getMeasurementId(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mMeasurements->at(static_cast<unsigned int>(pn))->getMeasurementId();
    }
    return -9;
}

bool Alignment::removeMeasurements(const int row, const int count)
{
    if(isMeasurementPointValid(row) && isMeasurementPointValid(row+count-1)){
        mMeasurements->erase(mMeasurements->begin()+row,mMeasurements->begin()+row+count-1);
        return true;
    }
    return false;
}

void Alignment::setMeasurementId(const int pn, const int id) const
{
    if(isMeasurementPointValid(pn)){
        mMeasurements->at(static_cast<unsigned int>(pn))->setMeasurementId(id);
    }
}

void Alignment::addMeasurementPoint(const QPointF &pointOne, const QPointF &pointTwo)
{
    std::unique_ptr<ProfileMeasurement> meas(new ProfileMeasurement(pointOne.x(), pointOne.y(), pointTwo.x(), pointTwo.y()));
    mMeasurements->push_back(move(meas));
}

int Alignment::getPointCount() const
{
    if(mPointVector == nullptr){
        return 0;
    }
    return static_cast<int>(mPointVector->size());
}


AlignmentPoint Alignment::getPoint(const int i) const
{
    AlignmentPoint returnPoint;
    returnPoint.setEasting(mPointVector->at(static_cast<size_t>(i))->getEasting());
    returnPoint.setNorthing(mPointVector->at(static_cast<size_t>(i))->getNorthing());
    return returnPoint;
}

std::unique_ptr<AlignmentPoint> & Alignment::getPointStdPtr(const int pn) const
{
    return mPointVector->at(static_cast<size_t>(pn));
}

QRectF Alignment::boundingRect() const
{
    return QRectF(QPointF(mMinEasting, mMinNorthing),QPointF(mMaxEasting, mMaxNorthing));
}

void Alignment::calculateBoundingRect()
{
    if(mPointVector == nullptr){
        mMinNorthing = -9;
        mMaxNorthing = -9;
        mMinEasting = -9;
        mMaxEasting = -9;
        return;

    }
    if(mPointVector->size()==0){
        mMinNorthing = -9;
        mMaxNorthing = -9;
        mMinEasting = -9;
        mMaxEasting = -9;
        return;
    }
    mMinNorthing = mPointVector->at(0)->getNorthing();
    mMaxNorthing = mMinNorthing;
    mMinEasting = mPointVector->at(0)->getEasting();
    mMaxEasting = mMinEasting;
    for(size_t i = 1; i < mPointVector->size(); i++){
        if(mPointVector->at(i)->getNorthing() < mMinNorthing){
            mMinNorthing = mPointVector->at(i)->getNorthing();
        }
        else if (mPointVector->at(i)->getNorthing() > mMaxNorthing){
            mMaxNorthing = mPointVector->at(i)->getNorthing();
        }

        if(mPointVector->at(i)->getEasting() < mMinEasting){
            mMinEasting = mPointVector->at(i)->getEasting();
        }
        else if (mPointVector->at(i)->getEasting() > mMaxEasting){
            mMaxEasting = mPointVector->at(i)->getEasting();
        }
    }
}
