//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "alignmentdao.h"
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QVariant>
#include <QDebug>
#include <QVector>

#include "gatdatamanager.h"

AlignmentDao::AlignmentDao(QSqlDatabase & database): mDatabase(database)
{

}

void AlignmentDao::init() const
{
    if(!mDatabase.tables().contains("alignments")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE alignments (alignmentId INTEGER PRIMARY KEY AUTOINCREMENT,"
                   "description TEXT,"
                   "lastModified DATETIME)");
    }

    if(!mDatabase.tables().contains("alignmentPoints")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE alignmentPoints "
                   "(alignmentId INTEGER NOT NULL REFERENCES alignments( alignmentId), "
                   "northing REAL, "
                   "easting REAL, "
                   "centerNorthing REAL, "
                   "centerEasting REAL, "
                   "curveType INT, "
                   "sortOrder INTEGER NOT NULL, "
                   "lastModified DATETIME)");
        GATDataManager::debugQuery(query);
    }
    if(!mDatabase.tables().contains("alignmentMeasurements")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE alignmentMeasurements ("
                   "measurementId INTEGER PRIMARY KEY AUTOINCREMENT, "
                   "pointOneX REAL, "
                   "pointOneY REAL, "
                   "pointTwoX REAL, "
                   "pointTwoY REAL,"
                   "type INTEGER, "
                   "alignmentId INTEGER NOT NULL REFERENCES alignments( alignmentId))");
        GATDataManager::debugQuery(query);
    }

}

void AlignmentDao::addAlignment(std::unique_ptr<Alignment> &alg) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO alignments(description, lastModified) VALUES (?,datetime('now'))");
    query.bindValue(0, alg->getDescription());
    query.exec();
    alg->setId(query.lastInsertId().toInt());
    addAlignmentPoints(alg);
    addMeasurements(alg);
    GATDataManager::debugQuery(query);

}
void AlignmentDao::removeAlignment(int id) const
{
    removeAlignmentPoints(id);
    qDebug() << "AlignmentDao::removeAlignment enter";
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM alignments WHERE alignmentId=?");
    qDebug() << "AlignmentDao::removeAlignment about to bind to id " << id;
    query.bindValue(0, id);
    qDebug() << "AlignmentDao::removeAlignment bound about to execute";
    query.exec();
    GATDataManager::debugQuery(query);
}

void AlignmentDao::updateAlignment(const std::unique_ptr<Alignment> &alg) const
{
    //TODO: implament update Alignment routine
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE alignments SET description = ?, "
                  "lastModified = datetime('now') "
                  "WHERE alignmentID = ?");
    query.bindValue(0, alg->getDescription());
    //query.bindValue(1, this->toWKT(alg));
    query.bindValue(1, alg->getId());

    query.exec();
    GATDataManager::debugQuery(query);
    qDebug() << "AlignmentDao::updateAlignment getId " << alg->getId();
    GATDataManager::debugQuery(query);
}

    std::unique_ptr<std::vector<std::unique_ptr<Alignment>>> AlignmentDao::getAlignments() const
{
    QSqlQuery query("SELECT * FROM alignments", mDatabase);
    query.exec();
    std::unique_ptr<std::vector<std::unique_ptr<Alignment>>> list(new std::vector<std::unique_ptr<Alignment>>());
    while(query.next()){
        std::unique_ptr<Alignment> alignment(new Alignment());
        alignment->setId(query.value("alignmentID").toInt());
        alignment->setDescription(query.value("description").toString());
        alignment->setPointVector(getPoints(alignment->getId()));
        alignment->calculateStationing();
        alignment->calculateBoundingRect();
        list->push_back(move(alignment));
    }
    return list;
}

Alignment AlignmentDao::getAlignment(const int id) const
{
    QSqlQuery query(mDatabase);
    query.prepare("SELECT * FROM alignments WHERE alignmentID = ?");
    query.bindValue(0,id);
    query.exec();
    Alignment alignment;
    while(query.next()){
        alignment.setId(query.value("alignmentID").toInt());
        alignment.setDescription(query.value("description").toString());
        QString lineString = query.value("lineString").toString();
        alignment.setPointVector(getPoints(alignment.getId()));

    }
    alignment.calculateStationing();
    return alignment;
}

void AlignmentDao::updateAlignmentPoints(const std::unique_ptr<Alignment> &al) const
{
    removeAlignmentPoints(al->getId());
    addAlignmentPoints(al);
}

//This is used if the sort order did not change.
void AlignmentDao::updateAlignmentPoint(const std::unique_ptr<AlignmentPoint> &point, const int alignId, const int pn) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE alignmentPoints SET "
                  "northing = ?, "
                  "easting = ?, "
                  "centerNorthing = ?, "
                  "centerEasting = ?, "
                  "curveType = ?, "
                  "lastModified = dateTime('now') "
                  "WHERE alignmentId = ? AND sortOrder = ?");
    query.bindValue(0, point->getNorthing());
    query.bindValue(1, point->getEasting());
    query.bindValue(2, point->getCenter().y());
    query.bindValue(3, point->getCenter().x());
    query.bindValue(4, point->getPointType());
    query.bindValue(5, alignId);
    query.bindValue(6, pn);
    query.exec();
    GATDataManager::debugQuery(query);
}

void AlignmentDao::addMeasurement(std::unique_ptr<Alignment> &alignment, const int i) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO alignmentMeasurements ("
                  "pointOneX, pointOneY, "
                  "pointTwoX, pointTwoY, "
                  "type, alignmentId) VALUES "
                  "(?, ? , ?, ? ,? ,?) ");
    query.bindValue(0,alignment->getPointOneX(i));
    query.bindValue(1,alignment->getPointOneY(i));
    query.bindValue(2,alignment->getPointTwoX(i));
    query.bindValue(3,alignment->getPointTwoY(i));
    query.bindValue(4,static_cast<int>(alignment->getPointType(i)));
    query.bindValue(5,alignment->getId());
    query.exec();
    GATDataManager::debugQuery(query);
    alignment->setMeasurementId(i,query.lastInsertId().toInt());
}

void AlignmentDao::addMeasurements(std::unique_ptr<Alignment> &alignment) const
{
    for(int i = 0; i < alignment->getMeasurementCount(); i++){
        QSqlQuery query(mDatabase);
        query.prepare("INSERT INTO alignmentMeasurements ("
                      "pointOneX, pointOneY, "
                      "pointTwoX, pointTwoY, "
                      "type, alignmentId) VALUES "
                      "(?, ? , ?, ? ,? ,?) ");
        query.bindValue(0,alignment->getPointOneX(i));
        query.bindValue(1,alignment->getPointOneY(i));
        query.bindValue(2,alignment->getPointTwoX(i));
        query.bindValue(3,alignment->getPointTwoY(i));
        query.bindValue(4,static_cast<int>(alignment->getPointType(i)));
        query.bindValue(5,alignment->getId());
        query.exec();
        GATDataManager::debugQuery(query);
        alignment->setMeasurementId(i,query.lastInsertId().toInt());
    }
}

bool AlignmentDao::removeMeasurements(const std::unique_ptr<Alignment> &alignment, int row, int count) const
{
    bool returnValue = false;
    for(int i = row; i < row + count; i++){
        QSqlQuery query(mDatabase);
        query.prepare("DELETE FROM alignmentMeasurements WHERE measurementId =(?)");
        query.bindValue(0,alignment->getMeasurementId(i));
        returnValue = query.exec();
        GATDataManager::debugQuery(query);
    }
    return returnValue;
}

void AlignmentDao::removeAlignmentPoints(int id) const
{
    qDebug() << "AlignmentDao::removeAlignmentPoints id " << id;
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM alignmentPoints WHERE alignmentId=(?)");
    query.bindValue(0,id);
    query.exec();
    GATDataManager::debugQuery(query);
}

void AlignmentDao::addAlignmentPoints(const std::unique_ptr<Alignment> &al) const
{
    if(al->getPointCount()==0){
        //there are no points return
            return;
    }
    mDatabase.transaction();
    for(int i=0; i< al->getPointCount(); i++){
        qDebug() << "AlignmentDao::addAlignmentPoints center x y" << al->getCenterPoint(i);
       QSqlQuery query(mDatabase);
        query.prepare("INSERT INTO alignmentPoints ( "
                  "alignmentId, "
                      "northing, "
                      "easting, "
                      "centerNorthing, "
                      "centerEasting, "
                      "sortOrder, "
                      "curveType, "
                      "lastModified) "
                      "VALUES "
                      "(?, ?, ?, ?, ?, ?, ?, datetime('now'))"
                      );
        query.bindValue(0, al->getId());
        query.bindValue(1, al->getNorthing(i));
        query.bindValue(2, al->getEasting(i));
        query.bindValue(3, al->getCenterPoint(i).y());
        query.bindValue(4, al->getCenterPoint(i).x());
        query.bindValue(5,i);
        query.bindValue(6, al->getPointType(i));
        query.exec();
        GATDataManager::debugQuery(query);
    }
    mDatabase.commit();
    return;
}

std::unique_ptr<std::vector<std::unique_ptr<AlignmentPoint> > > AlignmentDao::getPoints(const int alignmentId) const
{
    std::unique_ptr<std::vector<std::unique_ptr<AlignmentPoint>>> list(new std::vector<std::unique_ptr<AlignmentPoint>>());
    QSqlQuery query("SELECT northing, "
                    "easting, "
                    "centerNorthing, "
                    "centerEasting, "
                    "curveType FROM "
                    "alignmentPoints WHERE alignmentId = ? ORDER BY sortOrder", mDatabase);

    query.bindValue(0, alignmentId);
    query.exec();
    GATDataManager::debugQuery(query);
    while(query.next()){
        std::unique_ptr<AlignmentPoint> newPoint(new AlignmentPoint);
        newPoint->setNorthing(query.value("northing").toDouble());
        newPoint->setEasting(query.value("easting").toDouble());
        newPoint->setCenter(query.value("centerEasting").toDouble(), query.value("centerNorthing").toDouble());
        if(query.value("curveType").toInt()==0){
            newPoint->setPointType(AlignmentPoint::Line);
        }
        else if(query.value("curveType").toInt()==1){
            newPoint->setPointType(AlignmentPoint::CurveLeft);
        }
        else {
            newPoint->setPointType(AlignmentPoint::CurveRight);
        }
        list->push_back(move(newPoint));
    }
    return list;
}

std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement> > > AlignmentDao::getMeasurments(const int alignmentId) const
{
    std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement>>> profileMeasurementsVector(new std::vector<std::unique_ptr<ProfileMeasurement>>);
    QSqlQuery query(mDatabase);
    query.prepare("SELECT * FROM alignmentMeasurements WHERE alignmentId = ?");
    query.bindValue(0, alignmentId);
    query.exec();
    GATDataManager::debugQuery(query);
    while(query.next()){
        std::unique_ptr<ProfileMeasurement> measurement(new ProfileMeasurement);
        measurement->setPointOne(query.value("pointOneX").toDouble(), query.value("pointOneY").toDouble());
        measurement->setPointTwo(query.value("pointTwoX").toDouble(), query.value("pointTwoY").toDouble());
        measurement->setType(query.value("type").toInt());
        measurement->setMeasurementId(query.value("measurementId").toInt());
        profileMeasurementsVector->push_back(move(measurement));
    }
    return profileMeasurementsVector;
}

