//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "alignmentmodel.h"
#include <QDebug>
#include <QVariant>
#include <QMessageBox>
#include "profilemeasurementtypes.h"
//#include <QFont>

#define POINTSCOLUMNS 6
//0 - Northing
//1 - Easting
//2 - Stationing
//3 - Type
//4 - Center Northing
//5 - Center Easting

#define ALIGNMENTCOLUMNS 3
//0 - Id
//1 - Description
//2 - Bounding Box

#define MEASUREMENTSCOLUMNS 6

#define POINTSCOL 0
//Points are found in the 0 column of the parent in the tree

#define MEASUREMENTSCOL 1
//Measurements are found in the 1 column of the parent in the tree

AlignmentModel::AlignmentModel(QObject *parent) : QAbstractItemModel(parent),
    mDb(GATDataManager::instance()),
    mAlignments(mDb.alignmentDao.getAlignments())
{
//    for(int i = 0; i < rowCount(); i++){
//        QObject::connect(mAlignments->at(static_cast<size_t>(i)), &Alignment::deleteAlignmentPoint, this, &AlignmentModel::deleteAlignmentPoint);
//    }
}

QModelIndex AlignmentModel::addAlignment(std::unique_ptr<Alignment> alignment)
{
    int rowIndex = rowCount();
    beginInsertRows(QModelIndex(), rowIndex, rowIndex);
    qDebug() << "Inserting Alignment";
    mDb.alignmentDao.addAlignment(alignment);
    mAlignments->push_back(move(alignment));
    endInsertRows();
    return index(rowIndex, 0);
}

void AlignmentModel::addAlignmentPoint(const double northing, const double easting, const QModelIndex &parent)
{
    if(isIndexValid(parent)){
        beginInsertRows(parent, rowCount(parent), rowCount(parent));
        AlignmentPoint point(northing, easting);
        qDebug() << "AlignmentModel::addPoint about to addPoint";
        mAlignments->at(static_cast<size_t>(parent.row()))->addPoint(point);
        mDb.alignmentDao.updateAlignmentPoints(mAlignments->at(static_cast<size_t>(parent.row())));
        endInsertRows();
    }
}

void AlignmentModel::addMeasurement(const QModelIndex &parent, const QPointF &pointOne, const QPointF &pointTwo)
{
    if(!parent.isValid()){
        QMessageBox msgBox;
        msgBox.setText("Invalid index, please select cross section to add measurement");
        msgBox.exec();
        return;
    }
    //convert this to the column for the measurements
    QModelIndex parentMeas = this->index(parent.row(), 2,QModelIndex());
    int insertRow = rowCount(parentMeas);
    this->beginInsertRows(parentMeas,insertRow,insertRow);
    mAlignments->at(static_cast<unsigned int>(parentMeas.row()))->addMeasurementPoint(pointOne,pointTwo);
    mDb.alignmentDao.addMeasurement(mAlignments->at(static_cast<unsigned int>(parentMeas.row())),insertRow);
    endInsertRows();
}


void AlignmentModel::addAlignmentDes(const QString &des)
{
    std::unique_ptr<Alignment> al(new Alignment());
    al->setDescription(des);
    addAlignment(move(al));
}

int AlignmentModel::rowCount(const QModelIndex &parent) const
{
    if(isIndexValid(parent)){
        //qDebug() << "AlignmentModel::rowCount parent is valid using point count";
        return static_cast<int> (mAlignments->at(static_cast<size_t>(parent.row()))->getPointCount());
    }
    return static_cast<int>(mAlignments->size());
}

int AlignmentModel::columnCount(const QModelIndex &parent) const
{
    //if the parent index is valid then it is the subtable
    if(isIndexValid(parent)){
        //qDebug() << "AlignmentModel::columnCount parent is valid 6 columns";
        if(parent.column()==POINTSCOL){
            return POINTSCOLUMNS;
        }
        if(parent.column() == MEASUREMENTSCOL){
            return MEASUREMENTSCOLUMNS;
        }
    }
    return ALIGNMENTCOLUMNS;
}

QVariant AlignmentModel::data(const QModelIndex &index, const int role) const
{
    //qDebug() << "AlighmentModel::data " << index << " role: " << role;
    //Note style sheet handles font and other formatting.
    if(!isIndexValid(index)){
        return QVariant();
    }

    if(role == Qt::DisplayRole || role== Qt::EditRole)
    {
        //qDebug() << "AlignmentModel::data displayRole";
        //qDebug() <<  index.parent();
        if(isIndexValid(index.parent())){
            if(index.parent().column() == POINTSCOL){
            //qDebug() << "AlignmentModel::data getting subTableData";
                return alignmentPointData(index);
            }
            else if(index.parent().column() == MEASUREMENTSCOL){
                return dataMeasurementParameters(index, role);
            }
        }

        else {
            if(index.column() == 0)
            {
                return mAlignments->at(static_cast<unsigned int>(index.row()))->getId();
            }
            else if(index.column() == 1){
                return mAlignments->at(static_cast<unsigned int>(index.row()))->getDescription();
            }
            else if(index.column() == 2){
                return mAlignments->at(static_cast<unsigned int>(index.row()))->boundingRect();
            }
            else
            {
                return QVariant();
            }
        }
    }
    //if the parent is invalid then we can return to the main data in the tree
    return QVariant();
}

bool AlignmentModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!isIndexValid(index)){
        return false;
    }
    qDebug() << "AlignmentModel::setData Role " << role << " column" << index.column();
    if(role == Roles::NameRole)
    {
        mAlignments->at(static_cast<unsigned int>(index.row()))->setDescription(value.toString());
        //this automatically sets up the pointer so the data is maintained.
        mDb.alignmentDao.updateAlignment(mAlignments->at(static_cast<unsigned int>(index.row())));
        emit dataChanged(index,index);
        return true;
    }
    if(role == Qt::EditRole){

        if(isIndexValid(index.parent())){
            //Has parent must be a point;
            return setAlignmentPointData(index, value);
        }

        else {
            //column 0 is the id-> it is not editable.
            //column 1 is the bounding box-> it is not editable
            //This is a parent
            if(index.column() == 1){
                mAlignments->at(static_cast<unsigned int>(index.row()))->setDescription(value.toString());
                //this automatically sets up the pointer so the data is maintained.

                mDb.alignmentDao.updateAlignment(mAlignments->at(static_cast<unsigned int>(index.row())));
                emit dataChanged(index,index);
                return true;
            }
        }
    }
    return false;
}

QVariant AlignmentModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    //qDebug() << "AlignmentModel::headerData " << role;
    if(role == Qt::DisplayRole){
        if(orientation == Qt::Horizontal){
            switch(section){
            case 0:
                return "Id";
            case 1:
                return "Name";
            case 2:
                return "Bounding Box";
            default:
                return section;
            }
        }
        return section;
    }
    return QVariant();
}

bool AlignmentModel::removeRows(int row, int count, const QModelIndex &parent)
{
    qDebug() << "AlignmentModel::removeRows";
    if(isIndexValid(parent)){
        //parent is valid so we are removing points.
        if(row < 0 || (row+count > rowCount(parent)) || count<0){
            return false;
        }
        beginRemoveRows(parent, row, row+count -1);
        qDebug() << "AlignmentModel::remove rows about to call removePoints";
        mAlignments->at(static_cast<unsigned int>(parent.row()))->removePoints(row, count);
        //write changes to the database
        //TODO:Implement updateAlignmentPoints and use that function instead
        mDb.alignmentDao.updateAlignmentPoints(mAlignments->at(static_cast<unsigned int>(parent.row())));
        endRemoveRows();
        return true;
    }
    else {
        if(row < 0
                || count<0
                || (row+count) > rowCount()){
            return false;
        }
        beginRemoveRows(parent, row, row + count -1);
        int countLeft = count;
        while(countLeft--){
            qDebug() << "AlignmentModel::removeRows removing at: " << row+countLeft;
            const Alignment& alignment = *mAlignments->at(static_cast<unsigned int>(row+countLeft));
            mDb.alignmentDao.removeAlignment(alignment.getId());
        }
        //qDebug() << "AlignmentModel::removeRows deleteing from the vector ";
        mAlignments->erase(mAlignments->begin() + row,
                           mAlignments->begin()+row+count);
        endRemoveRows();
        //qDebug() << "AlignmentModel::removeRows finished successfully";
        return true;
    }
}

bool AlignmentModel::insertRow(int row, const QModelIndex &parent)
{
    if(isIndexValid(parent)){
        //Do the point insert
        if(parent.column()==0){
            beginInsertRows(parent, row, row);
            mAlignments->at(static_cast<unsigned int>(parent.row()))->insertPoint(row);
            endInsertRows();
            return true;
        }
    }
    else{
        std::unique_ptr<Alignment> newAlignment(new Alignment);
        beginInsertRows(parent, row,row);
        //mAlignments->insert(mAlignments->begin()+row,newAlignment);
        mAlignments->insert(mAlignments->begin()+row, move(newAlignment));
        //Alignment* temp = mAlignments->at(row);
        //mAlignments->at(row). = newAlignment;
        //QObject::connect(mAlignments->at(row),&Alignment::deleteAlignmentPoint, this, &AlignmentModel::deleteAlignmentPoint);
        //delete temp;
        //this is going to add it at the end....
        mDb.alignmentDao.addAlignment(mAlignments->at(static_cast<unsigned int>(row)));
        endInsertRows();
        return true;
    }
    return false;
}

QHash<int, QByteArray> AlignmentModel::roleNames() const
{
    QHash<int, QByteArray> roles;
    roles[Roles::IdRole] = "id";
    roles[Roles::NameRole] = "name";
    return roles;
}

QModelIndex AlignmentModel::index(int row, int column, const QModelIndex & parent) const
{
    if(isIndexValid(parent))
    {
        //This was made more static during troubleshooting the last rows weren't working.
       if(column <  POINTSCOLUMNS && row < mAlignments->at(static_cast<unsigned int>(parent.row()))->getPointCount()){
            return createIndex(row, column, static_cast<quintptr>(parent.row()));
       }
    }
    else
        return createIndex(row, column,static_cast<quintptr>(-1));
    return QModelIndex();
}

bool AlignmentModel::hasChildren(const QModelIndex &index) const
{
    if(!isIndexValid(index)){
        return false;
    }
    else if(isIndexValid(index.parent())){
        //If the parent index is valid then this must be a point so
        //it doesn't have any children
        return false;
    }
    else if(mAlignments->at(static_cast<unsigned int>(index.row()))->getPointCount() > 0){
        //if the alignment doesn't have any points it won't have any children.
        return true;
    }
    return false;
}

//http://lynxline.com/designing-tree-like-models/
//THANK YOU THIS WAS A BIG HELP
QModelIndex AlignmentModel::parent(const QModelIndex &index) const
{
    if(!index.isValid()){
        return QModelIndex();
    }
    int column = 0;
    quintptr internalId = index.internalId();
    if(internalId >= 1000){
        column = 1;
        internalId = internalId - 1000;
    }

    if(static_cast<int>(index.internalId()) < rowCount() && static_cast<int>(index.internalId()) != -1){
        return this->createIndex(static_cast<int>(internalId),static_cast<quintptr>(column),-1);
    }
    return QModelIndex();
}

//QRectF AlignmentModel::boundingRect(const QModelIndex &index) const
//{
//    if(index.parent().isValid()){
//        //this is a point return
//        return QRectF();
//    }
//    return mAlignments->at(index.row())->boundingRect();
//}

Qt::ItemFlags AlignmentModel::flags(const QModelIndex & index) const
{
    if(!index.isValid())
        return Qt::ItemIsEnabled;
    //parent valid then it is a point

    if(index.parent().isValid()){
        //can only change start stationing
        if(index.column() == 2){
            if(index.row() == 0){
                return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable;
            }
            else {
                return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
            }
        }
        else {
            return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable;
        }
    }
    else if(index.column() ==0 || index.column() == 2)
        return Qt::ItemIsEnabled | Qt::ItemIsSelectable;

    return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable;
}

bool AlignmentModel::isIndexValid(const QModelIndex & index) const
{
    //qDebug() << "AlignmentMoel::isIndexValid parent row is: " << index.parent().row() << " row is: " << index.row();
    if(!index.isValid()){
        return false;
    }

    if(index.row() < 0){
        return false;
    }
    //Parent is valid need to check both values
    if(index.parent().isValid()){
        if(index.parent().row() >= mAlignments->size()){
            return false;

        }
        if(index.parent().column()== POINTSCOL){
            if(index.row() >= mAlignments->at(static_cast<size_t>(index.parent().row()))->getPointCount()){
                return false;
            }
            return true;
        }
        else if(index.parent().column() ==MEASUREMENTSCOL){
            if(index.row() >= mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getMeasurementCount()){
                return false;
            }
            return true;
        }
    }
    else {
        if(index.row() >= mAlignments->size())
        {
            return false;
        }
    }
    return true;
}

void AlignmentModel::databaseUpdated()
{
    this->beginResetModel();
    mAlignments = mDb.alignmentDao.getAlignments();
    this->endResetModel();
}

QVariant AlignmentModel::alignmentPointData(const QModelIndex &index) const
{
    if(index.column() == 0){
        return QString::number(
                    mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
                    getNorthing(index.row()),'f',3);
    }
    else if(index.column() == 1){
        return QString::number(
                    mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
                    getEasting(index.row()),'f',3);
    }
    else if(index.column() == 2){
        return QString::number(
                    mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
                    getStationing(index.row()),'f',3);
    }
    else if(index.column() == 3){
        return mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getPointType(index.row());
//        if(mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
//                getPointType(index.row())==AlignmentPoint::Line){
//            return "Line";
//        }
//        else if(mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
//            getPointType(index.row())==AlignmentPoint::Curve){
//            return "Curve";
//        }
    }
    else if(index.column() == 4){
        if(mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
                getPointType(index.row())!=AlignmentPoint::Line){
        return QString::number(
                    mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
                    getCenterPoint(index.row()).y(),'f',3);
        }
        else{
            return "--";
        }
    }
    else if(index.column() == 5){
        if(mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
                getPointType(index.row())!=AlignmentPoint::Line){
        return QString::number(
                    mAlignments->at(static_cast<unsigned int>(index.parent().row()))->
                    getCenterPoint(index.row()).x(),'f',3);
        }
        else{
            return "--";
        }
    }
    //Default
    return QVariant();
}

bool AlignmentModel::setAlignmentPointData(const QModelIndex &index, const QVariant &value, int role)
{
    //qDebug() << "AlignmentModel::setAlignmentPointData ";
    //Assume the parent has already been checked and is valid.
    Q_UNUSED(role)


    bool returnValue = false;
    //Jumped off a cliff and assumed the value could be converted to a string.
    double setValue = value.toString().toDouble(&returnValue);
    if(!returnValue){
        //qDebug() << "AlignmentModel""setAlignmentPointData couldn't convert " << value;
        return false;
    }

    if(index.column() == 0){
        //northing

        mAlignments->at(static_cast<size_t>(index.parent().row()))->setNorthing(index.row(), setValue);
        emit dataChanged(index, index);

    }
    else if(index.column() == 1){
        //easting

        mAlignments->at(static_cast<size_t>(index.parent().row()))->setEasting(index.row(), setValue);
        emit dataChanged(index, index);
    }
    else if(index.column() == 2){
        //station

        mAlignments->at(static_cast<size_t>(index.parent().row()))->setStationing(index.row(), setValue);
        emit dataChanged(index,index);
    }
    else if(index.column() == 3){
        //Type
        mAlignments->at(static_cast<size_t>(index.parent().row()))->setPointType(index.row(),static_cast<AlignmentPoint::PointType>(value.toInt()));
        emit dataChanged(index, index);
    }
    else if(index.column() == 4){
        //station

        mAlignments->at(static_cast<size_t>(index.parent().row()))->setCenterNorthing(index.row(), setValue);
        emit dataChanged(index,index);
    }
    else if(index.column() == 5){
        //station

        mAlignments->at(static_cast<size_t>(index.parent().row()))->setCenterEasting(index.row(), setValue);
        emit dataChanged(index,index);
    }
    else{
        return false;
    }
    if(returnValue){
        //Process the stationing.
            mAlignments->at(static_cast<size_t>(index.parent().row()))->calculateStationing();
            QModelIndex changedIndexTop = this->index(index.row(),2,index.parent());
            int endRow = mAlignments->at(static_cast<size_t>(index.parent().row()))->getPointCount() - 1;
            QModelIndex changedIndexBottom = this->index(endRow,2, index.parent());
            emit dataChanged(changedIndexTop, changedIndexBottom);
            mDb.alignmentDao.updateAlignmentPoint(mAlignments->at(static_cast<size_t>(index.parent().row()))->getPointStdPtr(index.row())
                                                  ,mAlignments->at(static_cast<size_t>(index.parent().row()))->getId()
                                                  ,index.row());
   }
    return returnValue;
}

QVariant AlignmentModel::dataMeasurementParameters(const QModelIndex &index, const int role) const
{
    if(role == Qt::DisplayRole){
        switch(index.column()){
        case 0:
            return mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getMeasurementValue(index.row());
        case 1:
            qDebug() << "CrossSectionModel::daataMeasurementParameters " << mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getPointOneX(index.row());
            return mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getPointOneX(index.row());
        case 2:
            return mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getPointOneY(index.row());
        case 3:
            return mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getPointTwoX(index.row());
        case 4:
            return mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getPointTwoY(index.row());
        case 5: {
            ProfileMeasurementTypes text;
            return text.getTypeName(static_cast<int>(mAlignments->at(static_cast<unsigned int>(index.parent().row()))->getPointType(index.row())));
        }
        default:
            return QVariant();
        }
    }
    return QVariant();
}

bool AlignmentModel::removeRowsMeasurements(int row, int count, const QModelIndex &parent)
{
    emit beginRemoveRows(parent, row, row+count - 1);
    bool result = mDb.alignmentDao.removeMeasurements(mAlignments->at(static_cast<unsigned int>(parent.row())),row, count);
    result = result && mAlignments->at(static_cast<unsigned int>(parent.row()))->removeMeasurements(row, count);
    endRemoveRows();
    return result;
}

#undef POINTSCOLUMNS
#undef ALIGNMENTCOLUMNS
#undef MEASUREMENTSCOL
#undef POINTSCOL
