//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "childtablemodel.h"
#include <QDebug>

ChildTableModel::ChildTableModel(QObject *parent) : QAbstractItemModel(parent)
{
    //default to 0
    mModelColumn = 0;
}

void ChildTableModel::setRootIndex(const QModelIndex & rootIndex)
{
    emit beginResetModel();
    mRootIndex = rootIndex;
    emit endResetModel();
}

void ChildTableModel::setColumn(const int column)
{
    mModelColumn = column;
}

void ChildTableModel::setRow(const int row)
{
    emit beginResetModel();
    mRootIndex = mParentModel->index(row, mModelColumn);
    emit endResetModel();
}

void ChildTableModel::setModel(QAbstractItemModel * parentModel)
{
    mParentModel = parentModel;
    QObject::connect(mParentModel, &QAbstractItemModel::dataChanged, this, &ChildTableModel::parentModelDataChanged);
    QObject::connect(mParentModel, &QAbstractItemModel::rowsAboutToBeRemoved, this, &ChildTableModel::forwardRowsAboutToBeRemoved);
    QObject::connect(mParentModel, &QAbstractItemModel::rowsRemoved, this, &ChildTableModel::forwardRowsRemoved);
    QObject::connect(mParentModel, &QAbstractItemModel::rowsInserted, this, &ChildTableModel::forwardRowsInserted);
}

QModelIndex ChildTableModel::getRootIndex()
{
    return mRootIndex;
}

int ChildTableModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    if(mRootIndex == QModelIndex()){
        return 0;
    }
    return mParentModel->rowCount(mRootIndex);
}

int ChildTableModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    if(mRootIndex == QModelIndex()){
        return 0;
    }
    return mParentModel->columnCount(mRootIndex);
}

QVariant ChildTableModel::data(const QModelIndex &index, const int role) const
{
    if(mRootIndex == QModelIndex()){
        return QVariant();
    }
    QModelIndex modelIndex = mParentModel->index(index.row(), index.column(), mRootIndex);
    //qDebug() << "ChildTableModel::data root index column " << mRootIndex.column();
    return mParentModel->data(modelIndex,role);
}

bool ChildTableModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(mRootIndex == QModelIndex()){
        return false;
    }
    //Feed this to the parent model.
    QModelIndex modelIndex = mParentModel->index(index.row(), index.column(), mRootIndex);
    return mParentModel->setData(modelIndex,value, role);
}

Qt::ItemFlags ChildTableModel::flags(const QModelIndex &index) const
{
    QModelIndex modelIndex = mParentModel->index(index.row(), index.column(), mRootIndex);
    return mParentModel->flags(modelIndex);
}

QModelIndex ChildTableModel::index(int row, int column, const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    if (hasIndex(row, column))
    {
        return createIndex(row, column,static_cast<quintptr>(-1));
    }
    return QModelIndex();
}

QModelIndex ChildTableModel::parent(const QModelIndex &index) const
{
    Q_UNUSED(index)
    //Doesn't matter what point it is the parent is invalid.
    return QModelIndex();
}

bool ChildTableModel::removeRows(int row, int count, const QModelIndex &parent)
{
    Q_UNUSED(parent)
    if(mRootIndex == QModelIndex()){
        return false;
    }
    return mParentModel->removeRows(row, count, mRootIndex);
}

bool ChildTableModel::removeRow(int row, const QModelIndex &parent)
{
    Q_UNUSED(parent)
    return removeRows(row, 1);
}

void ChildTableModel::parentModelDataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight)
{
    if(topLeft.parent().isValid()){
        //Then a child item was updated.
        if(mRootIndex == topLeft.parent()){
            //The child item
            emit dataChanged(index(topLeft.row(), topLeft.column()), index(bottomRight.row(),bottomRight.column()));
        }
    }
    if(topLeft.row() <= mRootIndex.row() && bottomRight.row() >= mRootIndex.row()){
        //Parent changed data so the model is probably out of date.
        emit beginResetModel();
        emit endResetModel();
    }
}

void ChildTableModel::forwardRowsAboutToBeRemoved(const QModelIndex &parent, int first, int last)
{
    if(parent == mRootIndex){
        emit beginRemoveRows(QModelIndex(), first, last);
    }
    if(!parent.isValid()){
        if(mRootIndex.row() >= first && mRootIndex.row() <= last){
            this->beginResetModel();
        }
    }
}

void ChildTableModel::forwardRowsRemoved(const QModelIndex &parent, int first, int last)
{
    //If the parent is the root then this affects this table and need to forward the signal.
    if(parent == mRootIndex){
        emit endRemoveRows();
    }
    //If the parent index is not valid then it is a top level entry.
    if(!mParentModel->hasChildren(parent)){
        //If the parent row was deleted this all of the children would be deleted too
        if(mRootIndex.row() >= first && mRootIndex.row() <= last){
            this->endResetModel();
        }
    }
}

void ChildTableModel::forwardRowsInserted(const QModelIndex &parent, int first, int last)
{
    //If the parent is the root then this affects this table and need to forward the signal.
    qDebug() << "ChildTableModel::forwardRowsInserted begin parent " << parent << " " << first << " " << last;
    qDebug() << "Root index " << mRootIndex;
    if(parent != QModelIndex()){
        if(parent.row() == mRootIndex.row() && parent.column()==mRootIndex.column()){
        beginInsertRows(QModelIndex(), first, last);
        endInsertRows();
        }
    }

}

bool ChildTableModel::hasIndex(const int row,const int column) const
{
    if(mRootIndex == QModelIndex()){
        return false;
    }
    if (row < 0 || column < 0)
        return false;
    return row < rowCount(mRootIndex) && column < columnCount(mRootIndex);
}
