//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "crosssection.h"
#include <QDebug>
#include <QtMath>
#include <QPointF>
#include <QLineF>

CrossSection::CrossSection() : Profile()
{
    std::unique_ptr<std::vector<std::unique_ptr<HydraulicParameters>>> list2(new std::vector<std::unique_ptr<HydraulicParameters>>());
    mHydraulicParameters.swap(list2);
    mD84 = 64; //millimeters
    mSlope = .02; //length/length
}

CrossSection::CrossSection(const QString &name) : Profile(name)
{

}

CrossSection::CrossSection(std::unique_ptr<Profile> profile) : Profile(move(profile))
{
    std::unique_ptr<std::vector<std::unique_ptr<HydraulicParameters>>> list2(new std::vector<std::unique_ptr<HydraulicParameters>>());
    mHydraulicParameters.swap(list2);
    mD84 = 64; //millimeters
    mSlope = .02; //length/length

}

void CrossSection::setSlope(const double slope)
{
    mSlope = slope;
}

void CrossSection::setD84(double d84)
{
    mD84 = d84;
}

void CrossSection::setMannings(const int i, double mannings)
{
    mHydraulicParameters->at(static_cast<unsigned int>(i))->setMannings(mannings);
}

void CrossSection::setStageDescription(const int i, QString value)
{
    mHydraulicParameters->at(static_cast<unsigned int>(i))->setDescription(value);
}

bool CrossSection::removeHydraulicParameters(const int row, const int count)
{
    if((row+count) < getHydraulicParameterCount() && row >= 0 && count > 0){
        mHydraulicParameters->erase(mHydraulicParameters->begin()+row, mHydraulicParameters->begin()+row+count);
        return true;
    }
    return false;
}

void CrossSection::setStage(const int i, const double stage)
{
    mHydraulicParameters->at(static_cast<unsigned int>(i))->setStage(stage);
    calculateHydraulicParameters(i);
}

int CrossSection::getHydraulicParameterCount() const
{
    //qDebug() << "CrossSEction::getHydraulicParameterCount enter";
    return static_cast<int>(mHydraulicParameters->size());
}

double CrossSection::getStage(const int i) const
{
    return mHydraulicParameters->at(static_cast<unsigned int>(i))->getStage();
}

QString CrossSection::getStageDescription(const int i) const
{
    return mHydraulicParameters->at(static_cast<unsigned int>(i))->getDescription();
}

double CrossSection::getMaxDepth(const int i) const
{
    return mHydraulicParameters->at(static_cast<unsigned int>(i))->getDepthMax();
}

double CrossSection::getArea(const int i) const
{
    return mHydraulicParameters->at(static_cast<unsigned int>(i))->getArea();
}

double CrossSection::getWettedPerimeter(const int i) const
{
    return mHydraulicParameters->at(static_cast<unsigned int>(i))->getWettedPerimeter();
}

double CrossSection::getWidth(const int i) const
{
    return mHydraulicParameters->at(static_cast<unsigned int>(i))->getWidth();
}

double CrossSection::getd84() const
{
    return mD84;
}

double CrossSection::getSlope() const
{
    return mSlope;
}

double CrossSection::getManningN(const int i) const
{
    return mHydraulicParameters->at(static_cast<unsigned int>(i))->getMannings();
}

double CrossSection::getHydraulicRadius(const int i) const
{
    return getArea(i)/getWettedPerimeter(i);
}

double CrossSection::getMeanDepth(const int i) const
{
    return getArea(i)/ getWidth(i);
}

double CrossSection::getRelativeRoughness(const int i) const
{
    return getHydraulicRadius(i)/getd84();
}

double CrossSection::getMeanVelocity(const int i) const
{
    return 1.486*qPow(getHydraulicRadius(i), 2.0/3.0)*qSqrt(getSlope())/getManningN(i);
}

double CrossSection::getDischarge(const int i) const
{
    return getMeanVelocity(i) * getArea(i);
}

double CrossSection::getShearStress(const int i) const
{
    return 62.4*getHydraulicRadius(i)*getSlope();
}

double CrossSection::getStreamPower(const int i) const
{
    return getShearStress(i)*getMeanVelocity(i);
}

double CrossSection::getFroudeNumber(const int i) const
{
    return getMeanVelocity(i)/qSqrt(32.2*getMeanDepth(i));
}

double CrossSection::getFrictionFactor(const int i) const
{
    return getMeanVelocity(i)/getShearVelocity(i);
}

double CrossSection::getShearVelocity(const int i) const
{
    return qSqrt(getShearStress(i));
}

void CrossSection::setHydraulicParameterList(std::unique_ptr<std::vector<std::unique_ptr<HydraulicParameters>>> hydraulicParameterList)
{
    mHydraulicParameters.swap(hydraulicParameterList);
    //after setting the list make sure all the parameters are calculated
    calculateMinPointElevation();
    calculateMaxPointElevation();
    for(int i = 0; i < getHydraulicParameterCount(); i++){
        calculateHydraulicParameters(i);
    }
}

void CrossSection::addHydraulicParameters(const double minStage, const double maxStage, const double increment)
{
    qDebug() << "CrossSection::calculateHydraulicParameters";
    mHydraulicParameters->resize(0);

    for(double stage = minStage; stage <= maxStage; stage += increment){
        addHydraulicParameters(stage);
    }
    qDebug() << "Hydarulic parameter count" << mHydraulicParameters->size();
}

void CrossSection::addHydraulicParameters(const double stage)
{
    std::unique_ptr<HydraulicParameters> hydraulicParms(new HydraulicParameters());
    hydraulicParms->setStage(stage);
    hydraulicParms->setMannings(.035); //default
    mHydraulicParameters->push_back(move(hydraulicParms));
    calculateHydraulicParameters(getHydraulicParameterCount()-1);
    qDebug() << "Stage: " << stage;
}

double CrossSection::calculateArea(const double stage)
{
    double area = 0;
    for(int i = 1; i < getPointCount(); i++){
        //case both points are below water surface;
        if(getElevation(i-1) <= stage && getElevation(i) <= stage){
            area+= calculateSectionArea(getStationing(i-1), getElevation(i-1), getStationing(i), getElevation(i),stage);
        }
        //first station only below surface
        else if(getElevation(i-1) <= stage){
            double intersect = calculateIntersect(getStationing(i-1), getElevation(i-1), getStationing(i), getElevation(i),stage);
            area += calculateSectionArea(getStationing(i-1), getElevation(i-1), intersect, stage,stage);
        }
        //second station only below surface
        else if(getElevation(i) <= stage){
            double intersect = calculateIntersect(getStationing(i-1), getElevation(i-1), getStationing(i), getElevation(i),stage);
            area += calculateSectionArea(getStationing(i), getElevation(i), intersect, stage,stage);
        }
        //Implied both above surface add 0
    }
    return area;
}

double CrossSection::calculateWettedPerimeter(const double stage)
{
    double wettedPerimeter = 0;
    for(int i = 1; i < getPointCount(); i++){
        //case both points are below water surface;
        wettedPerimeter+= calculateLength(calculateSegmentWidth(stage, i), calculateWetterPerimeterHeight(stage, i));
    }
    return wettedPerimeter;
}

double CrossSection::calculateWidth(const double stage)
{
    double width = 0;
    for(int i = 1; i < getPointCount(); i++){
        //case both points are below water surface;
        width += calculateSegmentWidth(stage, i);
    }

    return width;
}

double CrossSection::calculateSegmentWidth(const double stage, const int i){
    if(getElevation(i-1) <= stage && getElevation(i) <= stage){
        return (getStationing(i)- getStationing(i-1));
    }
    //second station only below surface
    else if(getElevation(i-1) <= stage){
        //double intersect = calculateIntersect(getStationing(i-1), getElevation(i-1), getStationing(i), getElevation(i),stage);
        return (stage-getElevation(i-1))*(getStationing(i)-getStationing(i-1))/(getElevation(i)-getElevation(i-1));
    }
    //first station only below surface
    else if(getElevation(i) <= stage){
        //double intersect = calculateIntersect(getStationing(i-1), getElevation(i-1), getStationing(i), getElevation(i),stage);
        return (stage-getElevation(i))*(getStationing(i)-getStationing(i-1))/(getElevation(i-1)-getElevation(i));
    }
    //Implied both above surface
    return 0;
}

double CrossSection::calculateWetterPerimeterHeight(const double stage, const int i){
    if(getElevation(i-1) <= stage && getElevation(i) <= stage){
        //doesn't matter which one is minimum abs will take care of this.
        return fabs(getElevation(i)-getElevation(i-1));
    }
    //second station only below surface
    else if(getElevation(i-1) <= stage){
        return (stage-getElevation(i-1));
    }
    //first station only below surface
    else if(getElevation(i) <= stage){
        return (stage-getElevation(i));
    }
    //Implied both above surface
    return 0;
}


double CrossSection::calculateLength(const double w, const double h) const
{
    if(w==0){
        return h;
    }
    if(h==0){
        return w;
    }
    return qSqrt(w*w+h*h);
}

double CrossSection::calculateSectionArea(const double x1, const double y1, const double x2, const double y2, const double stage) const
{
    //note this works even if the section is a triangle or not.
    return((stage - 0.5*(y2+y1))*abs(x2-x1));
}

double CrossSection::calculateIntersect(const double x1, const double y1, const double x2, const double y2, const double stage) const
{
    return (stage - y1)*(x2-x1)/(y2-y1)+x1;
}



void CrossSection::calculateHydraulicParameters(const int i)
{
    double stage = mHydraulicParameters->at(static_cast<unsigned int>(i))->getStage();
    double minElevation = getMinPointElevation();
    if(stage > minElevation){
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setDepthMax(stage - minElevation);
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setArea(calculateArea(stage));
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setWettedPerimeter(calculateWettedPerimeter(stage));
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setWidth(calculateWidth(stage));
    }
    else {
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setDepthMax(0);
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setArea(0);
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setWettedPerimeter(0);
        mHydraulicParameters->at(static_cast<unsigned int>(i))->setWidth(0);
    }
}
