//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include <QSqlDatabase>
#include <QSqlQuery>
#include <QVariant>
#include <QDebug>

#include "crosssectiondao.h"
#include "gatdatamanager.h"
#include "alignment.h"

CrossSectionDao::CrossSectionDao(QSqlDatabase & database): mDatabase(database)
{

}

void CrossSectionDao::init() const
{
    if(!mDatabase.tables().contains("crossSections")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE crossSections (crossSectionId INTEGER PRIMARY KEY AUTOINCREMENT, "
                   "description TEXT, "
                   "d84 REAL, "
                   "slope REAL, "
                   "tolerance REAL, "
                   "filterString TEXT, "
                   "channelFeature INT, "
                   "alignmentId INTEGER NOT NULL REFERENCES alignments(alignmentId), "
                   "lastModified DATETIME)");
            GATDataManager::debugQuery(query);
    }
    if(!mDatabase.tables().contains("crossSectionHydraulicStages")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE crossSectionHydraulicStages (crossSectionId INTEGER, "
                   "stage REAL, "
                   "mannings REAL, "
                   "description TEXT,"
                   "lastModified DATETIME)");
            GATDataManager::debugQuery(query);
    }
    if(!mDatabase.tables().contains("crossSectionPoints")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE crossSectionPoints ("
                   "northing REAL,"
                   "easting REAL,"
                   "elevation REAL,"
                   "stationing REAL,"
                   "sortOrder INTEGER NOT NULL, "
                   "description TEXT, "
                   "pointNumber REAL NOT NULL REFERENCES points ( pointNumber ), "
                   "crossSectionId INTEGER NOT NULL REFERENCES crossSections( crossSectionId ), "
                   "lastModified DATETIME)");
        GATDataManager::debugQuery(query);
    }
    if(!mDatabase.tables().contains("crossSectionMeasurements")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE crossSectionMeasurements ("
                   "measurementId INTEGER PRIMARY KEY AUTOINCREMENT, "
                   "pointOneX REAL, "
                   "pointOneY REAL, "
                   "pointTwoX REAL, "
                   "pointTwoY REAL,"
                   "type INTEGER, "
                   "crossSectionId INTEGER NOT NULL REFERENCES crossSections( crossSectionId))");
        GATDataManager::debugQuery(query);
    }
}

int CrossSectionDao::addCrossSection(std::unique_ptr<CrossSection> &xs) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO crossSections(description, "
                  "d84, "
                  "slope, "
                  "tolerance, "
                  "filterString, "
                  "alignmentId, "
                  "channelFeature, "
                  "lastModified ) VALUES (?, ?, ?, ?, ?, ?, ?, datetime('now'))");
    //qDebug() << "CrossSectionDao::addCrossSection about to bind value";
    query.bindValue(0, xs->getName());
    query.bindValue(1, xs->getd84());
    query.bindValue(2, xs->getSlope());
    query.bindValue(3, xs->getTolerance());
    query.bindValue(4, xs->getFilterString());
    query.bindValue(5, xs->getAlignmentId());

    //Place holder
    query.bindValue(6, 0);
    query.exec();
    //qDebug() << "CrossSectionDao::addCrossSection query executed";

    GATDataManager::debugQuery(query);
    int id = query.lastInsertId().toInt();
    xs->setId(id);
    addCrossSectionPoints(xs);
    return id;

    //qDebug() << "CrossSectionDao::addCrossSection exiting";
}

void CrossSectionDao::removeCrossSection(int id) const
{
    //Remove the cross section points
    removeCrossSectionPoints(id);
    //Remove the cross section hydraulic parameters
    removeCrossSectionHydraulicParameters(id);
    //remove the cross section
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM crossSections WHERE crossSectionId=(?)");
    query.bindValue(0, id);
    query.exec();
    GATDataManager::debugQuery(query);
}

void CrossSectionDao::updateCrossSection(std::unique_ptr<CrossSection> &xs) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE crossSections SET description = ?,"
                  "d84 = ?, "
                  "slope = ?, "
                  "tolerance = ?, "
                  "filterString = ?, "
                  "alignmentId = ?, "
                  "channelFeature =?, "
                  "lastModified = datetime('now') "
                  "WHERE crossSectionId = ?");
    query.bindValue(0, xs->getName());
    query.bindValue(1, xs->getd84());
    query.bindValue(2, xs->getSlope());
    query.bindValue(3, xs->getTolerance());
    query.bindValue(4, xs->getFilterString());
    query.bindValue(5, xs->getAlignmentId());
    query.bindValue(6, 0);
    query.bindValue(7, xs->getId());
    query.exec();
    GATDataManager::debugQuery(query);
    //updateCrossSectionHydraulicParameters(xs);
    //updateCrossSectionPoints(xs);
    //GATDataManager::debugQuery(query);
    return;
}

std::unique_ptr<std::vector<std::unique_ptr<CrossSection>>> CrossSectionDao::getCrossSections() const
{
    QSqlQuery query("SELECT * FROM crossSections", mDatabase);
    query.exec();
    std::unique_ptr<std::vector<std::unique_ptr<CrossSection>>> list(new std::vector<std::unique_ptr<CrossSection>>());
    while(query.next()){
        std::unique_ptr<CrossSection> xs(new CrossSection());
        xs->setId(query.value("crossSectionId").toInt());
        xs->setName(query.value("description").toString());
        xs->setD84(query.value("d84").toDouble());
        xs->setSlope(query.value("slope").toDouble());
        xs->setTolerance(query.value("tolerance").toDouble());
        xs->setPointVector(getPoints(xs->getId()));
        xs->setAlignmentId(query.value("alignmentId").toInt());
        xs->setHydraulicParameterList(this->getHydraulicParameters(xs->getId()));
        xs->setMeasurementList(this->getProfileMeasurments(xs->getId()));
        list->push_back(move(xs));
    }

    return list;
}

std::unique_ptr<std::vector<std::unique_ptr<ProjectedPoint> > > CrossSectionDao::getPoints(const int xsId) const
{
    std::unique_ptr<std::vector<std::unique_ptr<ProjectedPoint>>> list(new std::vector<std::unique_ptr<ProjectedPoint>>());
    QSqlQuery query("SELECT northing, "
                    "easting, "
                    "elevation, "
                    "stationing, "
                    "description, "
                    "pointNumber FROM crossSectionPoints WHERE crossSectionId = ? ORDER BY sortOrder", mDatabase);
    query.bindValue(0, xsId);
    query.exec();
    GATDataManager::debugQuery(query);
    while(query.next()){
        std::unique_ptr<ProjectedPoint> newPoint(new ProjectedPoint);
        newPoint->setNorthing(query.value("northing").toDouble());
        newPoint->setEasting(query.value("easting").toDouble());
        newPoint->setElevation(query.value("elevation").toDouble());
        newPoint->setStation(query.value("stationing").toDouble());
        newPoint->setDescription(query.value("description").toString());
        newPoint->setOriginatingPointId(query.value("pointNumber").toInt());
        list->push_back(move(newPoint));
    }
    return list;
}

std::unique_ptr<std::vector<std::unique_ptr<HydraulicParameters>>> CrossSectionDao::getHydraulicParameters(const int crossSectionId) const
{
    std::unique_ptr<std::vector<std::unique_ptr<HydraulicParameters>>> hydraulicParametersVector(new std::vector<std::unique_ptr<HydraulicParameters>>());
    QSqlQuery hydraulicStagesQuery(mDatabase);
    hydraulicStagesQuery.prepare("SELECT * FROM crossSectionHydraulicStages WHERE crossSectionId = ?");
    hydraulicStagesQuery.bindValue(0,crossSectionId);
    hydraulicStagesQuery.exec();
    GATDataManager::debugQuery(hydraulicStagesQuery);
    while(hydraulicStagesQuery.next()){
        std::unique_ptr<HydraulicParameters> hydraulicParameter(new HydraulicParameters);
        hydraulicParameter->setStage(hydraulicStagesQuery.value("stage").toDouble());
        hydraulicParameter->setMannings(hydraulicStagesQuery.value("mannings").toDouble());
        hydraulicParameter->setDescription(hydraulicStagesQuery.value("description").toString());
        hydraulicParametersVector->push_back(move(hydraulicParameter));
    }
    return hydraulicParametersVector;
}

std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement>>> CrossSectionDao::getProfileMeasurments(const int crossSectionId) const
{
    std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement>>> profileMeasurementsVector(new std::vector<std::unique_ptr<ProfileMeasurement>>);
    QSqlQuery query(mDatabase);
    query.prepare("SELECT * FROM crossSectionMeasurements WHERE crossSectionId = ?");
    query.bindValue(0, crossSectionId);
    query.exec();
    GATDataManager::debugQuery(query);
    while(query.next()){
        std::unique_ptr<ProfileMeasurement> measurement(new ProfileMeasurement);
        measurement->setPointOne(query.value("pointOneX").toDouble(), query.value("pointOneY").toDouble());
        measurement->setPointTwo(query.value("pointTwoX").toDouble(), query.value("pointTwoY").toDouble());
        measurement->setType(query.value("type").toInt());
        measurement->setMeasurementId(query.value("measurementId").toInt());
        profileMeasurementsVector->push_back(move(measurement));
    }
    return profileMeasurementsVector;
}

void CrossSectionDao::updateCrossSectionHydraulicParameters(std::unique_ptr<CrossSection> &xs) const
{
    removeCrossSectionHydraulicParameters(xs->getId());
    for(int i = 0 ; i < xs->getHydraulicParameterCount(); i++){
        QSqlQuery query(mDatabase);
        query.prepare("INSERT INTO crossSectionHydraulicStages(crossSectionId, stage, mannings, description, lastModified ) VALUES (?, ?, ? , ?, datetime('now'))");
        query.bindValue(0, xs->getId());
        query.bindValue(1, xs->getStage(i));
        query.bindValue(2, xs->getManningN(i));
        query.bindValue(3, xs->getStageDescription(i));
        query.exec();
        GATDataManager::debugQuery(query);
    }

}

void CrossSectionDao::updateMeasurement(std::unique_ptr<CrossSection> &xs, int i) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE crossSectionMeasurements SET "
                  "pointOneX = ?, "
                  "pointOneY = ?, "
                  "pointTwoX = ?, "
                  "pointTwoY = ?, "
                  "type = ? "
                  "WHERE measurementId = ?");
    query.bindValue(0, xs->getPointOneX(i));
    query.bindValue(1, xs->getPointOneY(i));
    query.bindValue(2, xs->getPointTwoX(i));
    query.bindValue(3, xs->getPointTwoY(i));
    query.bindValue(4, xs->getPointType(i));
    query.bindValue(5, xs->getMeasurementId(i));
    query.exec();
    GATDataManager::debugQuery(query);
}

void CrossSectionDao::addCrossSectionPoints(std::unique_ptr<CrossSection> &xs) const
{
    if(xs->getPointCount()==0){
        return;
    }
    for(int i = 0; i < xs->getPointCount(); i++){
        QSqlQuery query(mDatabase);
        query.prepare("INSERT INTO crossSectionPoints ("
                      "northing, "
                      "easting, "
                      "elevation, "
                      "stationing, "
                      "sortOrder, "
                      "description, "
                      "pointNumber, "
                      "crossSectionId, "
                      "lastModified) "
                      "VALUES "
                      "(?, ?, ?, ?, ?, ?, ?, ?, datetime('now'))"
                      );
        query.bindValue(0, xs->getNorthing(i));
        query.bindValue(1, xs->getEasting(i));
        query.bindValue(2, xs->getElevation(i));
        query.bindValue(3, xs->getStationing(i));
        query.bindValue(4, i);
        query.bindValue(5, xs->getDescription(i));
        query.bindValue(6, xs->getOriginatingPointId(i));
        query.bindValue(7, xs->getId());
        query.exec();
        GATDataManager::debugQuery(query);
        qDebug() << "CrossSectionDao::addCrossSectionPoints " << xs->getDescription(i);
    }
    return;
}

void CrossSectionDao::removeCrossSectionPoints(const int id) const
{
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM crossSectionPoints WHERE crossSectionId=(?)");
    query.bindValue(0,id);
    query.exec();
    GATDataManager::debugQuery(query);
}

void CrossSectionDao::updateCrossSectionPoints(std::unique_ptr<CrossSection> &xs) const
{
    removeCrossSectionPoints(xs->getId());
    addCrossSectionPoints(xs);
}

void CrossSectionDao::updateProfilePoint(const std::unique_ptr<CrossSection> &xs, const int pn) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE crossSectionPoints SET "
                  "northing=?, "
                  "easting=?, "
                  "elevation=?, "
                  "stationing=?, "
                  "description=?, "
                  "pointNumber=?, "
                  "lastModified = dateTime('now') "
                  "WHERE longProfileId = ? and sortOrder =?");
        query.bindValue(0,xs->getNorthing(pn));
        query.bindValue(1,xs->getEasting(pn));
        query.bindValue(2, xs->getElevation(pn));
        query.bindValue(3, xs->getStationing(pn));
        query.bindValue(4, xs->getDescription(pn));
        query.bindValue(5, xs->getOriginatingPointId(pn));
        query.bindValue(6, xs->getId());
        query.bindValue(7,pn);
        query.exec();
        GATDataManager::debugQuery(query);
}

void CrossSectionDao::addCrossSectionHydraulicParameters(std::unique_ptr<CrossSection> & xs, int i) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO crossSectionHydraulicStages(crossSectionId, stage, mannings, description, lastModified ) VALUES (?, ?, ? , ?, datetime('now'))");
    query.bindValue(0, xs->getId());
    query.bindValue(1, xs->getStage(i));
    query.bindValue(2, xs->getManningN(i));
    query.bindValue(3, xs->getStageDescription(i));
    query.exec();
    GATDataManager::debugQuery(query);
}

void CrossSectionDao::addMeasurement(std::unique_ptr<CrossSection> &xs, int i) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO crossSectionMeasurements ("
                  "pointOneX, pointOneY, "
                  "pointTwoX, pointTwoY, "
                  "type, crossSectionId) VALUES "
                  "(?, ? , ?, ? ,? ,?) ");
    query.bindValue(0,xs->getPointOneX(i));
    query.bindValue(1,xs->getPointOneY(i));
    query.bindValue(2,xs->getPointTwoX(i));
    query.bindValue(3,xs->getPointTwoY(i));
    query.bindValue(4,static_cast<int>(xs->getPointType(i)));
    query.bindValue(5,xs->getId());
    query.exec();
    GATDataManager::debugQuery(query);
    xs->setMeasurementId(i,query.lastInsertId().toInt());
}

void CrossSectionDao::removeCrossSectionHydraulicParameters(const int id) const
{
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM crossSectionHydraulicStages WHERE crossSectionId=(?)");
    query.bindValue(0,id);
    query.exec();
}

bool CrossSectionDao::removeMeasurements(const std::unique_ptr<CrossSection> &xs, int row, int count) const
{
    bool returnValue = false;
    for(int i = row; i < row + count; i++){
        QSqlQuery query(mDatabase);
        query.prepare("DELETE FROM crossSectionMeasurements WHERE measurementId =(?)");
        query.bindValue(0,xs->getMeasurementId(i));
        returnValue = query.exec();
        GATDataManager::debugQuery(query);
    }
    return returnValue;
}
