#include "gatlandxml.h"
#include <QDebug>
#include "alignmentpoint.h"
#include <memory>

GATLandXML::GATLandXML(QString fileName)
{
    inLandXMLFile = new QFile(fileName);
    std::unique_ptr<Alignment> temp(new Alignment());
    alignment.swap(temp);
}

void GATLandXML::readXML()
{
    if(!inLandXMLFile->open(QIODevice::ReadOnly | QIODevice::Text)){
        return;
    }
    xmlReader.setDevice(inLandXMLFile);
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name()=="LandXML"){
            processLandXML();
        }
    }
}

void GATLandXML::readAlignmentXML()
{
    if(!xmlReader.isStartElement() || xmlReader.name() != "Alignment"){
        return;
    }
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name() == "CoordGeom"){
            processCoordGeom();
        }
        else {
            xmlReader.skipCurrentElement();
        }
    }
}

void GATLandXML::processLandXML()
{
    qDebug() << "GATLandXML::processLandXML";
    if(!xmlReader.isStartElement() || xmlReader.name() != "LandXML"){
        return;
    }
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name() == "Alignments"){
            processAlignments();
        }
        else {
            xmlReader.skipCurrentElement();
        }
    }
}

void GATLandXML::processAlignments()
{
    qDebug() << "GATLandXML::processAlignments";
    if(!xmlReader.isStartElement() || xmlReader.name() != "Alignments"){
        return;
    }
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name() == "Alignment"){
            alignment->setDescription(xmlReader.attributes().value("name").toString());
            processAlignment();
        }
        else {
            xmlReader.skipCurrentElement();
        }
    }
}

void GATLandXML::processAlignment()
{
    qDebug() << "GATLandXML::processAlignment";
    if(!xmlReader.isStartElement() || xmlReader.name() != "Alignment"){
        return;
    }
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name() == "CoordGeom"){
            processCoordGeom();
        }
        else {
            xmlReader.skipCurrentElement();
        }
    }
}

void GATLandXML::processCoordGeom()
{
    qDebug() << "GATLandXML::processCoordGeom";
    if(!xmlReader.isStartElement() || xmlReader.name() != "CoordGeom"){
        return;
    }
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name() == "Line"){
            processLine();
        }
        else if(xmlReader.name() == "Curve"){
            processCurve();
        }
    }
}

void GATLandXML::processLine()
{
    qDebug() << "GATLandXML::processLine";
    if(!xmlReader.isStartElement() || xmlReader.name() != "Line"){
        return;
    }
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name() == "Start"){
            QStringList pointList = xmlReader.readElementText().split(' ');
            if(pointList.length() == 2){
                AlignmentPoint point;
                point.setNorthing(pointList.at(0).toDouble());
                point.setEasting(pointList.at(1).toDouble());
                point.setPointType(AlignmentPoint::Line);
                alignment->addPoint(point);
                qDebug() << "Start Northing: " << pointList.at(0) << "easting: " << pointList.at(1);
            }
            else {
                qDebug() << "More than two elements " << xmlReader.readElementText();
            }
        }
        else if(xmlReader.name() == "End"){
            QStringList pointList = xmlReader.readElementText().split(' ');
            if(pointList.length()==2){
                //Store these points in case this is the last element.
                endNorthing = pointList.at(0).toDouble();
                endEasting = pointList.at(1).toDouble();
            }
            qDebug() << "END: " << xmlReader.readElementText();
        }
        //xmlReader.readNextStartElement();
    }
}

void GATLandXML::processCurve()
{
    qDebug() << "GATLandXML::processCurve";
    if(!xmlReader.isStartElement() || xmlReader.name() != "Curve"){
        return;
    }
    AlignmentPoint point;
    if(xmlReader.attributes().value("rot").toString() == "cw"){
        point.setPointType(AlignmentPoint::CurveRight);
    }
    else if(xmlReader.attributes().value("rot").toString() == "ccw"){
        point.setPointType(AlignmentPoint::CurveLeft);
    }
    while(xmlReader.readNextStartElement()){
        if(xmlReader.name() == "Start"){
            QStringList pointList = xmlReader.readElementText().split(' ');
            if(pointList.length() == 2){
                point.setNorthing(pointList.at(0).toDouble());
                point.setEasting(pointList.at(1).toDouble());
            }
            else {
                qDebug() << "More than two elements " << xmlReader.readElementText();
            }
        }
        else if(xmlReader.name() == "End"){
            QStringList pointList = xmlReader.readElementText().split(' ');
            if(pointList.length() == 2){
                endNorthing = pointList.at(0).toDouble();
                endEasting = pointList.at(1).toDouble();
            }
            else {
                qDebug() << "More than two elements " << xmlReader.readElementText();
            }
        }

        else if(xmlReader.name() == "Center"){
            QStringList pointList = xmlReader.readElementText().split(' ');
            if(pointList.length() == 2){
                point.setCenterNorthing(pointList.at(0).toDouble());
                point.setCenterEasting(pointList.at(1).toDouble());
            }
            else {
                qDebug() << "More than two elements " << xmlReader.readElementText();
            }
        }
        else if(xmlReader.name() == "PI"){
            qDebug() << "PI: " << xmlReader.readElementText();
            qDebug() << "Not doing anything with this at this time";
        }
        //xmlReader.readNextStartElement();
    }
    alignment->addPoint(point);
}

std::unique_ptr<Alignment> GATLandXML::getAlignment()
{
    //this will reset the reader to be ready for the next round.
    std::unique_ptr<Alignment> returnAlignment(new Alignment());
    returnAlignment.swap(alignment);
    return returnAlignment;
}

