#include "longprofiledao.h"
#include <QSqlQuery>
#include <QDebug>

#include "gatdatamanager.h"

LongProfileDao::LongProfileDao(QSqlDatabase & database): mDatabase(database)
{

}

void LongProfileDao::init() const
{
    if(!mDatabase.tables().contains("longProfiles")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE longProfiles (longProfileId INTEGER PRIMARY KEY AUTOINCREMENT,"
                   "description TEXT,"
                   "tolerance REAL,"
                   "filterString TEXT,"
                   "lastModified DATETIME)");
        GATDataManager::debugQuery(query);
    }
    if(!mDatabase.tables().contains("longProfilePoints")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE longProfilePoints("
                   "northing REAL, "
                   "easting REAL, "
                   "elevation REAL, "
                   "stationing REAL, "
                   "sortOrder INTEGER NOT NULL, "
                   "description TEXT, "
                   "pointNumber REAL NULL REFERENCES points (pointNumber), "
                   "longProfileId INTEGER NOT NULL REFERENCES longProfiles(longProfileId), "
                   "lastModified DATETIME)");
        GATDataManager::debugQuery(query);
    }
    if(!mDatabase.tables().contains("longProfileMeasurements")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE longProfileMeasurements ("
                   "measurementId INTEGER PRIMARY KEY AUTOINCREMENT, "
                   "pointOneX REAL, "
                   "pointOneY REAL, "
                   "pointTwoX REAL, "
                   "pointTwoY REAL,"
                   "type INTEGER, "
                   "longProfileId INTEGER NOT NULL REFERENCES longProfiles( longProfileId))");
        GATDataManager::debugQuery(query);
    }
}

int LongProfileDao::addLongProfile(std::unique_ptr<CrossSection> &profile) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO longProfiles(description, tolerance, filterString, lastModified ) VALUES (?, ?, ?, datetime('now'))");
    //qDebug() << "LongProfileDao::addCrossSection about to bind value";
    query.bindValue(0, profile->getName());
    query.bindValue(1, profile->getTolerance());
    query.bindValue(2, profile->getFilterString());
    query.exec();

    GATDataManager::debugQuery(query);
    int id = query.lastInsertId().toInt();
    profile->setId(id);
    addProfilePoints(profile);
    return id;
}

void LongProfileDao::updateLongProfile(std::unique_ptr<CrossSection> &profile) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE longProfiles SET description = ?, "
                  "tolerance = ?, "
                  "filterString = ?, "
                  "lastModified = datetime('now') "
                  "WHERE longProfileId = ?");
    query.bindValue(0, profile->getName());
    query.bindValue(1, profile->getTolerance());
    query.bindValue(2, profile->getFilterString());
    query.bindValue(3, profile->getId());
    query.exec();
    GATDataManager::debugQuery(query);

}

void LongProfileDao::removeLongProfie(const int id) const
{
    removeProfilePoints(id);
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM longProfiles WHERE longProfileId =(?)");
    query.bindValue(0,id);
    query.exec();
    GATDataManager::debugQuery(query);
}
std::unique_ptr<std::vector<std::unique_ptr<CrossSection> > > LongProfileDao::getLongProfiles() const
{
    QSqlQuery query("SELECT * FROM longProfiles", mDatabase);
    query.exec();
    std::unique_ptr<std::vector<std::unique_ptr<CrossSection>>> list(new std::vector<std::unique_ptr<CrossSection>>());
    while(query.next()){
        std::unique_ptr<CrossSection> xs(new CrossSection());
        xs->setId(query.value("longProfileId").toInt());
        xs->setName(query.value("description").toString());
        xs->setTolerance(query.value("tolerance").toDouble());
        xs->setFilterString(query.value("filterString").toString());
        xs->setPointVector(getPoints(xs->getId()));
        xs->setMeasurementList(getProfileMeasurments(xs->getId()));
        xs->calculateMaxPointElevation();
        xs->calculateMinPointElevation();
        list->push_back(move(xs));
    }

    return list;
}

std::unique_ptr<std::vector<std::unique_ptr<ProjectedPoint> > > LongProfileDao::getPoints(const int xsId) const
{
    qDebug() << "LongProfileDao::getPoints enter";
    std::unique_ptr<std::vector<std::unique_ptr<ProjectedPoint>>> list(new std::vector<std::unique_ptr<ProjectedPoint>>());
    QSqlQuery query("SELECT northing, "
                    "easting, "
                    "elevation, "
                    "stationing, "
                    "description, "
                    "pointNumber FROM "
                    "longProfilePoints WHERE longProfileId = ? ORDER BY sortOrder", mDatabase);
    query.bindValue(0, xsId);
    query.exec();
    GATDataManager::debugQuery(query);
    while(query.next()){
        std::unique_ptr<ProjectedPoint> newPoint(new ProjectedPoint);
        newPoint->setNorthing(query.value("northing").toDouble());
        newPoint->setEasting(query.value("easting").toDouble());
        newPoint->setElevation(query.value("elevation").toDouble());
        newPoint->setStation(query.value("stationing").toDouble());
        newPoint->setDescription(query.value("description").toString());
        newPoint->setOriginatingPointId(query.value("pointNumber").toInt());
        list->push_back(move(newPoint));
    }
    return list;
}

void LongProfileDao::addProfilePoints(std::unique_ptr<CrossSection> &profile) const
{
    if(profile->getPointCount()==0){
        return;
    }
    for(int i = 0; i < profile->getPointCount(); i++){
        QSqlQuery query(mDatabase);
        query.prepare("INSERT INTO longProfilePoints ("
                      "northing, "
                      "easting, "
                      "elevation, "
                      "stationing, "
                      "sortOrder, "
                      "description, "
                      "pointNumber, "
                      "longProfileId, "
                      "lastModified) "
                      "VALUES "
                      "(?, ?, ?, ?, ?, ?, ?, ?, datetime('now'))"
                      );
        query.bindValue(0, profile->getNorthing(i));
        query.bindValue(1, profile->getEasting(i));
        query.bindValue(2, profile->getElevation(i));
        query.bindValue(3, profile->getStationing(i));
        query.bindValue(4, i);
        query.bindValue(5, profile->getDescription(i));
        query.bindValue(6, profile->getOriginatingPointId(i));
        query.bindValue(7, profile->getId());
        query.exec();
        GATDataManager::debugQuery(query);
        qDebug() << "LongProfileDao::addProfilePoints " << profile->getDescription(i);
    }
    return;
}

void LongProfileDao::updateProfilePoints(std::unique_ptr<CrossSection> &profile) const
{
    removeProfilePoints(profile->getId());
    addProfilePoints(profile);
}

void LongProfileDao::updateProfilePoint(const std::unique_ptr<CrossSection> &profile, const int pn) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE longProfilePoints SET northing = ?, "
                  "easting = ?, "
                  "elevation = ?, "
                  "stationing = ?, "
                  "description = ?, "
                  "pointNumber = ?, "
                  "lastModified = datetime('now') "
                  "WHERE longProfileId = ? AND sortOrder = ?");
    query.bindValue(0,profile->getNorthing(pn));
    query.bindValue(1, profile->getEasting(pn));
    query.bindValue(2, profile->getElevation(pn));
    query.bindValue(3, profile->getStationing(pn));
    query.bindValue(4, profile->getDescription(pn));
    query.bindValue(5, profile->getOriginatingPointId(pn));
    query.bindValue(6, profile->getId());
    query.bindValue(7, pn);
    query.exec();
    GATDataManager::debugQuery(query);
}

bool LongProfileDao::removeMeasurements(const std::unique_ptr<CrossSection> &profile, int row, int count) const
{
    bool returnValue = false;
    for(int i = row; i < row + count; i++){
        QSqlQuery query(mDatabase);
        query.prepare("DELETE FROM longProfileMeasurements WHERE measurementId =(?)");
        query.bindValue(0,profile->getMeasurementId(i));
        returnValue = query.exec();
        GATDataManager::debugQuery(query);
    }
    return returnValue;
}

void LongProfileDao::addMeasurement(std::unique_ptr<CrossSection> &profile, int i) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO longProfileMeasurements ("
                  "pointOneX, pointOneY, "
                  "pointTwoX, pointTwoY, "
                  "type, longProfileId) VALUES "
                  "(?, ? , ?, ? ,? ,?) ");
    query.bindValue(0,profile->getPointOneX(i));
    query.bindValue(1,profile->getPointOneY(i));
    query.bindValue(2,profile->getPointTwoX(i));
    query.bindValue(3,profile->getPointTwoY(i));
    query.bindValue(4,profile->getPointType(i));
    query.bindValue(5,profile->getId());
    query.exec();
    GATDataManager::debugQuery(query);
    profile->setMeasurementId(i,query.lastInsertId().toInt());
}

void LongProfileDao::updateMeasurement(std::unique_ptr<CrossSection> &xs, int i) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE longProfileMeasurements SET "
                  "pointOneX = ?, "
                  "pointOneY = ?, "
                  "pointTwoX = ?, "
                  "pointTwoY = ?, "
                  "type = ? "
                  "WHERE measurementId = ?");
    query.bindValue(0, xs->getPointOneX(i));
    query.bindValue(1, xs->getPointOneY(i));
    query.bindValue(2, xs->getPointTwoX(i));
    query.bindValue(3, xs->getPointTwoY(i));
    query.bindValue(4, xs->getPointType(i));
    query.bindValue(5, xs->getMeasurementId(i));
    query.exec();
    GATDataManager::debugQuery(query);
}

void LongProfileDao::removeProfilePoints(const int id) const
{
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM longProfilePoints WHERE longProfileId = (?)");
    query.bindValue(0, id);
    query.exec();
    GATDataManager::debugQuery(query);
}

std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement> > > LongProfileDao::getProfileMeasurments(const int Id) const
{
    std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement>>> profileMeasurementsVector(new std::vector<std::unique_ptr<ProfileMeasurement>>);
    QSqlQuery query(mDatabase);
    query.prepare("SELECT * FROM longProfileMeasurements WHERE longProfileId = ?");
    query.bindValue(0, Id);
    query.exec();
    GATDataManager::debugQuery(query);
    while(query.next()){
        std::unique_ptr<ProfileMeasurement> measurement(new ProfileMeasurement);
        measurement->setPointOne(query.value("pointOneX").toDouble(), query.value("pointOneY").toDouble());
        measurement->setPointTwo(query.value("pointTwoX").toDouble(), query.value("pointTwoY").toDouble());
        measurement->setType(query.value("type").toInt());
        measurement->setMeasurementId(query.value("measurementId").toInt());
        qDebug() << "measurement point " << measurement->getPointOne();
        profileMeasurementsVector->push_back(move(measurement));

    }
    return profileMeasurementsVector;
}
