//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "pointsdao.h"

#include <QSqlDatabase>
#include <QSqlQuery>
#include <QVariant>
#include <QString>
#include <QDebug>

#include "point.h"
#include "gatdatamanager.h"

PointsDao::PointsDao(QSqlDatabase & database): mDatabase(database)
{

}

void PointsDao::init() const
{
    if(!mDatabase.tables().contains("points")){
        QSqlQuery query(mDatabase);
        query.exec("CREATE TABLE points (pointNumber INTEGER PRIMARY KEY AUTOINCREMENT, "
                   "northing REAL, easting REAL, elevation REAL, description TEXT, lastModified DATETIME)");
        GATDataManager::debugQuery(query);

    }
}

void PointsDao::addPoint(Point& point) const
{
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO points(northing, easting, elevation, description, lastModified)"
                  " VALUES (?, ?, ?, ?, datetime('now'))");
    query.bindValue(0, point.getNorthing());
    query.bindValue(1, point.getEasting());
    query.bindValue(2, point.getElevation());
    query.bindValue(3, point.getDescription());
    query.exec();
    point.setPointNumber(query.lastInsertId().toInt());
    GATDataManager::debugQuery(query);
}

void PointsDao::addPoints(std::unique_ptr<std::vector<std::unique_ptr<Point>>> &pointList, int first, int last) const
{
//    QVariantList northing;
//    QVariantList easting;
//    QVariantList elevation;
//    QVariantList description;

    qDebug() << "addPoints start: " << QTime::currentTime().toString();
    if(first < 0 || last >= pointList->size()){
        return;
    }
    //reserve to avoid repetative allocation
//    northing.reserve(last-first+1);
//    easting.reserve(last-first+1);
//    elevation.reserve(last-first+1);
//    description.reserve(last-first+1);
    mDatabase.transaction();
    QSqlQuery query(mDatabase);
    query.prepare("INSERT INTO points(northing, easting, elevation, description, lastModified)"
                  " VALUES (?, ?, ?, ?, datetime('now'))");
    for(int i = first; i <= last; i++){
//        northing.append(pointList->at(i)->getNorthing());
//        easting.append(pointList->at(i)->getEasting());
//        elevation.append(pointList->at(i)->getElevation());
//        description.append(pointList->at(i)->getDescription());
        query.bindValue(0,pointList->at(i)->getNorthing());
        query.bindValue(1,pointList->at(i)->getEasting());
        query.bindValue(2,pointList->at(i)->getElevation());
        query.bindValue(3,pointList->at(i)->getDescription());
        query.exec();
    }
    mDatabase.commit();

//    query.addBindValue(northing);
//    query.addBindValue(easting);
//    query.addBindValue(elevation);
//    query.addBindValue(description);

//    if(!query.execBatch()){
//        GATDataManager::debugQuery(query);
//    }
    int Id = query.lastInsertId().toInt();
    for(int i = last; i >= first; i--){
        pointList->at(static_cast<size_t>(i))->setPointNumber(Id);
        Id--;
    }
    qDebug() << "End: " << QTime::currentTime().toString();
    return;
}

void PointsDao::updatePoint(const std::unique_ptr<Point> & point) const
{
    QSqlQuery query(mDatabase);
    query.prepare("UPDATE points SET northing = ?, "
                  "easting = ?, "
                  "elevation = ?, "
                  "description = ?, "
                  "lastModified = datetime('now') "
                  "WHERE pointNumber = ?;");
    qDebug() << "PointsDao::updatePoint query prepared";
    query.bindValue(0, point->getNorthing());
    query.bindValue(1, point->getEasting());
    query.bindValue(2, point->getElevation());
    qDebug() << "PointsDAO::updatePoint description" << point->getDescription();
    query.bindValue(3, point->getDescription());
    query.bindValue(4, point->getPointNumber());
        qDebug() << "PointsDao::updatePoint values bound";
    query.exec();
    GATDataManager::debugQuery(query);
    qDebug() << "PointsDao::updatePoint about to return";
    return;
}

void PointsDao::removePoint(const int id) const
{
    QSqlQuery query(mDatabase);
    query.prepare("DELETE FROM points WHERE pointNumber = ?");
    query.bindValue(0, id);
    query.exec();
    GATDataManager::debugQuery(query);
    //TODO: need to add exceptions
}

std::unique_ptr<std::vector<std::unique_ptr<Point>>> PointsDao::getPoints() const
{
    QSqlQuery query("SELECT * FROM points", mDatabase);
    //query.exec();
    return runQuery(query);
}

std::unique_ptr<std::vector<std::unique_ptr<Point>>> PointsDao::getFilteredPoints(const double minX,
                                                                                  const double minY,
                                                                                  const double maxX,
                                                                                  const double maxY,
                                                                                  const QString &descriptionFilters) const
{
//this assumes the points are normalized

    qDebug() << "PointsDao::getFilteredPoints minx, maxx, miny, maxy " << minX << " " << maxX << " " << minY << " " << maxY;
    QSqlQuery query(mDatabase);
    QString queryText = "SELECT * FROM points WHERE "
                        "northing > ? AND "
                        "northing < ? AND "
                        "easting > ? AND "
                        "easting < ?";
    QStringList descFilters;

    if(descriptionFilters != ""){
        queryText += " AND (";

        //build the where clause
        descFilters = descriptionFilters.split(",");
        //qDebug() << "PointsDao::getFilteredPoints descFilters " << descFilters;
        for(int i = 0; i < descFilters.size() -1; i++){
            queryText += " description LIKE ? OR ";
        }
        queryText += "description LIKE ?)";
    }

    query.prepare(queryText);
    query.bindValue(0, minY);
    query.bindValue(1, maxY);
    query.bindValue(2, minX);
    query.bindValue(3, maxX);
    if(descriptionFilters != ""){
        for(int i = 0; i < descFilters.size(); i++){
            //qDebug() << "PointsDao::getFilteredPoints desFilters " << descFilters.at(i);
            query.bindValue(i+4,  descFilters.at(i));
        }
    }
    return runQuery(query);
}

std::unique_ptr<std::vector<std::unique_ptr<Point>>> PointsDao::runQuery(QSqlQuery &query) const
{
    query.exec();
    GATDataManager::debugQuery(query);
    std::unique_ptr<std::vector<std::unique_ptr<Point>>> list(new std::vector<std::unique_ptr<Point>>());
    while(query.next()){
        std::unique_ptr<Point> point(new Point());
        point->setPointNumber(query.value("pointNumber").toInt());
        point->setNorthing(query.value("northing").toDouble());
        point->setEasting(query.value("easting").toDouble());
        point->setElevation(query.value("elevation").toDouble());
        point->setDescription(query.value("description").toString());
        list->push_back(move(point));
    }
    //qDebug() << "PointsDao::runQuery point count " << list->size();
    return list;
}
