//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "pointsmodel.h"
#include <QTextStream>
#include <QFile>
#include <QString>
#include <QUrl>
#include <QDebug>
#include <QVariant>
#include <QFontMetrics>
#include <QFont>
#include "fileutilities.h"


PointsModel::PointsModel(QObject *parent): QAbstractListModel(parent),
    mDb(GATDataManager::instance()),
    mCoordinateTransform(GATCoordinateTransform::instance()),
    mPoints(mDb.pointsDao.getPoints())
{

}

QModelIndex PointsModel::addPoint(const Point& point)
{
    int rowIndex = rowCount();
    beginInsertRows(QModelIndex(), rowIndex, rowIndex);
    //qDebug() << "Inserting Point";
    std::unique_ptr<Point> newPoint(new Point(point));
    mDb.pointsDao.addPoint(*newPoint);
    mPoints->push_back(move (newPoint));
    //qDebug() << "PointsModel::addPoint about to emit endInsertRows vector size is " << mPoints->size();
    endInsertRows();
    //qDebug() << "PointsModel::addPoint about to return row " << rowIndex;
    return index(rowIndex, 0);
}

void PointsModel::addPointFromNEZD(const double n, const double e, const double el, const QString d){
    qDebug() << "In addPointFrom NEZD";
    Point *newPoint = new Point(n,e,el, d);
    addPoint(*newPoint);
}

void PointsModel::addPointFromMouse(const QPointF mousePoint)
{
    //Need to have the dialog for the description
    qDebug() << "PointsModel::addPointFromMouse adding point";
    QPointF transformedPoint = mCoordinateTransform.toDataCoordinates(mousePoint);
    Point *newPoint = new Point(transformedPoint.y(),transformedPoint.x(), 0,"");
    addPoint(*newPoint);
}

//File import code adapted from: https://evileg.com/en/post/158/ accessed July 18, 2018
bool PointsModel::addPointsFromFile(const QString& delimiter, const QString& fileName){

    //QString fileName = url.toLocalFile();

    QFile file(fileName);
    if( !file.open(QFile::ReadOnly |QFile::Text)){
        qDebug() << "File doesn't exist";
        file.close();
        return false;
    } else {
        QTextStream in(&file);
        //first check to make sure there aren't any problems
        int pointCount = 0;
        qDebug() << "PointsModel::addPointsFromFile begin check file " << QTime::currentTime().toString();
        while(!in.atEnd())
        {
            QString line = in.readLine();
            if(line.split(delimiter).length() != 5){
                qDebug() << "Wrong number of parameters" << line.split(delimiter).length();
                file.close();
                return false;
            }
            pointCount++;
        }
        qDebug() << "PointsModel::addPointsFromFile end check file begin insert" << QTime::currentTime().toString();

        in.seek(0);
        int rowIndex = rowCount();
        //Do all operations in this function to reduce calls to beginInsertRows
        beginInsertRows(QModelIndex(), rowIndex, rowIndex+pointCount-1);
        while(!in.atEnd()){
            QString line = in.readLine();
            QStringList items = line.split(delimiter);
            std::unique_ptr<Point> newPoint(new Point());
            //newPoint->setPointNumber(items.at(0).toInt());
            newPoint->setNorthing(items.at(1).toDouble());
            newPoint->setEasting(items.at(2).toDouble());
            newPoint->setElevation(items.at(3).toDouble());
            newPoint->setDescription(items.at(4));
            mPoints->push_back(move (newPoint));
        }
        qDebug() << "PointsModel::addPointsFromFile end read file begin insert" << QTime::currentTime().toString();
        mDb.pointsDao.addPoints(mPoints, rowIndex, rowIndex+pointCount-1);
        qDebug() << "PointsModel::addPointsFromFile end database begin insert" << QTime::currentTime().toString();
        endInsertRows();
        qDebug() << "PointsModel::addPointsFromFile after endInsertRows" << QTime::currentTime().toString();

    }
    file.close();
    return true;
}

int PointsModel::rowCount(const QModelIndex &parent)const
{
    Q_UNUSED(parent);
    return static_cast<int>(mPoints->size());
}

int PointsModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent);
    return 5;
}

//adjusted this for the table view when I moved away from QtQuick.
QVariant PointsModel::data(const QModelIndex & index, const int role) const
{
    if(!isIndexValid(index)){
        return QVariant();
    }

    const Point& point = *mPoints->at(static_cast<unsigned int>(index.row()));
    if (role == Qt::DisplayRole){
        if(index.column() == 0)
            {
            return point.getPointNumber();
            }
        else if(index.column()==1)
        {
            return QString::number(point.getNorthing(),'f',3);
        }
        else if(index.column()==2)
        {
            return QString::number(point.getEasting(),'f',3);
        }
        else if(index.column()==3)
        {
            return QString::number(point.getElevation(),'f',3);
        }
        else if(index.column()==4)
        {
            return point.getDescription();
        }
        else
            return QVariant();
        }
    else if(role == Roles::IdRole){
            return point.getPointNumber();
    }
    else if(role == Roles::NorthingRole){
            return point.getNorthing();
    }
    else if (role == Roles::EastingRole){
             return point.getEasting();
    }
    else if (role == Roles::ElevationRole){
            return point.getElevation();
    }
    else
            return QVariant();

}

//This could be extended for more roles
bool PointsModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    qDebug() << "PointsModel::setData enter index column: " << index.column();
    if(!isIndexValid(index)){
        return false;
    }
    if (value == "" && index.column() != 4){
        return false;
    }
    if(role == Qt::EditRole){
        if(index.column() == 4){
            mPoints->at(index.row())->setDescription(value.toString());
        }
        else if (index.column() == 1){
            mPoints->at(index.row())->setNorthing(value.toDouble());
        }
        else if(index.column() == 2){
            mPoints->at(index.row())->setEasting(value.toDouble());
        }
        else if(index.column() == 3){
            mPoints->at(index.row())->setElevation(value.toDouble());
        }
        qDebug() << "PointsModel::setData about to update Point";
        mDb.pointsDao.updatePoint(mPoints->at(index.row()));
        emit dataChanged(index,index);
        return true;
    }
    return false;
}

QVariant PointsModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if(role == Qt::DisplayRole){
        if(orientation == Qt::Horizontal){
        if(section == 0){
            return "Id";
        }
        else if (section == 1){
            return "Northing";
        }
        else if (section == 2){
            return "Easting";
        }
        else if (section == 3){
            return "Elevation";
        }
        else if (section == 4){
            return "Description";
        }
        else
            return section;
    }
    if(orientation == Qt::Vertical){
        return section;
    }
    }
    return QVariant();
}

bool PointsModel::removeRows(int row, int count, const QModelIndex &parent){
    if(row<0 || row >= rowCount()
            || count<0
            || (row+count) > rowCount()){
        return false;
    }
    beginRemoveRows(parent, row, row + count -1);
    int countLeft = count;
    while(countLeft--){
        const Point& point = *mPoints->at(row+countLeft);
        mDb.pointsDao.removePoint(point.getPointNumber());
    }
    mPoints->erase(mPoints->begin() + row, mPoints->begin()+row+count);
    endRemoveRows();
    return true;
}

QHash<int, QByteArray> PointsModel::roleNames() const
{
    QHash<int, QByteArray> roles;
    roles[Roles::IdRole] = "id";
    roles[Roles::NorthingRole]= "northing";
    roles[Roles::EastingRole]= "easting";
    roles[Roles::ElevationRole]= "elevation";
    roles[Roles::DescriptionRole]= "description";
    return roles;
}


Qt::ItemFlags PointsModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::ItemIsEnabled;
    if(index.row() >= 0){
      return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable;
    }
    //Point number is not editable
    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

bool PointsModel::isIndexValid(const QModelIndex & index) const
{
    if(index.row() <0 || index.row() >= rowCount()){ //|| !index.isValid()){
        qDebug() << "PointsModel::isIndexValid index is invalid";
        return false;
    }
    return true;
}

void PointsModel::databaseUpdated()
{

    this->beginResetModel();
    mPoints = mDb.pointsDao.getPoints();
    this->endResetModel();
}
