#include "profile.h"
#include <QDebug>
#include <QLineF>
#include <QPointF>
#include <QtMath>

Profile::Profile()
{
    mId = -9;
    mName ="";
    mTolerance =-9;
    std::unique_ptr<std::vector<std::unique_ptr<ProjectedPoint>>> list(new std::vector<std::unique_ptr<ProjectedPoint>>());
    mProfilePointVector.swap(list);

    std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement>>> measurementList(new std::vector<std::unique_ptr<ProfileMeasurement>>());
    mProfileMeasurements.swap(measurementList);
    mPlot = true;
}

Profile::Profile(const QString &name)
{
    mName = name;
}

Profile::Profile(std::unique_ptr<Profile> profile)
{
    mProfilePointVector.swap(profile->mProfilePointVector);
    mProfileMeasurements.swap(profile->mProfileMeasurements);
    mId = profile->mId;
    mName = profile->mName;
    mTolerance = profile->mTolerance;
    mFilterString = profile->mFilterString;
    mPlot = profile->mPlot;
    mAlignmentId = profile->mAlignmentId;
}

void Profile::setName(const QString &name)
{
    mName = name;
}

void Profile::setTolerance(const double tolerance)
{
    mTolerance = tolerance;
}

void Profile::setFilterString(const QString filterString)
{
    mFilterString = filterString;
}

void Profile::setId(const int id)
{
    mId = id;
}

void Profile::setAlignmentId(const int id)
{
    mAlignmentId = id;
}

void Profile::setPlot(const bool plotState)
{
    mPlot = plotState;
}

void Profile::setChannelFeature(const int feat)
{
    mChannelFeature = feat;
}

QString Profile::getName() const
{
    return mName;
}

double Profile::getTolerance() const
{
    return mTolerance;
}

QString Profile::getFilterString() const
{
    return mFilterString;
}

int Profile::getId() const
{
    return mId;
}

int Profile::getAlignmentId() const
{
 return mAlignmentId;
}

bool Profile::getPlot() const
{
    return mPlot;
}

int Profile::getChannelFeature() const
{
    return mChannelFeature;
}

void Profile::setNorthing(const int pn, const double value)
{
    mProfilePointVector->at(static_cast<unsigned int>(pn))->setNorthing(value);
    if(pn < this->getPointCount()-1){
        calculateStationing(pn+1);
    }
}

void Profile::setEasting(const int pn, const double value)
{
    mProfilePointVector->at(static_cast<unsigned int>(pn))->setEasting(value);
    if(pn < this->getPointCount()-1){
        calculateStationing(pn+1);
    }
}

void Profile::setElevation(const int pn, const double value)
{
    mProfilePointVector->at(static_cast<unsigned int>(pn))->setElevation(value);
    calculateMinPointElevation();
    calculateMaxPointElevation();
}

void Profile::setStationing(const int pn, const double value)
{
    mProfilePointVector->at(static_cast<unsigned int>(pn))->setStation(value);
    if(pn < this->getPointCount()-1){
        calculateStationing(pn+1);
    }

}

double Profile::getNorthing(const int pn) const
{
    if (isPointValid(pn)){
        return mProfilePointVector->at(static_cast<unsigned int>(pn))->getNorthing();
    }
    return -9;
}

double Profile::getEasting(const int pn) const
{
    if (isPointValid(pn)){
        return mProfilePointVector->at(static_cast<unsigned int>(pn))->getEasting();
    }
    return -9;
}

double Profile::getElevation(const int pn) const
{
    if (isPointValid(pn)){
        return mProfilePointVector->at(static_cast<unsigned int>(pn))->getElevation();
    }
    return -9;
}

double Profile::getStationing(const int pn) const
{
    if (isPointValid(pn)){
        return mProfilePointVector->at(static_cast<unsigned int>(pn))->getStationing();
    }
    return -9;
}

int Profile::getOriginatingPointId(const int pn) const
{
    if (isPointValid(pn)){
        return mProfilePointVector->at(static_cast<unsigned int>(pn))->getOriginatingPointId();
    }
    return -9;
}

QString Profile::getDescription(const int pn) const
{
    if (isPointValid(pn)){
        return mProfilePointVector->at(static_cast<unsigned int>(pn))->getDescription();
    }
    return "";
}


bool Profile::removePoint(const int pn)
{
    if(pn < getPointCount() && pn >= 0){
        mProfilePointVector->erase(mProfilePointVector->begin()+pn);
        calculateMinPointElevation();
        calculateMaxPointElevation();
        return true;
    }
    return false;
}

bool Profile::removePoints(const int pn, const int count)
{
    if((pn+count) < getPointCount() && pn >=0){
        //qDebug() << "CrossSection::removePoints about to remove points " << pn << " " << count;
        //This is probably a leak
        //qDebug() << "length " << getPointCount();
        mProfilePointVector->erase(mProfilePointVector->begin()+pn,mProfilePointVector->begin()+pn+count);
        calculateMinPointElevation();
        calculateMaxPointElevation();
        return true;
    }
    return false;
}

int Profile::getPointCount() const
{
    if(mProfilePointVector == nullptr){
        qDebug() << "Prfile::getPointCount nullptr";
        return 0;
    }
    return static_cast<int>(mProfilePointVector->size());
}


void Profile::addPoint(std::unique_ptr<ProjectedPoint> point)
{
    if(mMinPointElevation > point->getElevation()){
        mMinPointElevation = point->getElevation();
    }
    if(mMaxPointElevation < point->getElevation()){
        mMaxPointElevation = point->getElevation();
    }
    mProfilePointVector->push_back(move(point));
}

void Profile::setPointVector(std::unique_ptr<std::vector<std::unique_ptr<ProjectedPoint> > > newPointVector)
{
    mProfilePointVector.swap(newPointVector);
}

QRectF Profile::getBoundingRect()
{
    return QRectF(0,mMinPointElevation, getMaxStationing(),mMaxPointElevation-mMinPointElevation);
}

void Profile::setMeasurementList(std::unique_ptr<std::vector<std::unique_ptr<ProfileMeasurement> > > newMeasurementVector)
{
    mProfileMeasurements.swap(newMeasurementVector);
}

void Profile::sortPointVector()
{
    int i = 1;
    while (i < getPointCount()){
        int j = i;
        while (j > 0 && mProfilePointVector->at(static_cast<unsigned int>(j-1))->getStationing() > mProfilePointVector->at(static_cast<size_t>(j))->getStationing()){
            std::unique_ptr<ProjectedPoint> temp;
            temp.swap(mProfilePointVector->at(static_cast<unsigned int>(j-1)));
            mProfilePointVector->at(static_cast<unsigned int>(j-1)).swap(mProfilePointVector->at(static_cast<size_t>(j)));
            mProfilePointVector->at(static_cast<unsigned int>(j)).swap(temp);
            j = j - 1;
        }
        i = i + 1;
    }
}

int Profile::getMeasurementCount() const
{
    if(mProfileMeasurements == nullptr){
        qDebug() << "Prfile::getPointCount nullptr";
        return 0;
    }
    return static_cast<int>(mProfileMeasurements->size());
}

qreal Profile::getPointOneX(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mProfileMeasurements->at(static_cast<unsigned int>(pn))->getPointOneX();
    }
    return -9;
}

qreal Profile::getPointOneY(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mProfileMeasurements->at(static_cast<unsigned int>(pn))->getPointOneY();
    }
    return -9;
}

qreal Profile::getPointTwoX(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mProfileMeasurements->at(static_cast<unsigned int>(pn))->getPointTwoX();
    }
    return -9;
}

qreal Profile::getPointTwoY(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mProfileMeasurements->at(static_cast<unsigned int>(pn))->getPointTwoY();
    }
    return -9;
}

int Profile::getPointType(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mProfileMeasurements->at(static_cast<unsigned int>(pn))->getType();
    }
    //TODO: Make invalid value
    return 0;
}

qreal Profile::getMeasurementValue(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mProfileMeasurements->at(static_cast<unsigned int>(pn))->getMeasurement();
    }
    return -9;
}

int Profile::getMeasurementId(const int pn) const
{
    if(isMeasurementPointValid(pn)){
        return mProfileMeasurements->at(static_cast<unsigned int>(pn))->getMeasurementId();
    }
    return -9;
}

void Profile::setPointOneX(qreal value, const int pn)
{
    if(isMeasurementPointValid(pn)){
            mProfileMeasurements->at(static_cast<unsigned int>(pn))->setPointOneX(value);
    }
}

void Profile::setPointTwoX(qreal value, const int pn)
{
    if(isMeasurementPointValid(pn)){
            mProfileMeasurements->at(static_cast<unsigned int>(pn))->setPointTwoX(value);
    }
}

void Profile::setPointOneY(qreal value, const int pn)
{
    if(isMeasurementPointValid(pn)){
            mProfileMeasurements->at(static_cast<unsigned int>(pn))->setPointOneY(value);
    }
}

void Profile::setPointTwoY(qreal value, const int pn)
{
    if(isMeasurementPointValid(pn)){
            mProfileMeasurements->at(static_cast<unsigned int>(pn))->setPointTwoY(value);
    }
}

void Profile::setPointType(int type, const int pn)
{
    if(isMeasurementPointValid(pn)){
            mProfileMeasurements->at(static_cast<unsigned int>(pn))->setType(type);
    }
}

bool Profile::removeMeasurements(const int row, const int count)
{
    if(isMeasurementPointValid(row) && isMeasurementPointValid(row+count-1)){
        mProfileMeasurements->erase(mProfileMeasurements->begin()+row,mProfileMeasurements->begin()+row+count);
        qDebug() << "Profile::removeMeasurements measurements removed";
        return true;
    }
    qDebug() << "Profile::removeMeasurements measurements not removed";
    return false;
}

void Profile::setMeasurementId(const int pn, const int id) const
{
    if(isMeasurementPointValid(pn)){
        mProfileMeasurements->at(static_cast<unsigned int>(pn))->setMeasurementId(id);
    }
}

void Profile::addMeasurementPoint(const QPointF &pointOne, const QPointF &pointTwo)
{
    std::unique_ptr<ProfileMeasurement> meas(new ProfileMeasurement(pointOne.x(), pointOne.y(), pointTwo.x(), pointTwo.y()));
    mProfileMeasurements->push_back(move(meas));
}

bool Profile::isPointValid(const int pn) const
{
    if(pn >= 0 && pn < getPointCount()){
        return true;
    }
    return false;
}

bool Profile::isMeasurementPointValid(const int pn) const
{
    if(pn >= 0 && pn < getMeasurementCount()){
        return true;
    }
    return false;
}

QPointF Profile::getPointF(const int pn) const
{
    return mProfilePointVector->at(static_cast<unsigned int>(pn))->getPointF();
}

void Profile::calculateStationing(const int startPoint)
{
    if(getPointCount() < startPoint +1){
        //ensures the vector isn't walked off
        return;
    }
    if(startPoint == 0){
            mProfilePointVector->at(0)->setStation(0);
            for(qint32 i = startPoint+1; i < this->getPointCount(); i++){
                QLineF line(mProfilePointVector->at(static_cast<size_t>(i-1))->getPointF(),mProfilePointVector->at(static_cast<size_t>(i))->getPointF());
                mProfilePointVector->at(static_cast<size_t>(i))->setStation(mProfilePointVector->at(static_cast<size_t>(i-1))->getStationing() + line.length());
            }
            return;
    }
    for(qint32 i = startPoint; i < this->getPointCount(); i++){
        QLineF line(mProfilePointVector->at(static_cast<size_t>(i-1))->getPointF(),mProfilePointVector->at(static_cast<size_t>(i))->getPointF());
        mProfilePointVector->at(static_cast<size_t>(i))->setStation(mProfilePointVector->at(static_cast<size_t>(i-1))->getStationing() + line.length());
    }
}

double Profile::getMaxStationing() const
{
    if(getPointCount() > 0){
        return mProfilePointVector->at(static_cast<unsigned int>(getPointCount()-1))->getStationing();
    }
    return 0;
}

double Profile::getMinPointElevation() const
{
    return mMinPointElevation;
}

void Profile::calculateMinPointElevation()
{
    //qDebug() << "CrossSection::calculateMinPointElevation start";
    if(getPointCount() == 0){
       mMinPointElevation = -9;
       return;
    }
    //qDebug() << "CrossSection::calculateMinElevation more than 1 point";
    mMinPointElevation = getElevation(0);
    for(int i = 1; i < getPointCount(); i++){
        if(getElevation(i) < mMinPointElevation){
            mMinPointElevation = getElevation(i);
        }
    }
}

void Profile::calculateMaxPointElevation()
{
    //qDebug() << "CrossSection::calculateMinPointElevation start";
    if(getPointCount() == 0){
       mMaxPointElevation = -9;
       return;
    }
    //qDebug() << "CrossSection::calculateMinElevation more than 1 point";
    mMaxPointElevation = getElevation(0);
    for(int i = 1; i < getPointCount(); i++){
        if(getElevation(i) > mMaxPointElevation){
            mMaxPointElevation = getElevation(i);
        }
    }
}

