//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "profilemodel.h"
#include <QDebug>
#include <QFile>
#include <QMessageBox>
#include "profilemeasurementtypes.h"


ProfileModel::ProfileModel(ProfileType type, QObject *parent) : QAbstractItemModel (parent),
    mDb(GATDataManager::instance())

{
    mType = type;
    if(type == ProfileType::crossSection){
        mProfiles = mDb.crossSectionDao.getCrossSections();
    }
    if(type == ProfileType::longProfile){
        mProfiles = mDb.longProfileDao.getLongProfiles();
    }
}

int ProfileModel::rowCount(const QModelIndex &parent) const
{
    if(isIndexValid(parent)){
        //qDebug() << "CrossSectionModel::rowCount parent is valid using point count";
        if(parent.column() == 0){
            return (mProfiles->at(static_cast<size_t>(parent.row()))->getPointCount());
        }
        else if(parent.column() == 1){
            return mProfiles->at(static_cast<size_t>(parent.row()))->getHydraulicParameterCount();
        }
        else if(parent.column() == 2){
            return mProfiles->at(static_cast<unsigned int>(parent.row()))->getMeasurementCount();
        }
    }

    if(mProfiles == nullptr){
        return 0;
    }
    return static_cast<int>(mProfiles->size());
}

int ProfileModel::columnCount(const QModelIndex &parent) const
{
    //if the parent index is valid then it is the subtable
    //qDebug() << "CrossSectionModel::columnCount parent row " << parent.row();
    if(isIndexValid(parent)){
        if(parent.column() == 0){
            //qDebug() << "CrossSection::columnCount parent is valid 3 columns";
            return 6;
        }
        if(parent.column() == 1){
            //This is the hydraulic parameters
            return 16;
        }
        if(parent.column() == 2){
            //This is the measurement parameters
            return 6;
        }
    }
    return 7;
}

QVariant ProfileModel::data(const QModelIndex &index, const int role) const
{
    if(!isIndexValid(index)){
        return QVariant();
    }


    //Points and hydraulic parameters handle their own data
    if(isIndexValid(index.parent()))
    {
        //qDebug() << "CrossSectionModel::data parent is valid parent column is" << index.parent().column();
        if(index.parent().column() == 0){
            //qDebug() << "CrossSectionModel::data parent column is 0";
            return dataProfilePoints(index, role);
        }
        if(index.parent().column() == 1){
            //qDebug() << "CrossSectionModel::data returning hydraulicParameters Data";
            return dataHydraulicParameters(index, role);
        }
        if(index.parent().column() == 2){
            return dataMeasurementParameters(index,role);
        }
            return QVariant();
    }

    //Editor widgets use this role to get the data for the widget
    //Placed below the processing of the children to allow the children to process their own.
    if(role == Qt::EditRole){
        //override for the mapper
        return data(index);
    }

    if(role == Qt::DisplayRole)
    {
        //qDebug() << "CrossSectionModel::data before parent valid check";
        //qDebug() << "Parent Row" << index.parent().row() << " Column " << index.parent().column();
        //qDebug() << "isIndexValid parent? " << isIndexValid(index.parent());
        switch (index.column())
        {
        case 0:
            return mProfiles->at(static_cast<unsigned int>(index.row()))->getId();
        case 1:
            return mProfiles->at(static_cast<unsigned int>(index.row()))->getName();
        case 2:
            return mProfiles->at(static_cast<unsigned int>(index.row()))->getSlope();
        case 3:
            return mProfiles->at(static_cast<unsigned int>(index.row()))->getd84();
        case 4:
            return mProfiles->at(static_cast<unsigned int>(index.row()))->getTolerance();
        case 5:
            return mProfiles->at(static_cast<unsigned int>(index.row()))->getFilterString();
        case 6:
            return mProfiles->at(static_cast<unsigned int>(index.row()))->getBoundingRect();
        default:
            return QVariant();
        }
    }
    return QVariant();
}

QVariant ProfileModel::headerData(int section, const Qt::Orientation orientation, int role) const
{
    if(role == Qt::DisplayRole){
        if(orientation == Qt::Horizontal){
                switch(section) {
                case 0:
                    return "Id";
                case 1:
                    return "Description";
                case 2:
                    return "Slope";
                case 3:
                    return "D84";
                case 4:
                    return "Tolerance";
                case 5:
                    return "Filter";
                case 6:
                    return "Bounding Rect";
                default:
                    return section;
                }

                }
        if(orientation == Qt::Vertical){
            return section;
        }

    }
    return QVariant();
}

bool ProfileModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(!isIndexValid(index)){
        return false;
    }
    //if the parent is valid then delegate to the correct funciton.
    if(isIndexValid(index.parent()))
    {
        //then it is either a point edit or a hydraulic parameter edit
        if(index.parent().column() == 0)
        {
            return setDataCrossSectionPoints(index, value, role);
        }
        if(index.parent().column() == 1)
        {
            //need to process this as a hydraulic parameter edit.
            return setDataHydraulicParameters(index,value,role);
        }
        if(index.parent().column() == 2){
            return setDataMeasurement(index, value, role);
        }

        //parent column is invalid
        return false;
    }
    qDebug() << "CrossSectionModel::setData Role " << role << " column" << index.column();
    if(role == Roles::DescriptionRole)
    {
        mProfiles->at(static_cast<unsigned int>(index.row()))->setName(value.toString());
        //this automatically sets up the pointer so the data is maintained.
        //xs.setName(value.toString());
        if(mType == ProfileType::crossSection){
            mDb.crossSectionDao.updateCrossSection(mProfiles->at(static_cast<unsigned int>(index.row())));
        }
        else if(mType == ProfileType::longProfile){
            mDb.longProfileDao.updateLongProfile(mProfiles->at(static_cast<unsigned int>(index.row())));
        }
            emit dataChanged(index,index);
        return true;
    }

    else if(role == Qt::EditRole){

        if(index.column() == 1){
            mProfiles->at(static_cast<unsigned int>(index.row()))->setName(value.toString());
            //this automatically sets up the pointer so the data is maintained.
            //xs.setName(value.toString());
            mDb.crossSectionDao.updateCrossSection(mProfiles->at(static_cast<unsigned int>(index.row())));
            emit dataChanged(index,index);
            return true;
        }
        else if(index.column() == 2){
            //Slope
            if(value.canConvert(QMetaType::Double)){
                mProfiles->at(static_cast<unsigned int>(index.row()))->setSlope(value.toDouble());
                mDb.crossSectionDao.updateCrossSection(mProfiles->at(static_cast<unsigned int>(index.row())));
                emit dataChanged(index, index);
                hydraulicParametersUpdated(index);
            }
            else {
                return false;
            }
        }
        else if(index.column() == 3){
            //d84
            if(value.canConvert(QMetaType::Double)){
                mProfiles->at(static_cast<unsigned int>(index.row()))->setD84(value.toDouble());
                mDb.crossSectionDao.updateCrossSection(mProfiles->at(static_cast<unsigned int>(index.row())));
                emit dataChanged(index, index);
                hydraulicParametersUpdated(index);
            }
            else {
                return false;
            }
        }
        else if(index.column() == 4){
            //tolerance
            if(value.canConvert(QMetaType::Double)){
                mProfiles->at(static_cast<unsigned int>(index.row()))->setTolerance(value.toDouble());
                mDb.crossSectionDao.updateCrossSection(mProfiles->at(static_cast<unsigned int>(index.row())));
                emit dataChanged(index, index);
                hydraulicParametersUpdated(index);
            }
            else {
                return false;
            }
        }
        else if(index.column() == 5){
            //FilterString
            if(value.canConvert(QMetaType::QString)){
                mProfiles->at(static_cast<unsigned int>(index.row()))->setFilterString(value.toString());
                mDb.crossSectionDao.updateCrossSection(mProfiles->at(static_cast<unsigned int>(index.row())));
                emit dataChanged(index, index);
                hydraulicParametersUpdated(index);
            }
            else {
                return false;
            }
        }

    }
    return false;
}

bool ProfileModel::removeRows(int row, int count, const QModelIndex &parent)
{
    if(isIndexValid(parent)){
        if(parent.column() == 0){
            return removeRowsPoints(row, count, parent);
        }
        else if(parent.column() == 1){
            return removeRowsHydraulicParameters(row, count, parent);
        }
        else if(parent.column() == 2){
            return removeRowsMeasurements(row, count, parent);
        }
        else {
            return false;
        }

    }
    if(row<0 || row >= rowCount()
            || count<0
            || (row+count) > rowCount()){
        return false;
    }
    beginRemoveRows(parent, row, row + count -1);
    int countLeft = count;
    while(countLeft--){
        const CrossSection& xs = *mProfiles->at(static_cast<unsigned int>(row+countLeft));
        mDb.crossSectionDao.removeCrossSection(xs.getId());
    }
    mProfiles->erase(mProfiles->begin() + row, mProfiles->begin()+row+count);
    endRemoveRows();
    return true;
}

Qt::ItemFlags ProfileModel::flags(const QModelIndex &index) const
{
    if(!index.isValid())
        return Qt::ItemIsEnabled;

    return Qt::ItemIsEnabled | Qt::ItemIsEditable | Qt::ItemIsSelectable;
}

void ProfileModel::databaseUpdated()
{
    this->beginResetModel();
    if(mType == ProfileType::crossSection){
        mProfiles = mDb.crossSectionDao.getCrossSections();
    }
    if(mType == ProfileType::longProfile){
        mProfiles = mDb.longProfileDao.getLongProfiles();
    }
    this->endResetModel();
}

QModelIndex ProfileModel::index(int row, int column, const QModelIndex & parent) const
{
    if(isIndexValid(parent))
    {
        if (hasIndex(row, column, parent)){
            if(parent.column() == 0){
                return createIndex(row, column, static_cast<quintptr>(parent.row()));
            }
            if(parent.column() == 1){
                return createIndex(row, column, static_cast<quintptr>(parent.row())+1000);

            }
            if(parent.column() == 2){
                return createIndex(row,column, static_cast<quintptr>(parent.row())+2000);
            }
        }
    }
    else {
        return createIndex(row, column,static_cast<quintptr>(-1));
    }
    return QModelIndex();
}

void ProfileModel::addCrossSection(std::unique_ptr<CrossSection> xs)
{
    qDebug() << "ProfileModel::addCrossSection filter string " << xs->getFilterString();
    int rowIndex = rowCount();
    beginInsertRows(QModelIndex(), rowIndex, rowIndex);
    //need to add int assignment.
    xs->calculateMinPointElevation();
    xs->calculateMaxPointElevation();
    if(mType == ProfileType::crossSection){
        mDb.crossSectionDao.addCrossSection(xs);
    }
    else if (mType == ProfileType::longProfile){
        mDb.longProfileDao.addLongProfile(xs);
    }

    qDebug() << "CrossSectionModel::addCrossSection dao added the cross section";
    mProfiles->push_back(move(xs));
    endInsertRows();
}

void ProfileModel::addProfile(std::unique_ptr<Profile> profile)
{
    std::unique_ptr<CrossSection> newCrossSection(new CrossSection(move(profile)));
    addCrossSection(move(newCrossSection));
}

bool ProfileModel::hasChildren(const QModelIndex &index) const
{
    //If the index is invalid then it doesn't have any children.
    if(!isIndexValid(index)){
        return false;
    }
    //If the index has a parent then it is a child and has no children in this dataset
    if(isIndexValid(index.parent())){
        return false;
    }
    //Just because it is a parent doesn't mean any points have been added to it
    if(index.column()== 0){
        if(mProfiles->at(static_cast<unsigned int>(index.row()))->getPointCount() > 0){
            return true;
        }
        return false;
    }
    if(index.column()==1){
        if(mProfiles->at(static_cast<unsigned int>(index.row()))->getHydraulicParameterCount()>0){
            return true;
        }
        return false;
    }
    if(index.column() == 2){
        if(mProfiles->at(static_cast<unsigned int>(index.row()))->getMeasurementCount()>0){
            return true;
        }
        return false;
    }
    return false;
}

QModelIndex ProfileModel::parent(const QModelIndex &index) const
{
    if(!index.isValid()){
        // qDebug() << "CrossSectionModel:: index is not valid return invalid index";
        return QModelIndex();
    }

    if(index.internalId() == quintptr(-1)){
        //Then this is a parent so it doesn't have a parent.
        return QModelIndex();
    }
    int column = 0;
    int internalId = static_cast<int>(index.internalId());
    if(internalId >= 2000){
        column = 2;
        internalId = internalId - 2000;
    }
    else if(internalId >= 1000){
        column = 1;
        internalId = internalId - 1000;
    }
    if(internalId < rowCount() && internalId >= 0)
    {
        //qDebug() << "CrossSectionModel:: index is valid return parent index";
        //qDebug() << index;
        return createIndex(internalId,column,quintptr(-1));
    }
    return QModelIndex();
}

bool ProfileModel::isIndexValid(const QModelIndex & index) const
{
    //qDebug() << "AlignmentMoel::isIndexValid parent row is: " << index.parent().row() << " row is: " << index.row();
    if(!index.isValid()){
        return false;
    }
    if(index.row() < 0){
        return false;
    }
    //Has valid parent
    if(index.parent().isValid()){
            if(index.parent().column() == 0){
                if(index.row() >= mProfiles->at(static_cast<size_t>(index.parent().row()))->getPointCount()){
                    return false;
                }
            }
            else if(index.parent().column() == 1){
                if(index.row() >= mProfiles->at(static_cast<size_t>(index.parent().row()))->getHydraulicParameterCount()){
                    return false;
                }
            }
            else if(index.parent().column() == 2){
                if(index.row() >= mProfiles->at(static_cast<size_t>(index.parent().row()))->getMeasurementCount()){
                    return false;
                }
            }
            else {
                return false;
            }
            return true;
        }
    //Doesn't have valid parent
    if(index.row() >= rowCount()){
        return false;
    }
    return true;
}

void ProfileModel::calculateHydraulicParameters(const QModelIndex & parent, double beginStage, double endStage, const double increment)
{
    if(beginStage > endStage){
        double temp = endStage;
        endStage = beginStage;
        beginStage = temp;
    }
    if(!isIndexValid(parent)){
            return;
    }

    qDebug() << "CrossSectionModel::calculateHydraulicParameters" << (endStage-beginStage)/increment +1;
    this->beginInsertRows(parent, rowCount(parent), rowCount(parent) + (endStage-beginStage)/increment +1);
    mProfiles->at(static_cast<unsigned int>(parent.row()))->addHydraulicParameters(beginStage, endStage, increment);
    hydraulicParametersUpdated(parent);
    endInsertRows();
}

void ProfileModel::calculateHydraulicParameter(const QModelIndex &parent, double stage)
{
    if(!parent.isValid()){
        QMessageBox msgBox;
        msgBox.setText("Invalid index, please select cross section to add stage");
        msgBox.exec();
        return;
    }
    if(mType == ProfileType::longProfile){
        QMessageBox msgBox;
        msgBox.setText("Longitudinal profiles don't have hydraulic parameters");
        msgBox.exec();
        return;
    }
    //Correct row
    QModelIndex hydParent = this->index(parent.row(), 1);
    int insertRow = rowCount(hydParent);
    this->beginInsertRows(hydParent, insertRow,insertRow);
    mProfiles->at(static_cast<unsigned int>(parent.row()))->addHydraulicParameters(stage);
    hydraulicParametersUpdated(parent);
    mDb.crossSectionDao.addCrossSectionHydraulicParameters(mProfiles->at(static_cast<unsigned int>(parent.row())), insertRow);
    endInsertRows();
}

void ProfileModel::addMeasurement(const QModelIndex &parent, const QPointF &pointOne, const QPointF &pointTwo)
{
    if(!parent.isValid()){
        QMessageBox msgBox;
        msgBox.setText("Invalid index, please select cross section to add measurement");
        msgBox.exec();
        return;
    }
    //convert this to the column for the measurements
    QModelIndex parentMeas = this->index(parent.row(), 2,QModelIndex());
    int insertRow = rowCount(parentMeas);
    this->beginInsertRows(parentMeas,insertRow,insertRow);
    mProfiles->at(static_cast<unsigned int>(parentMeas.row()))->addMeasurementPoint(pointOne,pointTwo);
    if(mType == ProfileType::crossSection){
        mDb.crossSectionDao.addMeasurement(mProfiles->at(static_cast<unsigned int>(parentMeas.row())),insertRow);
    }
    else if (mType == ProfileType::longProfile){
        mDb.longProfileDao.addMeasurement(mProfiles->at(static_cast<unsigned int>(parentMeas.row())),insertRow);
    }
    endInsertRows();
}

QVariant ProfileModel::dataProfilePoints(const QModelIndex &index, const int role) const
{
    if(role==Qt::DisplayRole){
        const CrossSection& crossSection = *mProfiles->at(static_cast<unsigned int>(index.parent().row()));
        if(index.column() == 0){
            return QString::number(crossSection.getNorthing(index.row()), 'f',3);
        }
        else if(index.column() == 1){
            return QString::number(
                        crossSection.getEasting(index.row()),'f',3);
        }
        else if(index.column() == 2){
            return QString::number(
                        crossSection.getElevation(index.row()),'f', 3);
        }
        else if(index.column() == 3){
            return QString::number(
                        crossSection.getStationing(index.row()),'f',2);
        }
        else if(index.column() == 4){
            return QString::number(crossSection.getOriginatingPointId(index.row()));
        }
        else if(index.column() == 5){
            return crossSection.getDescription(index.row());
        }
        else {
            return QVariant();
        }
    }
    return QVariant();
}

QVariant ProfileModel::dataHydraulicParameters(const QModelIndex &index, const int role) const
{
    if(role ==Qt::DisplayRole){
        const CrossSection& crossSection = *mProfiles->at(static_cast<unsigned int>(index.parent().row()));
        switch(index.column()){
        case Description:
            return crossSection.getStageDescription(index.row());
        case Stage:
            return crossSection.getStage(index.row());
        case DepthMax:
            return crossSection.getMaxDepth(index.row());
        case Area:
            return crossSection.getArea(index.row());
        case WettedPerimeter:
            return crossSection.getWettedPerimeter(index.row());
        case Width:
            return crossSection.getWidth(index.row());
        case HydraulicRadius:
            return crossSection.getHydraulicRadius(index.row());
        case DepthMean:
            return crossSection.getMeanDepth(index.row());
        case ManningsN:
            return crossSection.getManningN(index.row());
        case RelativeRoughness:
            return crossSection.getRelativeRoughness(index.row());
        case VelocityMean:
            return crossSection.getMeanVelocity(index.row());
        case FrictionFactor:
            return crossSection.getFrictionFactor(index.row());
        case ShearVelocity:
            return crossSection.getShearVelocity(index.row());
        case Discharge:
            return crossSection.getDischarge(index.row());
        case ShearStress:
            return crossSection.getShearStress(index.row());
        case StreamPower:
            return crossSection.getStreamPower(index.row());
        case FroudeNumber:
            return crossSection.getFroudeNumber(index.row());

        default:
            return QVariant();
        }
    }
    if(role ==Qt::BackgroundRole){
        if(index.column() == ManningsN){
            if(mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getManningN(index.row())>=1){
                return QBrush(QColor(Qt::red),Qt::SolidPattern);
            }
        }
    }
    return QVariant();
}

QVariant ProfileModel::dataMeasurementParameters(const QModelIndex &index, const int role) const
{
    if(role == Qt::DisplayRole){
        switch(index.column()){
        case 0:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getMeasurementValue(index.row());
        case 1:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointOneX(index.row());
        case 2:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointOneY(index.row());
        case 3:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointTwoX(index.row());
        case 4:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointTwoY(index.row());
        case 5: {
            ProfileMeasurementTypes text;
            return text.getTypeName(static_cast<int>(mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointType(index.row())));
        }
        default:
            return QVariant();
        }
    }
    if(role == Qt::EditRole){
        switch(index.column()){
        case 0:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getMeasurementValue(index.row());
        case 1:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointOneX(index.row());
        case 2:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointOneY(index.row());
        case 3:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointTwoX(index.row());
        case 4:
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointTwoY(index.row());
        case 5: {
            //In the edit role the number should be returned.
            return mProfiles->at(static_cast<unsigned int>(index.parent().row()))->getPointType(index.row());
        }
        default:
            return QVariant();
        }
    }
    return QVariant();
}

bool ProfileModel::removeRowsPoints(int row, int count, const QModelIndex &parent)
{
    //Updates all the points in the profile to maintain the sort order.
    emit beginRemoveRows(parent, row, row+count - 1);
    mProfiles->at(static_cast<unsigned int>(parent.row()))->removePoints(row, count);
    //mCrossSections->at(static_cast<unsigned int>(parent.row()));
    //this automatically sets up the pointer so the data is maintained.
    if(mType==ProfileType::crossSection){
        mDb.crossSectionDao.updateCrossSectionPoints(mProfiles->at(static_cast<unsigned int>(parent.row())));
    }
    else if(mType == ProfileType::longProfile){
        mDb.longProfileDao.updateProfilePoints(mProfiles->at(static_cast<unsigned int>(parent.row())));
    }
    emit endRemoveRows();
    return true;
}

bool ProfileModel::removeRowsHydraulicParameters(int row, int count, const QModelIndex &parent)
{
    if(mType == ProfileType::crossSection){
        emit beginRemoveRows(parent, row, row+count - 1);
        bool result = mProfiles->at(static_cast<unsigned int>(parent.row()))->removeHydraulicParameters(row, count);
        for(int i = row; i < row+count; i++){
            mDb.crossSectionDao.removeCrossSectionHydraulicParameters(i);
        }
        endRemoveRows();
        return result;
    }
    //If it is not a cross section hydraulic parameters do not make sense return false.
    return false;
}

bool ProfileModel::removeRowsMeasurements(int row, int count, const QModelIndex &parent)
{
    emit beginRemoveRows(parent, row, row+count - 1);
    bool result = mDb.crossSectionDao.removeMeasurements(mProfiles->at(static_cast<unsigned int>(parent.row())),row, count);
    result = result && mProfiles->at(static_cast<unsigned int>(parent.row()))->removeMeasurements(row, count);
    endRemoveRows();
    return result;
}

void ProfileModel::hydraulicParametersUpdated(const QModelIndex &parent)
{
    //this sends notification of the hydraulic parameter table update when the
    // dependent parameters are adjusted.
    QModelIndex parentReference = this->createIndex(parent.row(),2);
    qint32 lastRow = rowCount(parentReference);
    qint32 lastColumn = columnCount(parentReference);
    emit dataChanged(this->createIndex(0,0,&parentReference),createIndex(lastRow,lastColumn, &parentReference));
}

bool ProfileModel::setDataCrossSectionPoints(const QModelIndex &index, const QVariant &value, int role)
{
    //assume the parent checks have already been performed to get here
    //qDebug() << "CrossSectionModel::setDataCrossSectionPoints";
    if(role != Qt::EditRole){
        return false;
    }
    bool returnValue;
    double setValue = value.toString().toDouble(&returnValue);
    if(!returnValue){
        return returnValue;
    }
    switch(index.column()){
    case 0: {
        //Assuming that all values passed can convert to string.
        //Northing
        mProfiles->at(static_cast<unsigned int>(index.parent().row()))->setNorthing(index.row(), setValue);
        //this will exit so we won't ever get to the bottom.
        return true;
    }
    case 1: {
        //Easting
        mProfiles->at(static_cast<unsigned int>(index.parent().row()))->setEasting(index.row(), setValue);
        break;
    }
    case 2: {
        //Elevation
        mProfiles->at(static_cast<unsigned int>(index.parent().row()))->setElevation(index.row(), setValue);
        break;
    }
    case 3: {
        //Station. Only allow setting of first station.
        if(index.row() == 0){
            mProfiles->at(static_cast<unsigned int>(index.parent().row()))->setStationing(index.row(), setValue);
            break;
        }
        return false;
    }
    default: {
        //all the un-editable parameters return false.
        return false;
    }
    }
    if(returnValue){
        //write changes
        if(mType == ProfileType::crossSection){
            mDb.crossSectionDao.updateProfilePoint(mProfiles->at(static_cast<unsigned int>(index.parent().row())),index.row());
            hydraulicParametersUpdated(index.parent());
        }
        else if(mType == ProfileType::longProfile){
            mDb.longProfileDao.updateProfilePoint(mProfiles->at(static_cast<unsigned int>(index.parent().row())),index.row());
        }
    }
    return returnValue;
}

bool ProfileModel::setDataHydraulicParameters(const QModelIndex &index, const QVariant &value, int role)
{
    //assume the parent checks have already been performed to get here
    if(role != Qt::EditRole){
        return false;
    }
    bool returnValue;
    switch(index.column()){
    case ProfileModel::Description: {
        //Updated the description does not change any of the other data in the table so this case may return true.
        //Assuming that all values passed can convert to string.
        mProfiles->at(index.parent().row())->setStageDescription(index.row(), value.toString());
        //this will exit so we won't ever get to the bottom.
        mDb.crossSectionDao.updateCrossSectionHydraulicParameters(mProfiles->at(static_cast<unsigned int>(index.parent().row())));
        emit dataChanged(index,index);
        return true;
    }
        //Updating stage or ManningsN will result in a chnage to the rest of the table so need to emit data
        //chnaged for everything but the description.
    case ProfileModel::Stage: {
        double setValue = value.toString().toDouble(&returnValue);
        if(returnValue){
            mProfiles->at(index.parent().row())->setStage(index.row(), setValue);
        }
        break;
    }
    case ProfileModel::ManningsN: {
        double setValue = value.toString().toDouble(&returnValue);
        if(returnValue){
            mProfiles->at(index.parent().row())->setMannings(index.row(), setValue);
        }
        break;
    }
    default:
        //all the un-editable parameters return false.
        return false;
    }
    if(returnValue){
        QModelIndex parentReference = this->index(index.parent().row(),1);
        int row = index.row();
        //subtract 1 for 0 based counting
        int lastColumn = columnCount(parentReference)-1;
        //this will emit the data change for all columns to cause a recalculate.
        //write data changes.
        mDb.crossSectionDao.updateCrossSectionHydraulicParameters(mProfiles->at(static_cast<unsigned int>(index.parent().row())));
        QModelIndex firstIndex = this->index(row,0,parentReference);
        QModelIndex lastIndex = this->index(row,lastColumn,parentReference);
        emit dataChanged(firstIndex,lastIndex);

    }
    return returnValue;
}

bool ProfileModel::setDataMeasurement(const QModelIndex &index, const QVariant &value, int role)
{
    if(role == Qt::EditRole){
        switch(index.column()){
        case 0:
        {
            //Value is calculated
            return false;
        }
        case 1:
        {
            mProfiles->at(index.parent().row())->setPointOneX(value.toDouble(), index.row());
            if(mType == ProfileType::crossSection){
                mDb.crossSectionDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            else if(mType == ProfileType::longProfile){
                 mDb.longProfileDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            //value changed and the point changed
            emit dataChanged(index,index);
            emit dataChanged(this->index(index.row(),0,index.parent()),this->index(index.row(),0,index.parent()));
            break;
        }
        case 2:
        {
            mProfiles->at(index.parent().row())->setPointOneY(value.toDouble(),index.row());
            if(mType == ProfileType::crossSection){
                mDb.crossSectionDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            else if(mType == ProfileType::longProfile){
                 mDb.longProfileDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            emit dataChanged(index,index);
            emit dataChanged(this->index(index.row(),0,index.parent()),this->index(index.row(),0,index.parent()));
            break;
        }
        case 3:
        {
            mProfiles->at(index.parent().row())->setPointTwoX(value.toDouble(), index.row());
            if(mType == ProfileType::crossSection){
                mDb.crossSectionDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            else if(mType == ProfileType::longProfile){
                 mDb.longProfileDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            emit dataChanged(index,index);
            emit dataChanged(this->index(index.row(),0,index.parent()),this->index(index.row(),0,index.parent()));
            break;
        }
        case 4:
        {
            mProfiles->at(index.parent().row())->setPointTwoY(value.toDouble(), index.row());
            if(mType == ProfileType::crossSection){
                mDb.crossSectionDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            else if(mType == ProfileType::longProfile){
                 mDb.longProfileDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            emit dataChanged(index,index);
            emit dataChanged(this->index(index.row(),0,index.parent()),this->index(index.row(),0,index.parent()));
            break;
        }
        case 5:{
            mProfiles->at(index.parent().row())->setPointType(value.toInt(), index.row());
            if(mType == ProfileType::crossSection){
                mDb.crossSectionDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            else if(mType == ProfileType::longProfile){
                 mDb.longProfileDao.updateMeasurement(mProfiles->at(index.parent().row()),index.row());
            }
            emit dataChanged(index,index);
            emit dataChanged(this->index(index.row(),0,index.parent()),this->index(index.row(),0,index.parent()));
            break;
        }
        default:
            break;
        }
    }
    return false;
}

bool ProfileModel::readCSVCrossSectionStationOffset(QUrl fileUrl, double alignmentStationing, QString description)
{
    QString fileName = fileUrl.toLocalFile();
    QString delimiter =",";
    QFile file(fileName);
    if( !file.open(QFile::ReadOnly |QFile::Text)){
        qDebug() << "File doesn't exist";
        file.close();
        return false;
    }
    else {
        std::unique_ptr<CrossSection> newCrossSection(new CrossSection);
        QTextStream in(&file);
        //first check to make sure there aren't any problems
        newCrossSection->setName(description);
        while(!in.atEnd()){
            QString line = in.readLine();
            QStringList items = line.split(delimiter);
            std::unique_ptr<ProjectedPoint> newPoint;
            newPoint->setNorthing(items.at(0).toDouble());
            newPoint->setStation(items.at(0).toDouble());
            newPoint->setEasting(alignmentStationing);
            newPoint->setElevation(items.at(1).toDouble());
            newCrossSection->addPoint(move(newPoint));
        }
        file.close();
        addCrossSection(move(newCrossSection));
        return true;

    }
}

bool ProfileModel::readCSVCrossSectionNEZ(QUrl fileUrl, QString description)
{
    QString fileName = fileUrl.toLocalFile();
    QString delimiter =",";
    QFile file(fileName);
    if( !file.open(QFile::ReadOnly |QFile::Text)){
        qDebug() << "File doesn't exist";
        file.close();
        return false;
    }
    else {
        std::unique_ptr<CrossSection> newCrossSection(new CrossSection);
        QTextStream in(&file);
        //first check to make sure there aren't any problems
        newCrossSection->setName(description);
        while(!in.atEnd()){
            QString line = in.readLine();
            QStringList items = line.split(delimiter);
            std::unique_ptr<ProjectedPoint> newPoint;
            newPoint->setNorthing(items.at(0).toDouble());
            newPoint->setEasting(items.at(1).toDouble());
            newPoint->setElevation(items.at(2).toDouble());
            newCrossSection->addPoint(move(newPoint));
        }
        file.close();
        addCrossSection(move(newCrossSection));
        return true;

    }
}
