//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#ifndef CROSSSECTIONMODEL_H
#define CROSSSECTIONMODEL_H

#include <QAbstractItemModel>
#include <QHash>
#include "projectedpoint.h"
#include "gatdatamanager.h"
#include <memory>
#include <vector>
#include <QUrl>
#include "gat-core_global.h"

class GATCORESHARED_EXPORT ProfileModel : public QAbstractItemModel
{
    Q_OBJECT
public:
    enum Roles{
        IdRole = Qt::UserRole + 1,
        DescriptionRole,
        PointRole
    };

    enum HydraulicParameterNames{
        Description = 0,
        Stage,
        DepthMax,
        Area,
        WettedPerimeter,
        Width,
        HydraulicRadius,
        DepthMean,
        ManningsN,
        RelativeRoughness,
        VelocityMean,
        FrictionFactor,
        ShearVelocity,
        Discharge,
        ShearStress,
        StreamPower,
        FroudeNumber
    };

    enum ProfileType{
        crossSection = 0,
        longProfile
    };


    //CrossSectionModel(QObject *parent = nullptr);
    ProfileModel(ProfileType type = ProfileType::crossSection, QObject *parent = nullptr);
    int rowCount(const QModelIndex&parent = QModelIndex())const override;
    int columnCount(const QModelIndex& parent = QModelIndex())const override;
    QVariant data(const QModelIndex& index, const int role = Qt::DisplayRole) const override;
    QVariant headerData(int section, const Qt::Orientation orientation, int role = Qt::DisplayRole) const override;
    bool setData(const QModelIndex& index, const QVariant& value, int role = Qt::EditRole) override;
    bool removeRows(int row, int count, const QModelIndex& parent) override;
    Qt::ItemFlags flags(const QModelIndex & index) const override;
    QModelIndex index(int row, int column, const QModelIndex & parent = QModelIndex()) const override;
    void addCrossSection(std::unique_ptr <CrossSection> xs);
    void addProfile(std::unique_ptr<Profile> profile);
    bool hasChildren(const QModelIndex& index) const override;
    QModelIndex parent(const QModelIndex& index) const override;
    bool isIndexValid(const QModelIndex & index) const;
    bool readCSVCrossSectionStationOffset(QUrl fileUrl, double alignmentSetationing = 0, QString description = "");
    bool readCSVCrossSectionNEZ(QUrl fileUrl, QString description = "");

public slots:
    void databaseUpdated();
    void calculateHydraulicParameters(const QModelIndex & parent, double beginStage, double endStage, const double increment);
    void calculateHydraulicParameter(const QModelIndex & parent, double stage);
    void addMeasurement(const QModelIndex & parent, const QPointF &pointOne, const QPointF &pointTwo);

private:
    QVariant dataProfilePoints(const QModelIndex &index, const int role) const;
    QVariant dataHydraulicParameters(const QModelIndex &index, const int role) const;
    QVariant dataMeasurementParameters(const QModelIndex &index, const int role) const;

    GATDataManager & mDb;
    std::unique_ptr<std::vector<std::unique_ptr<CrossSection>>> mProfiles;
    bool removeRowsPoints(int row, int count, const QModelIndex &parent);
    bool removeRowsHydraulicParameters(int row, int count, const QModelIndex &parent);
    bool removeRowsMeasurements(int row, int count, const QModelIndex &parent);
    void hydraulicParametersUpdated(const QModelIndex &parent);
    bool setDataCrossSectionPoints(const QModelIndex& index, const QVariant& value, int role = Qt::EditRole);
    bool setDataHydraulicParameters(const QModelIndex& index, const QVariant& value, int role = Qt::EditRole);
    bool setDataMeasurement(const QModelIndex &index, const QVariant& value, int role = Qt::EditRole);
    ProfileType mType;
};

#endif // CROSSSECTIONMODEL_H
