#include "alignmentlayer.h"
#include <QGraphicsScene>
#include <QDebug>
#include <iterator>

AlignmentLayer::AlignmentLayer(QGraphicsItem* parent): QGraphicsItemGroup(parent)
  , QObject(), mIndexPointHash(), mIndexLineHash()
{
    mMode = GATPlanViewScene::noMode;
    setHandlesChildEvents(false);
}

void AlignmentLayer::setModel(AlignmentModel *alignmentModel)
{
    mAlignmentModel = alignmentModel;
    QObject::connect(mAlignmentModel,&AlignmentModel::modelReset, this, &AlignmentLayer::modelReset);
    QObject::connect(mAlignmentModel, &AlignmentModel::rowsAboutToBeRemoved, this, &AlignmentLayer::rowsAboutToBeRemoved);
    QObject::connect(mAlignmentModel, &AlignmentModel::modelAboutToBeReset, this, &AlignmentLayer::modelAboutToBeReset);
    QObject::connect(mAlignmentModel, &AlignmentModel::dataChanged, this, &AlignmentLayer::dataChanged);
    QObject::connect(mAlignmentModel, &AlignmentModel::rowsInserted, this, &AlignmentLayer::rowsInserted);

    loadAlignments();
}

void AlignmentLayer::loadAlignments(){
    loadAlignmentPoints();
    loadAlignmentLines();
    calculateBoundingRect();
}

void AlignmentLayer::updatePoint(AlignmentGraphicsItem *point, qreal northing, qreal easting)
{
    QPersistentModelIndex pIdx = mIndexPointHash.value(point);
    mAlignmentModel->setData(pIdx, northing, Qt::EditRole);
    mAlignmentModel->setData(mAlignmentModel->index(pIdx.row(),1, pIdx.parent()),easting, Qt::EditRole);

}

void AlignmentLayer::deletePoint(AlignmentGraphicsItem *point)
{
    QPersistentModelIndex pIdx = mIndexPointHash.value(point);
    mAlignmentModel->removeRow(pIdx.row(),pIdx.parent());
}

void AlignmentLayer::insertPoint(AlignmentLineGraphicsItem * line, double northing, double easting)
{
    QPersistentModelIndex pIdx = mIndexLineHash.value(line);
    int row = pIdx.row()+1;
    QModelIndex parent = pIdx.parent();
    qDebug() << "AlignmentLayer::insertPoint " << row << " " << parent;
    if(mAlignmentModel->insertRow(row, parent)){
        mAlignmentModel->setData(mAlignmentModel->index(row, 0,parent),northing, Qt::EditRole);
        mAlignmentModel->setData(mAlignmentModel->index(row, 1,parent),easting, Qt::EditRole);
    }
}

GATPlanViewScene::GATSceneMode AlignmentLayer::getCurrentMode()
{
    return mMode;
}

void AlignmentLayer::setCurrentMode(GATPlanViewScene::GATSceneMode mode)
{
    mMode = mode;
}

void AlignmentLayer::clearAlignments()
{
    if(mIndexPointHash.size() >0){
        QHash<AlignmentGraphicsItem*, QPersistentModelIndex>::const_iterator i;
        for(i = mIndexPointHash.begin(); i!= mIndexPointHash.end();++i){
            AlignmentGraphicsItem *point = i.key();
            removePoint(point);
            delete point;
        }
        mIndexPointHash.clear();
    }
    if(mIndexLineHash.size() >0){

        QHash<AlignmentLineGraphicsItem*, QPersistentModelIndex>::const_iterator j;
        for(j = mIndexLineHash.begin(); j!= mIndexLineHash.end(); ++j){
            AlignmentLineGraphicsItem *line = j.key();
            removeLine(line);
            delete line;
        }
        mIndexLineHash.clear();
    }
}

void AlignmentLayer::calculateBoundingRect()
{
    //qDebug() << "GATPlanViewScene::itemsBoundingRect()";
    //This only includes the points because the lines are all within the bounds of the points.
    if(mIndexLineHash.isEmpty())
    {
        mBoundingRect = QRectF();
        return;
    }
    if(mIndexLineHash.size()==0){
        mBoundingRect = QRectF();
        return;
    }

    QHash<AlignmentLineGraphicsItem*, QPersistentModelIndex>::const_iterator i;
    //Grab the first item.
    mBoundingRect = mIndexLineHash.begin().key()->boundingRect();
    for(i =mIndexLineHash.begin()+1; i != mIndexLineHash.end(); ++i){
        AlignmentLineGraphicsItem* item = i.key();
            //qDebug() << "AlignmentLayer::boundingRect " << item->boundingRect();
            mBoundingRect = mBoundingRect.united(item->boundingRect());
    }
}

QRectF AlignmentLayer::boundingRect() const
{
    //qDebug() << "AlignmentLayer::boundingRect " << mBoundingRect;
    return mBoundingRect;
}



void AlignmentLayer::modelReset()
{
    //clearAlignments();
    loadAlignments();
}

void AlignmentLayer::modelAboutToBeReset()
{
    clearAlignments();
}

void AlignmentLayer::dataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight, const QVector<int> &roles)
{
//    if(roles.at(0)!=Qt::DisplayRole){
//        //only care if the DisplayRole changed
//        return;
//    }
    if(!mAlignmentModel->isIndexValid(topLeft.parent())){
        //If parent is where the change occured this will not affect the points
        return;
    }
    if(topLeft.parent().column()!=0){
        //Only changes to the points make a difference other tables no change.
        return;
    }
    if(topLeft.column() == 2){
        //The position of the point did not change just the stationing
        //this really doesn't change much.
        return;
    }
    for(int i = topLeft.row(); i <= bottomRight.row(); i++){
        QPersistentModelIndex pIdx(mAlignmentModel->index(i,0,topLeft.parent()));
        AlignmentGraphicsItem* point = mIndexPointHash.key(pIdx);
        double easting = mAlignmentModel->data(mAlignmentModel->index(i,1,topLeft.parent())).toDouble();
        double northing = mAlignmentModel->data(mAlignmentModel->index(i,0,topLeft.parent())).toDouble();


        point->setPosition(easting,northing);
        if(i >0){
            QPersistentModelIndex pIdxPrev(mAlignmentModel->index(i-1,0,topLeft.parent()));
            AlignmentLineGraphicsItem* line = mIndexLineHash.key(pIdxPrev);
            AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(mAlignmentModel->index(i-1,3,topLeft.parent())).toInt());
            double northingCenter = mAlignmentModel->data(mAlignmentModel->index(i-1,4,topLeft.parent())).toDouble();
            double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(i-1,5,topLeft.parent())).toDouble();

            qreal easting0 = mAlignmentModel->data(mAlignmentModel->index(i-1,1,topLeft.parent())).toDouble();
            qreal northing0 = mAlignmentModel->data(mAlignmentModel->index(i-1,0,topLeft.parent())).toDouble();
            line->setLine(easting0, northing0, easting, northing);
            line->setPointType(pointType);
            line->setCenter(eastingCenter, northingCenter);
        }
        if(i < (mAlignmentModel->rowCount(topLeft.parent())-1)){
            AlignmentLineGraphicsItem* line = mIndexLineHash.key(pIdx);
            double easting2 = mAlignmentModel->data(mAlignmentModel->index(i+1,1,topLeft.parent())).toDouble();
            double northing2 = mAlignmentModel->data(mAlignmentModel->index(i+1,0,topLeft.parent())).toDouble();
            line->setLine(easting, northing, easting2, northing2);
            AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(mAlignmentModel->index(i,3,topLeft.parent())).toInt());
            double northingCenter = mAlignmentModel->data(mAlignmentModel->index(i,4,topLeft.parent())).toDouble();
            double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(i,5,topLeft.parent())).toDouble();
            line->setPointType(pointType);
            line->setCenter(eastingCenter, northingCenter);
        }
    }
    this->calculateBoundingRect();
}

void AlignmentLayer::rowsInserted(const QModelIndex &parent, int first, int last)
{
    if(!mAlignmentModel->isIndexValid(parent)){
        //Parent is not valid so it is an alignment that was added
        //Need to add all of the points
        for(int i = first; i <= last; i++){
            QModelIndex parent = mAlignmentModel->index(i, 0);
            for(int j = 0; j < mAlignmentModel->rowCount(parent);j++){
                AlignmentGraphicsItem* newPoint = new AlignmentGraphicsItem();
                QModelIndex eastingIndex = mAlignmentModel->index(j,1 , parent);
                QModelIndex northingIndex = mAlignmentModel->index(j,0,parent);
                double easting = mAlignmentModel->data(eastingIndex).toDouble();
                double northing= mAlignmentModel->data(northingIndex).toDouble();
                newPoint->setPosition(easting ,  northing);
                newPoint->setAlignmentLayer(this);
                addToGroup(newPoint);
                QPersistentModelIndex pIdx(northingIndex);
                mIndexPointHash.insert(newPoint,pIdx);
                if(j < mAlignmentModel->rowCount(parent)-1){
                    //Add line
                    AlignmentLineGraphicsItem* line = new AlignmentLineGraphicsItem();
                    QModelIndex eastingIndex1 = mAlignmentModel->index(j+1,1 , parent);
                    QModelIndex northingIndex1 = mAlignmentModel->index(j+1,0,parent);

                    AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(mAlignmentModel->index(j,3,parent)).toInt());
                    double northingCenter = mAlignmentModel->data(mAlignmentModel->index(j,4,parent)).toDouble();
                    double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(j,5,parent)).toDouble();
                    double easting1 = mAlignmentModel->data(eastingIndex1).toDouble();
                    double northing1= mAlignmentModel->data(northingIndex1).toDouble();
                    line->setLine(easting, northing, easting1, northing1);
                    line->setAlignmentLayer(this);
                    line->setPointType(pointType);
                    line->setCenter(eastingCenter, northingCenter);
                    addToGroup(line);
                    mIndexLineHash.insert(line, pIdx);
                }
            }
        }
    }
    if(parent.column()!=0){
        return;
    }
    for(int i = first; i <=  last; i++){
        AlignmentGraphicsItem* newPoint = new AlignmentGraphicsItem();
        QModelIndex eastingIndex = mAlignmentModel->index(i, 1, parent);
        QModelIndex northingIndex = mAlignmentModel->index(i, 0, parent);
        double easting = mAlignmentModel->data(eastingIndex).toDouble();
        double northing= mAlignmentModel->data(northingIndex).toDouble();
        newPoint->setPosition(easting,northing);
        newPoint->setAlignmentLayer(this);
        addToGroup(newPoint);
        QPersistentModelIndex pIdx(northingIndex);
        mIndexPointHash.insert(newPoint,pIdx);
        if(i < (mAlignmentModel->rowCount(parent)-1)){
            //Add line
            AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(mAlignmentModel->index(i,3,parent)).toInt());
            double northingCenter = mAlignmentModel->data(mAlignmentModel->index(i,4,parent)).toDouble();
            double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(i,5,parent)).toDouble();

            AlignmentLineGraphicsItem* line = new AlignmentLineGraphicsItem();
            QModelIndex eastingIndex1 = mAlignmentModel->index(i+1, 1 , parent);
            QModelIndex northingIndex1 = mAlignmentModel->index(i+1,0,parent);
            double easting1 = mAlignmentModel->data(eastingIndex1).toDouble();
            double northing1= mAlignmentModel->data(northingIndex1).toDouble();
            line->setLine(easting, northing, easting1, northing1);
            line->setAlignmentLayer(this);
            line->setPointType(pointType);
            line->setCenter(eastingCenter, northingCenter);
            addToGroup(line);
            mIndexLineHash.insert(line, pIdx);
        }
    }
    if(mAlignmentModel->rowCount(parent)== 1){
        //Below is adding lines if there is only one point then no line to be added
        return;
    }
    if(first == mAlignmentModel->rowCount(parent)-1){
        //Need to add the last line
        QModelIndex eastingIndex = mAlignmentModel->index(first-1,1 , parent);
        QModelIndex northingIndex = mAlignmentModel->index(first-1,0,parent);
        double easting = mAlignmentModel->data(eastingIndex).toDouble();
        double northing= mAlignmentModel->data(northingIndex).toDouble();
        AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(mAlignmentModel->index(first-1,3,parent)).toInt());
        double northingCenter = mAlignmentModel->data(mAlignmentModel->index(first-1,4,parent)).toDouble();
        double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(first-1,5,parent)).toDouble();

        AlignmentLineGraphicsItem* line = new AlignmentLineGraphicsItem();
        QModelIndex eastingIndex1 = mAlignmentModel->index(first, 1, parent);
        QModelIndex northingIndex1 = mAlignmentModel->index(first, 0, parent);
        double easting1 = mAlignmentModel->data(eastingIndex1).toDouble();
        double northing1= mAlignmentModel->data(northingIndex1).toDouble();
        line->setLine(easting, northing, easting1, northing1);
        line->setAlignmentLayer(this);
        line->setPointType(pointType);
        line->setCenter(eastingCenter, northingCenter);
        addToGroup(line);
        QPersistentModelIndex pIdx(northingIndex);
        mIndexLineHash.insert(line, pIdx);
    }
    else if(first > 0){
        QModelIndex eastingIndex1 = mAlignmentModel->index(first-1,1, parent);
        QModelIndex northingIndex1 = mAlignmentModel->index(first-1,0,parent);
        QModelIndex eastingIndex2 = mAlignmentModel->index(first,1 , parent);
        QModelIndex northingIndex2 = mAlignmentModel->index(first,0,parent);

        double easting1 = mAlignmentModel->data(eastingIndex1).toDouble();
        double northing1= mAlignmentModel->data(northingIndex1).toDouble();
        AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(mAlignmentModel->index(first-1,3,parent)).toInt());
        double northingCenter = mAlignmentModel->data(mAlignmentModel->index(first-1,4,parent)).toDouble();
        double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(first-1,5,parent)).toDouble();

        double easting2 = mAlignmentModel->data(eastingIndex2).toDouble();
        double northing2= mAlignmentModel->data(northingIndex2).toDouble();

        QPersistentModelIndex pIdx(northingIndex1);
        AlignmentLineGraphicsItem* line = mIndexLineHash.key(pIdx);
        line->setLine(easting1, northing1, easting2, northing2);
        line->setPointType(pointType);
        line->setCenter(eastingCenter, northingCenter);
    }
    calculateBoundingRect();
}

void AlignmentLayer::rowsAboutToBeRemoved(const QModelIndex &parent, int first, int last)
{
    qDebug() << "AlignmentLayer::rowsAboutToBeRemoved";
    //Remove rows in preperation for the model to be deleted.
    if(!mAlignmentModel->isIndexValid(parent)){
        //Alignment being removed
        for(int i = first; i <= last; i++){
            QModelIndex alignmentIndex = mAlignmentModel->index(i,0);
            for(int j = 0; j < mAlignmentModel->rowCount(alignmentIndex); j++){
                QPersistentModelIndex pIdx(mAlignmentModel->index(j,0,alignmentIndex));
                //Remove point
                AlignmentGraphicsItem* point = mIndexPointHash.key(pIdx);
                removePoint(point);
                mIndexPointHash.remove(point);
                delete point;
            }
            for(int k = 0; k < (mAlignmentModel->rowCount(alignmentIndex)-1); k++){
                QPersistentModelIndex pIdx = QPersistentModelIndex(mAlignmentModel->index(k,0,alignmentIndex));
                AlignmentLineGraphicsItem* line = mIndexLineHash.key(pIdx);
                removeLine(line);
                mIndexLineHash.remove(line);
                delete line;
            }
        }
        calculateBoundingRect();
        return;
    }
    if(parent.column()!= 0){
        //Other tables removing data don't need to do anything
        return;
    }
    qDebug() << "AlignmentLayer::rowsAboutToBeRemoved first and last " << first << ", " << last;

    for(int i = first; i <= last; i++){
        QPersistentModelIndex pIdx = QPersistentModelIndex(mAlignmentModel->index(i,0,parent));
        AlignmentGraphicsItem* point = mIndexPointHash.key(pIdx);
        removePoint(point);
        mIndexPointHash.remove(point);
        delete point;
        //There is one fewer lines than rows
        if(i < (mAlignmentModel->rowCount(parent)-1)){
            AlignmentLineGraphicsItem* line = mIndexLineHash.key(pIdx);
            removeLine(line);
            mIndexLineHash.remove(line);
            delete line;
        }
    }
    //Adjust end lines if applicable
    if(first > 0 ){
        if(last == (mAlignmentModel->rowCount(parent)-1)){
            qDebug() << "AlignmentLayer::rowsAboutToBeRemoved last point deleted";
            //If there is a point before the first and the last point is deleted
            //Need to remove one additonal line
            QPersistentModelIndex pIdx = QPersistentModelIndex(mAlignmentModel->index(first-1,0,parent));
            AlignmentLineGraphicsItem* line = mIndexLineHash.key(pIdx);
            removeLine(line);
            mIndexLineHash.remove(line);
            delete line;
        }
        if(last < (mAlignmentModel->rowCount(parent)-1)){
            qDebug() << "AlignmentLayer::rowsAboutToBeRemoved last point remaining";
            //Need to adjust the prior line end point if the point was in the middle
            QPersistentModelIndex pIdx(mAlignmentModel->index(first-1, 0, parent));
            AlignmentLineGraphicsItem* line = mIndexLineHash.key(pIdx);
            double northing1 = mAlignmentModel->data(mAlignmentModel->index(first-1,0,parent)).toDouble();
            double easting1 = mAlignmentModel->data(mAlignmentModel->index(first-1,1,parent)).toDouble();
            AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(mAlignmentModel->index(first-1,3,parent)).toInt());
            double northingCenter = mAlignmentModel->data(mAlignmentModel->index(first-1,4,parent)).toDouble();
            double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(first-1,5,parent)).toDouble();

            double northing2 = mAlignmentModel->data(mAlignmentModel->index(last+1,0,parent)).toDouble();
            double easting2 = mAlignmentModel->data(mAlignmentModel->index(last+1,1,parent)).toDouble();
            line->setLine(easting1,northing1, easting2, northing2);
            line->setPointType(pointType);
            line->setCenter(eastingCenter, northingCenter);
        }

    }
    calculateBoundingRect();
    prepareGeometryChange();
}

void AlignmentLayer::removePoint(AlignmentGraphicsItem *point)
{
    this->removeFromGroup(point);
    //point->scene()->removeItem(point);
}

void AlignmentLayer::removeLine(AlignmentLineGraphicsItem *line)
{
    this->removeFromGroup(line);
    //line->scene()->removeItem(line);
}

void AlignmentLayer::loadAlignmentPoints()
{
    qDebug() << "AlignmentLayer::loadAlignmentPoints";
    //TODO use the model to build the graphics items.
    for(int i = 0; i < mAlignmentModel->rowCount(); i++){
        QModelIndex parent = mAlignmentModel->index(i, 0);
        for(int j = 0; j < mAlignmentModel->rowCount(parent);j++){
            AlignmentGraphicsItem* newPoint = new AlignmentGraphicsItem();
            QModelIndex eastingIndex = mAlignmentModel->index(j,1,parent);
            QModelIndex northingIndex = mAlignmentModel->index(j,0,parent);
            newPoint->setPosition(mAlignmentModel->data(eastingIndex).toDouble(),
                                  mAlignmentModel->data(northingIndex).toDouble());

            newPoint->setAlignmentLayer(this);
            addToGroup(newPoint);

            QPersistentModelIndex pIdx(northingIndex);
            mIndexPointHash.insert(newPoint,pIdx);
        }
    }
}

void AlignmentLayer::loadAlignmentLines()
{
    for(int i = 0; i < mAlignmentModel->rowCount(); i++){
        QModelIndex parent = mAlignmentModel->index(i, 0);
        if(mAlignmentModel->rowCount(parent)<=1){
            continue;
        }
        else {
            //Know there are at least 2 points
            //Northing is the persistentModelIndex so it is stored and used again.
            QModelIndex firstNorthingIndex = mAlignmentModel->index(0,0,parent);
            double firstEasting = mAlignmentModel->data(mAlignmentModel->index(0,1,parent)).toDouble();
            double firstNorthing = mAlignmentModel->data(firstNorthingIndex).toDouble();
            for(int j = 1; j < mAlignmentModel->rowCount(parent); j++){
                AlignmentLineGraphicsItem* newLine = new AlignmentLineGraphicsItem();
                QModelIndex eastingIndex = mAlignmentModel->index(j,1,parent);
                QModelIndex northingIndex = mAlignmentModel->index(j,0,parent);
                QModelIndex pointTypeIndex = mAlignmentModel->index(j-1,3,parent);
                double easting = mAlignmentModel->data(eastingIndex).toDouble();
                double northing = mAlignmentModel->data(northingIndex).toDouble();
                AlignmentPoint::PointType pointType = static_cast<AlignmentPoint::PointType>(mAlignmentModel->data(pointTypeIndex).toInt());
                double eastingCenter = mAlignmentModel->data(mAlignmentModel->index(j-1,5,parent)).toDouble();
                double northingCenter = mAlignmentModel->data(mAlignmentModel->index(j-1,4,parent)).toDouble();
                newLine->setLine(firstEasting, firstNorthing, easting, northing);
                newLine->setAlignmentLayer(this);
                newLine->setPointType(pointType);
                newLine->setCenter(eastingCenter, northingCenter);
                addToGroup(newLine);
                QPersistentModelIndex pIdx(firstNorthingIndex);
                mIndexLineHash.insert(newLine, pIdx);
                //shuffle forward for next line.
                firstEasting = easting;
                firstNorthing = northing;
                firstNorthingIndex = northingIndex;
            }
        }
    }
}

