//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "gatplanview.h"
#include <QWheelEvent>
#include <QDebug>
#include <QScrollBar>



GATPlanView::GATPlanView(QWidget* parent) : QGraphicsView(parent), totalScaleFactor(1)
{
    scale(1,-1);
    setPanMode();
    setAttribute(Qt::WA_AcceptTouchEvents,true);
    setTransformationAnchor(QGraphicsView::AnchorUnderMouse);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    //By defult the drag is handled by the touch event.    
    setInteractive(true);
    grabGesture(Qt::PinchGesture);
    //setSizeAdjustPolicy(QAbstractScrollArea::AdjustIgnored);
    //This is for the mouse press/release
    mPan = false;
}

//https://riptutorial.com/qt/example/24869/pan--zoom--and-rotate-with-qgraphicsview

void GATPlanView::wheelEvent(QWheelEvent *event)
{
    setTransformationAnchor(QGraphicsView::AnchorUnderMouse);
    if(event->delta() > 0){
        scale(1.25, 1.25);
        mScaleBar->updatePosition();
    }
    else {
        scale(0.8, 0.8);
        mScaleBar->updatePosition();
    }
    //qDebug() << "GATPlanView::wheelEvent m11 " << this->transform().m11();
}
//re-implamented from
//https://bugreports.qt.io/browse/QTBUG-7328
void GATPlanView::mousePressEvent(QMouseEvent *event)
{
    //qDebug() << "GATPlanView::mousePressEvent enter mouse press event";
    if(event->button()== Qt::MidButton || (mMode == GATPlanViewScene::panMode && event->button() == Qt::LeftButton))
    {
        setCursor(Qt::ClosedHandCursor);
        mPan = true;
        mMousePressCoordinate = mapToScene(event->pos());
        event->accept();
        return;
    }
    QGraphicsView::mousePressEvent(event);
}

void GATPlanView::mouseReleaseEvent(QMouseEvent *event)
{
    //qDebug() << "GATPlanView::mouseReleaseEvent enter";
    if((event->button()==Qt::MidButton || (mMode == GATPlanViewScene::panMode && event->button() == Qt::LeftButton))&& mPan){
        setCursor(Qt::OpenHandCursor);
        mPan = false;
        //qDebug() << "GATPlanView::mouseReleaseEvent mMousePressCoordinate, event pos " << mMousePressCoordinate << " " << mapToScene(event->pos());
        setTransformationAnchor((QGraphicsView::AnchorUnderMouse));
        QPointF centerOn = mapToScene(viewport()->rect().center()) - mapToScene(event->pos())+mMousePressCoordinate;
        //horizontal scaling factor should be the same
        this->centerOn(centerOn);

        mScaleBar->updatePosition();
        return;
    }
    QGraphicsView::mouseReleaseEvent(event);
}

//With attribute to: https://doc.qt.io/archives/qt-4.8/qt-touch-pinchzoom-graphicsview-cpp.html
bool GATPlanView::viewportEvent(QEvent *event)
{
    //qDebug() << "GATPlanView::viewportEvent eventType: " << event->type();
    switch (event->type()) {
    case QEvent::TouchBegin:
        break;
    case QEvent::TouchUpdate:
        break;
    case QEvent::TouchEnd:
    {
        QTouchEvent *touchEvent = static_cast<QTouchEvent *>(event);
        QList<QTouchEvent::TouchPoint> touchPoints = touchEvent->touchPoints();
        if (touchPoints.count() == 2) {
            // determine scale factor
            const QTouchEvent::TouchPoint &touchPoint0 = touchPoints.first();
            const QTouchEvent::TouchPoint &touchPoint1 = touchPoints.last();
            qreal currentScaleFactor =
                    QLineF(touchPoint0.pos(), touchPoint1.pos()).length()
                    / QLineF(touchPoint0.startPos(), touchPoint1.startPos()).length();
            if (touchEvent->touchPointStates() & Qt::TouchPointReleased) {
                // if one of the fingers is released, remember the current scale
                // factor so that adding another finger later will continue zooming
                // by adding new scale factor to the existing remembered value.
                totalScaleFactor *= currentScaleFactor;
                currentScaleFactor = 1;
            }

            centerOn(QLineF(mapToScene(touchPoint0.startPos().toPoint()),mapToScene(touchPoint1.startPos().toPoint())).center());
            //updated this to just apply the scale
            setTransformationAnchor(QGraphicsView::AnchorViewCenter);

//            setTransform(QTransform().scale(totalScaleFactor * currentScaleFactor,
//                                            totalScaleFactor * currentScaleFactor));
            scale(currentScaleFactor,currentScaleFactor);
            mScaleBar->updatePosition();
            event->accept();

        }
        return true;
    }
    default:
        break;
    }
    return QGraphicsView::viewportEvent(event);
}

bool GATPlanView::event(QEvent *event)
{
    if(event->type() == QEvent::Gesture){
        return processGesture(static_cast<QGestureEvent*>(event));
    }
    return QGraphicsView::event(event);
}

bool GATPlanView::processGesture(QGestureEvent *event)
{

    if(QGesture *pinch=event->gesture(Qt::PinchGesture)){
        pinchTriggered(static_cast<QPinchGesture *>(pinch));
        event->accept();
    }
    return true;
}

void GATPlanView::pinchTriggered(QPinchGesture *gesture)
{
    qreal sf = gesture->scaleFactor();
    scale(sf,sf);
//    totalScaleFactor *= sf;
//    setTransformationAnchor(QGraphicsView::AnchorViewCenter);

//    setTransform(QTransform().scale(totalScaleFactor,
//                                    totalScaleFactor));
    mScaleBar->updatePosition();
    centerOn(mapToScene(gesture->centerPoint().toPoint()));
}

void GATPlanView::zoomExtents()
{
    //setDragMode(QGraphicsView::NoDrag);
    //This resets the view otherwise it is ever expanding.
    QRectF boundingRect = mScene->itemsBoundingRect();
    //Margin on frame is set to 10 units
    qreal scale;
    if(boundingRect.width() <= 0){
        scale = 1;
    }

    else {
        scale = (this->viewport()->width())/(boundingRect.width());
        if(scale > (this->viewport()->height())/(boundingRect.height()))
        {
            scale = (this->viewport()->height())/(boundingRect.height());
        }
        //make the scale slightly smaller to allow for margins.
        scale = scale * .9;
    }

    setTransformationAnchor(QGraphicsView::NoAnchor);
    setResizeAnchor(QGraphicsView::NoAnchor);
    QTransform transform(scale,0,0,-1.0*scale,0,0);
    this->setTransform(transform);
    //used center on to work around the anchor moving the view to either under the mouse or other.
    centerOn(boundingRect.center());
    mScaleBar->updatePosition();
}

void GATPlanView::setPanMode(){
   mMode = GATPlanViewScene::panMode;
   setDragMode(QGraphicsView::ScrollHandDrag);
   setCursor(Qt::OpenHandCursor);
}

void GATPlanView::setToolMode(const GATPlanViewScene::GATSceneMode mode)
{
    mMode = mode;
    setDragMode(QGraphicsView::NoDrag);
    setCursor(Qt::CrossCursor);
}

void GATPlanView::setScaleBar(GATScaleBar *scaleBar)
{
    mScaleBar = scaleBar;
}

void GATPlanView::setScene(GATPlanViewScene *scene)
{
    mScene = scene;
    QGraphicsView::setScene(mScene);
}

void GATPlanView::zoomIn(){
    //setDragMode(QGraphicsView::NoDrag);
    setTransformationAnchor(QGraphicsView::AnchorViewCenter);
    scale(1.25, 1.25);
    mScaleBar->updatePosition();
}

void GATPlanView::zoomOut(){
    //setDragMode(QGraphicsView::NoDrag);
    setTransformationAnchor(QGraphicsView::AnchorViewCenter);
    scale(.8, .8);
    mScaleBar->updatePosition();
}

