//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "mainwindow.h"

#include <QtCore/QVariant>
#include <QtCore>
#include <QtWidgets/QAction>
#include <QtWidgets/QApplication>
#include <QtWidgets/QMainWindow>

#include <QtDebug>
#include <QWindow>
#include <QKeySequence>
#include <QUrl>
#include <QDir>
#include <QDesktopWidget>
#include <QMessageBox>

#include "pointswidget.h"
#include "pointsmodel.h"
#include "alignmentmodel.h"
#include "profilemodel.h"
#include "gatplanviewscene.h"
#include "gatplanview.h"
#include "profilechartwidget.h"
#include "hydraulicparametersmodel.h"
#include "createprofiledialog.h"
#include "calculatehydraulicparametersdialog.h"
#include "fileutilities.h"
#include "importcsvfiledialog.h"
#include "about.h"
#include "gatdataexport.h"
#include "gatscalebar.h"
#include "alignmentlayer.h"
#include "pointlayer.h"
#include "gatlandxml.h"
#include "measurementpointmodel.h"
#include "measurementitemdelegate.h"
#include "profilecalculator.h"

MainWindow::MainWindow(QWidget * parent): QMainWindow(parent), ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    //Manually setup the long profile widget so that parameters may be passed
    mLongProfileChartWidget = new ProfileChartWidget(false,this);
    ui->mMainTabWidget->addTab(mLongProfileChartWidget,"Long Profiles");
    mGATPlanViewScene = new GATPlanViewScene();


    mAlignmentLayer = new AlignmentLayer();
    mPointLayer = new PointLayer();

    setupModels();

    mGATPlanViewScene->setAlignmentPointsGroup(mAlignmentLayer);
    mGATPlanViewScene->setPointsGroup(mPointLayer);
    //Put the data into the elements
    mMobileMode = false;
    ui->pointsTable->setModel(mPointsModel);
    ui->alignmentsTable->setModel(mAlignmentModel);

    ui->alignmentsTable->setColumnHidden(2, true);

    ui->mCrossSectionChartWidget->setModel(mCrossSectionModel);

    mLongProfileChartWidget->setModel(mLongProfileModel);

    //Enable context menus on parent tables
    ui->alignmentsTable->setContextMenuPolicy(Qt::CustomContextMenu);

    mScaleBar = new GATScaleBar();
    //mGATPlanViewScene->setPointsModel(mPointsModel);
    mGATPlanViewScene->setScaleBar(mScaleBar);
    ui->mGATPlanView->setScene(mGATPlanViewScene);
    mScaleBar->setView(ui->mGATPlanView);
    ui->mGATPlanView->setScaleBar(mScaleBar);
    connectSignals();
    readSettings();
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::connectSignals()
{
    //QObject::connect(ui->mPrintItems, &QAction::triggered, mGATPlanViewScene, &GATPlanViewScene::printItems);
    QObject::connect(mDataManager, &GATDataManagerQMLPlugin::databaseUpdated, mAlignmentModel, &AlignmentModel::databaseUpdated);
    QObject::connect(mDataManager, &GATDataManagerQMLPlugin::databaseUpdated, mPointsModel, &PointsModel::databaseUpdated);
    QObject::connect(mDataManager, &GATDataManagerQMLPlugin::databaseUpdated, mCrossSectionModel, &ProfileModel::databaseUpdated);
    QObject::connect(mDataManager, &GATDataManagerQMLPlugin::databaseUpdated, mLongProfileModel, &ProfileModel::databaseUpdated);
    QObject::connect(ui->alignmentsTable, &QTableView::clicked, this, &MainWindow::setCurrentAlignment);
    QObject::connect(ui->alignmentsTable->verticalHeader(), &QHeaderView::sectionClicked, this, &MainWindow::setCurrentAlignmentRow);

    QObject::connect(ui->mActionResetWindow, &QAction::triggered, this, &MainWindow::resetTabs);
    QObject::connect(ui->mActionSingleWindow, &QAction::triggered, this, &MainWindow::mobileView);

    //Zoom GATView to the scene extents
    QObject::connect(ui->mActionZoomExtents, &QAction::triggered, ui->mGATPlanView, &GATPlanView::zoomExtents);
    QObject::connect(ui->mButtonZoomExtents, &QToolButton::clicked, ui->mGATPlanView, &GATPlanView::zoomExtents);
    QObject::connect(ui->mActionZoomIn, &QAction::triggered, ui->mGATPlanView, &GATPlanView::zoomIn);
    QObject::connect(ui->mActionZoomOut, &QAction::triggered, ui->mGATPlanView, &GATPlanView::zoomOut);

    QObject::connect(ui->mActionImportPNEZD, &QAction::triggered, this, &MainWindow::getNEZDFileImport);
    //QObject::connect(ui->mActionAddPoint, &QAction::triggered, this, &MainWindow::setAddPointMode);

    //Add Alignment Point Action. This sets the mode so that mouse clicks on the scene add points.
    QObject::connect(ui->mActionAddAlignmentPoint, &QAction::triggered, this, &MainWindow::setAddAlignmentPointMode);
    QObject::connect(ui->mButtonAddAlignmentPoint, &QToolButton::clicked, this, &MainWindow::setAddAlignmentPointMode);
    QObject::connect(ui->mButtonInsertAlignmentPoint, &QPushButton::clicked, this, &MainWindow::setInsertAlignmentPointMode);

    //Delete Alignment Point
    QObject::connect(ui->mButtonDeleteAlignmentPoint, &QPushButton::clicked, this, &MainWindow::setDeleteAlignmentPointMode);

    //Move Alignment Point
    QObject::connect(ui->mButtonMovePoint, &QPushButton::clicked, this, &MainWindow::setMoveAlignmentPointMode);

    //Adds a cross section by projecting points on alignment.
    QObject::connect(ui->mActionCreateCrossSection, &QAction::triggered, this, &MainWindow::calculateProfile);

    //Adds a profile by projecting points on alignment
    QObject::connect(ui->createLongProfileButton, &QToolButton::clicked, this, &MainWindow::calculateProfile);

    //Calculates hydraulic parameters by bringing up ui to set max and min with interval
    //QObject::connect(ui->mActionCalculateHydraulicParameters, &QAction::triggered, this, &MainWindow::calculateHydraulicParametersDialog);

    //Adds alignment to the alignments brings up ui for input of the description
    QObject::connect(ui->addAlignmentButton, &QToolButton::clicked, this, &MainWindow::addAlignment);

    //This resets the view mode to pan, the default
    QObject::connect(ui->mActionClearPlanViewTool, &QAction::triggered, this, &MainWindow::clearGraphicsViewMode);
    QObject::connect(ui->mButtonPanView, &QToolButton::clicked, this, &MainWindow::clearGraphicsViewMode);
    QObject::connect(ui->mButtonZoomIn, &QToolButton::clicked, ui->mGATPlanView, &GATPlanView::zoomIn);
    QObject::connect(ui->mButtonZoomOut, &QToolButton::clicked, ui->mGATPlanView, &GATPlanView::zoomOut);

    QObject::connect(ui->mActionOpen, &QAction::triggered, this, &MainWindow::openFile);
    QObject::connect(ui->mActionNew, &QAction::triggered, this, &MainWindow::newFile);
    QObject::connect(mGATPlanViewScene, &GATPlanViewScene::addAlignmentPoint,this, &MainWindow::addAlignmentPoint);
    QObject::connect(ui->mActionImportCSVCrossSection, &QAction::triggered, this, &MainWindow::importCrossSection);
    QObject::connect(mGATPlanViewScene, &GATPlanViewScene::createCrossSection, this, &MainWindow::calculateProfile);
    QObject::connect(ui->mActionAbout, &QAction::triggered, this, &MainWindow::showAbout);
    //mActionLoadLandXML

    QObject::connect(ui->mActionLoadLandXML, &QAction::triggered, this, &MainWindow::readLandXML);

    QObject::connect(ui->mActionSetBackgroundImage, &QAction::triggered, this, &MainWindow::setBackgroundImage);
    QObject::connect(ui->mActionExit, &QAction::triggered, this, &MainWindow::close);
    QObject::connect(ui->mActionExport, &QAction::triggered, this, &MainWindow::exportTableData);

    //Context menus
    QObject::connect(ui->alignmentsTable, &GATTableView::customContextMenuRequested, this, &MainWindow::alignmentContextMenu);
}

void MainWindow::setupModels()
{
    mPointsModel = new PointsModel(this);
    mAlignmentModel = new AlignmentModel(this);
    mCrossSectionModel = new ProfileModel(ProfileModel::crossSection,this);
    mDataManager = new GATDataManagerQMLPlugin(this);
    mCrossSectionPointsModel = new ProfilePointsModel();
    mCrossSectionPointsModel->setModel(mCrossSectionModel);
    mCrossSectionPointsModel->setRootIndex(mCrossSectionModel->index(0,0));
    mCrossSectionMeasurementsModel = new MeasurementPointModel();
    mCrossSectionMeasurementsModel->setModel(mCrossSectionModel);
    mCrossSectionMeasurementsModel->setRootIndex(mCrossSectionModel->index(0,2));

    mHydraulicParametersModel = new HydraulicParametersModel();
    mHydraulicParametersModel->setModel(mCrossSectionModel);
    mHydraulicParametersModel->setRootIndex(mCrossSectionModel->index(0,1,QModelIndex()));
    mAlignmentPointsModel = new AlignmentPointsModel();
    mAlignmentPointsModel->setModel(mAlignmentModel);
    mAlignmentPointsModel->setRootIndex(mAlignmentModel->index(0,0));
    mAlignmentMeasurementModel = new MeasurementPointModel();
    mAlignmentMeasurementModel->setModel(mAlignmentModel);
    mAlignmentMeasurementModel->setRootIndex(mAlignmentModel->index(0,1));

    mLongProfileModel = new ProfileModel(ProfileModel::longProfile);
    mLongProfilePointsModel = new ProfilePointsModel();
    mLongProfilePointsModel->setModel(mLongProfileModel);
    mLongProfilePointsModel->setRootIndex(mLongProfileModel->index(0,0, QModelIndex()));
    //mLongProfileModel->setAlignmentModel(mAlignmentModel);

    mLongProfileMeasurementModel = new MeasurementPointModel();
    mLongProfileMeasurementModel->setModel(mLongProfileModel);
    mLongProfileMeasurementModel->setRootIndex(mLongProfileModel->index(0,1,QModelIndex()));

    mAlignmentLayer->setModel(mAlignmentModel);
    mPointLayer->setModel(mPointsModel);
}

void MainWindow::loadFileDialog()
{
    //Minimum size set so that these can be used with both mouse or touch device
    QPushButton* newFileButton = new QPushButton(tr("&New File"));
    newFileButton->setDefault(true);
    newFileButton->setMinimumSize(128,64);

    QPushButton* existingFileButton = new QPushButton(tr("&Select Existing File"));
    existingFileButton->setMinimumSize(128,64);

    QDialogButtonBox* introButtonBox = new QDialogButtonBox();
    introButtonBox->setCenterButtons(true);
    introButtonBox->addButton(newFileButton, QDialogButtonBox::AcceptRole);
    introButtonBox->addButton(existingFileButton, QDialogButtonBox::RejectRole);

    //introButtonBox->setMinimumWidth(300);
    //introButtonBox->setMinimumHeight(300);
    introButtonBox->setContentsMargins(10,30,10,30);

    QObject::connect(introButtonBox, &QDialogButtonBox::accepted, introButtonBox, &QDialogButtonBox::close);
    QObject::connect(introButtonBox, &QDialogButtonBox::rejected, introButtonBox, &QDialogButtonBox::close);
    QObject::connect(introButtonBox, &QDialogButtonBox::accepted, this, &MainWindow::newFile);
    QObject::connect(introButtonBox, &QDialogButtonBox::rejected, this, &MainWindow::openFile);
    introButtonBox->setWindowTitle("GAT");
    introButtonBox->setWindowModality(Qt::WindowModal);
    introButtonBox->setWindowFlags(Qt::WindowStaysOnTopHint);
    introButtonBox->move(this->pos());
    introButtonBox->show();
}

void MainWindow::setFile(QString fileName)
{
    QGuiApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    if(!fileName.isEmpty()){
        mDataManager->setDatabaseFileName(fileName);
        this->setWindowTitle("GAT " + fileName);
    }

    QGuiApplication::restoreOverrideCursor();
}

void MainWindow::setAddAlignmentPointMode()
{
    mGATPlanViewScene->setMode(GATPlanViewScene::addAlignmentPointMode);
    ui->mGATPlanView->setToolMode(GATPlanViewScene::addAlignmentPointMode);
}

void MainWindow::setMoveAlignmentPointMode()
{
    mGATPlanViewScene->setMode(GATPlanViewScene::moveAlignmentPointMode);
    ui->mGATPlanView->setToolMode(GATPlanViewScene::moveAlignmentPointMode);
}

void MainWindow::setDeleteAlignmentPointMode()
{
    mGATPlanViewScene->setMode(GATPlanViewScene::deleteAlignmentPointMode);
    ui->mGATPlanView->setToolMode(GATPlanViewScene::deleteAlignmentPointMode);
}

void MainWindow::setInsertAlignmentPointMode()
{
    mGATPlanViewScene->setMode(GATPlanViewScene::insertAlignmentPointMode);
    ui->mGATPlanView->setToolMode(GATPlanViewScene::insertAlignmentPointMode);

}

void MainWindow::clearGraphicsViewMode(){
    mGATPlanViewScene->setMode(GATPlanViewScene::panMode);
    ui->mGATPlanView->setPanMode();
}

void MainWindow::openFile()
{
    const QString DEFAULT_DIR_KEY("default_dir");

    QSettings settings;
    QString fileName = QFileDialog::getOpenFileName(this,
                                                tr("Open Database"),
                                                settings.value(DEFAULT_DIR_KEY).toString(),
                                                tr("SQLite (*.db)"));
    if(fileName.isEmpty()){
        return;
    }
    QDir dir;
    settings.setValue(DEFAULT_DIR_KEY, dir.absoluteFilePath(fileName));
    setFile(fileName);
    ui->mGATPlanView->zoomExtents();
}

void MainWindow::newFile()
{
    const QString DEFAULT_DIR_KEY("default_dir");

    QSettings settings;
    QString fileName = QFileDialog::getSaveFileName(this,
                                                tr("New Database"),
                                                settings.value(DEFAULT_DIR_KEY).toString(),
                                                tr("SQLite (*.db)"));

    if(fileName.isEmpty()){
        return;
    }
    QDir dir;
    settings.setValue(DEFAULT_DIR_KEY, dir.absoluteFilePath(fileName));
    if(QFile::exists(fileName)){
        QFile::remove(fileName);
    }
    setFile(fileName);
}

void MainWindow::readLandXML()
{
    const QString DEFAULT_DIR_KEY("default_dir");

    QSettings settings;
    QString fileName = QFileDialog::getOpenFileName(this,
                                                tr("Open Land XML"),
                                                settings.value(DEFAULT_DIR_KEY).toString(),
                                                tr("LAND XML (*.xml)"));

    if(fileName.isEmpty()){
        return;
    }
    QDir dir;
    settings.setValue(DEFAULT_DIR_KEY, dir.absoluteFilePath(fileName));
    GATLandXML landxml(fileName);
    landxml.readXML();
    mAlignmentModel->addAlignment(landxml.getAlignment());
}

void MainWindow::addProfile(const QModelIndex &alignmentIndex, const QString pointFilters, const double tolerance, const int profileType)
{
    ProfileCalculator calculater;
    std::unique_ptr<CrossSection> newProfile =calculater.createProfilePointsAlignment(alignmentIndex,mAlignmentModel, pointFilters, tolerance);
    if(newProfile == nullptr){
        qDebug() << "MainWindow::addProfile null ptr not adding anything";
        return;
    }
    if(profileType == 0){
        mCrossSectionModel->addProfile(std::move(newProfile));
    }
    else if (profileType == 1){
        mLongProfileModel->addProfile(std::move(newProfile));
    }
}

void MainWindow::setCurrentAlignment(const QModelIndex & idx)
{
    //qDebug() << "AlignmentWidget::setCurrentAlignment root index " << mAlignmentModel->index(idx.row(),0,QModelIndex());
    QModelIndex currentAlignment = mAlignmentModel->index(idx.row(),0,QModelIndex());
    mAlignmentPointsModel->setRootIndex(currentAlignment);
    mAlignmentMeasurementModel->setRootIndex(mAlignmentModel->index(idx.row(),1,QModelIndex()));
    QModelIndex descriptionIndex = mAlignmentModel->index(idx.row(),1, QModelIndex());
    QString labelText = mAlignmentModel->data(currentAlignment).toString() + " " + mAlignmentModel->data(descriptionIndex).toString();
    ui->mPlanViewAlignmentLabel->setText(labelText);
}

void MainWindow::setCurrentAlignmentRow(const int row)
{
    QModelIndex currentAlignment = mAlignmentModel->index(row,0,QModelIndex());
    setCurrentAlignment(currentAlignment);
}

void MainWindow::closeEvent(QCloseEvent *event)
{
    writeSettings();
    QMainWindow::closeEvent(event);
}

void MainWindow::resetTabs()
{
    ui->alignmentsDockWidget->show();

    ui->pointsDockWidget->show();

    if(mMobileMode){
        ui->alignmentsDockWidget->setWidget(ui->alignmentContents);
        ui->pointsDockWidget->setWidget(ui->pointsContents);
        mMobileMode = false;
    }
}

void MainWindow::readSettings()
{
    restore();
}

void MainWindow::writeSettings()
{
    QSettings settings("Tailwater Limited", "GAT");
    settings.setValue("geometry", saveGeometry());
    settings.setValue("windowState", saveState());
}

void MainWindow::restore()
{
    QSettings settings("Tailwater Limited", "GAT");
    restoreGeometry(settings.value("geometry").toByteArray());
    restoreState(settings.value("windowState").toByteArray());
}

void MainWindow::getNEZDFileImport()
{
    const QString DEFAULT_DIR_KEY("default_dir");
    QSettings settings;
    QString fileName = QFileDialog::getOpenFileName(this,
                                                tr("Open PNEZD File"),
                                                settings.value(DEFAULT_DIR_KEY).toString(),
                                                tr("csv (*.csv);; txt (*.txt)"));
    if(!fileName.isEmpty()){
            QDir fileDir;
            settings.setValue(DEFAULT_DIR_KEY,
                              fileDir.absoluteFilePath(fileName));
        ui->centralwidget->setStatusTip("Begin File Import");
        QGuiApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
        mPointsModel->addPointsFromFile(",",fileName);
        QGuiApplication::restoreOverrideCursor();
        ui->centralwidget->setStatusTip("");
    }
}

void MainWindow::calculateProfile(){
    CreateProfileDialog* profileDialog = new CreateProfileDialog();
    profileDialog->setPointModel(mPointsModel);
    profileDialog->setAlignmentModel(mAlignmentModel);

    QObject::connect(profileDialog, &CreateProfileDialog::addProfile, this, &MainWindow::addProfile);
    profileDialog->exec();
    clearGraphicsViewMode();
}

//void MainWindow::calculateHydraulicParametersDialog()
//{
//    CalculateHydraulicParametersDialog* hydraulicParmsDialog = new CalculateHydraulicParametersDialog();
//    if(ui->crossSectionTable->selectionModel()->selectedIndexes().size() == 0){
//        QMessageBox msgBox;
//        msgBox.setText("No Cross Seciton Selected Exiting");
//        msgBox.exec();
//        return;
//    }
//    QModelIndex crossSection = ui->crossSectionTable->selectionModel()->selectedIndexes().value(0);
//    QModelIndex parent = mCrossSectionModel->index(crossSection.row(),1,QModelIndex());
//    hydraulicParmsDialog->setCurrentIndex(parent);
//    QObject::connect(hydraulicParmsDialog,
//                     &CalculateHydraulicParametersDialog::calculateHydraulicParameters,
//                     mCrossSectionModel,
//                     &ProfileModel::calculateHydraulicParameters);
//    hydraulicParmsDialog->exec();
//}

void MainWindow::addAlignment()
{
    QDialog * alignmentInputDialog = new QDialog;
    alignmentInputDialog->setModal(true);
    QFormLayout* dialogLayout = new QFormLayout();
    dialogLayout->setMargin(20);

    QLineEdit* alignmentDesInput = new QLineEdit();

    dialogLayout->addRow("Alignment Description", alignmentDesInput);

    QPushButton* okButton = new QPushButton(alignmentInputDialog);
    okButton->setText("Submit");
    dialogLayout->addRow(okButton);

    QPushButton* cancelButton = new QPushButton(alignmentInputDialog);
    cancelButton->setText("Cancel");
    dialogLayout->addRow(cancelButton);
    alignmentInputDialog->setLayout(dialogLayout);

    connect(cancelButton,&QPushButton::released,alignmentInputDialog,&QDialog::reject);
    connect(okButton, &QPushButton::released,alignmentInputDialog,&QDialog::accept);
    alignmentInputDialog->exec();
    if(alignmentInputDialog->result()){
        mAlignmentModel->addAlignmentDes(alignmentDesInput->text());
        ui->alignmentsTable->selectionModel()->setCurrentIndex(mAlignmentModel->index(mAlignmentModel->rowCount()-1,0),QItemSelectionModel::Select);
        setCurrentAlignment(mAlignmentModel->index(mAlignmentModel->rowCount()-1,0));
    }
}

void MainWindow::addAlignmentPoint(const double easting, const double northing)
{
    //If there isn't an alignment selected do nothing.
    if(ui->alignmentsTable->selectionModel()->selectedIndexes().size() > 0){
        QModelIndex parent = mAlignmentModel->index(ui->alignmentsTable->selectionModel()->selectedIndexes()[0].row(),0);
        mAlignmentModel->addAlignmentPoint(easting, northing, parent);
    }
}

void MainWindow::importCrossSection()
{
    //Need file dialog to select file
    QUrl fileName = QFileDialog::getOpenFileUrl(this,
                                                tr("Open NEZ File"),
                                                QUrl::fromLocalFile("C:\\"),
                                                tr("csv (*.csv)"));
    if(!fileName.isEmpty()){
        //Read through files to make sure all rows have same number of values
        int columns = GAT::getCSVColumnCount(fileName);
        //Have user select which column is which what about northing/easting vs station/offset
        importCSVFileDialog* optionSelectDialog = new importCSVFileDialog();
        optionSelectDialog->exec();

        if(optionSelectDialog->Accepted){
            if(optionSelectDialog->getImportType() == 0){
                if(columns == 2){
                    if(!mCrossSectionModel->readCSVCrossSectionStationOffset(fileName, 1))
                    {
                        QMessageBox errorMessageBox;
                        errorMessageBox.setText("Error Reading File");
                        errorMessageBox.exec();
                        return;
                    }
                }
                else {
                    QMessageBox errorMessageBox;
                    errorMessageBox.setText("Wrong number of columns should be 2");
                    errorMessageBox.exec();
                    return;
                }
            }
            else if (optionSelectDialog->getImportType() == 1){
                if (columns == 3){
                    if(!(mCrossSectionModel->readCSVCrossSectionNEZ(fileName)))
                    {
                            QMessageBox errorMessageBox;
                            errorMessageBox.setText("Error Reading File");
                            errorMessageBox.exec();
                            return;
                    }
                }
                else {
                    QMessageBox errorMessageBox;
                    errorMessageBox.setText("Wrong number of columns should be 3");
                    errorMessageBox.exec();
                    return;
                }

            }
        } 
    }

    //no file name routine will exit
}

void MainWindow::showAbout()
/*! \brief void MainWindow::showAbout() function builds and shows the license and credits text for the about box
 *
 * QDialog box is used to show the GNU LGPL licensing information. Credits are owed to many individuals including some people who may not be mentioned.
 * Text is formatted as an HTML document, it was copied from the GNU website with existing styles. Version number is manually set so this may contain errors.
 */
{
    QDialog* aboutBox = new QDialog();
    QScrollArea* scroll = new QScrollArea();
    QTextEdit *licenseText = new QTextEdit(); /**< member to hold the displayed text in a QTextEdit box */
    licenseText->setHtml(
                "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">"
                "<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">"
                "p, li { white-space: pre-wrap; }"
                "</style></head><body style=\" font-family:'MS Shell Dlg 2'; font-size:8.25pt; font-weight:400; font-style:normal;\">"
                "<h3 style=\" margin-top:14px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">Geomorphic Assment Tools is licensed under the following license.</span></h3>"
                "<h3 style=\" margin-top:14px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">Version 3.1</span></h3>"
                "<h3 style=\" margin-top:14px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">Source code for this project may be obtained by emailing andrea@tailwaterlimited.com or may be provided with the compiled version of this software.</span></h3>"
                "<h3 style=\" margin-top:14px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">This software was compiled with QT version 5.10.1</span></h3>"
                "<h3 style=\" margin-top:14px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">GNU LESSER GENERAL PUBLIC LICENSE</span></h3>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">Version 3, 29 June 2007</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">Copyright © 2007 Free Software Foundation, Inc. &lt;</span><a href=\"http://fsf.org/\"><span style=\" font-size:8pt; text-decoration: underline; color:#0000ff;\">http://fsf.org/</span></a><span style=\" font-size:8pt;\">&gt;</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">Everyone is permitted to copy and distribute verbatim copies of this license document, but changing it is not allowed.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">This version of the GNU Lesser General Public License incorporates the terms and conditions of version 3 of the GNU General Public License, supplemented by the additional permissions listed below.</span></p>"
                "<h4 style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">0. Additional Definitions.</span></h4>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">As used herein, “this License” refers to version 3 of the GNU Lesser General Public License, and the “GNU GPL” refers to version 3 of the GNU General Public License.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">“The Library” refers to a covered work governed by this License, other than an Application or a Combined Work as defined below.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">An “Application” is any work that makes use of an interface provided by the Library, but which is not otherwise based on the Library. Defining a subclass of a class defined by the Library is deemed a mode of using an interface provided by the Library.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">A “Combined Work” is a work produced by combining or linking an Application with the Library. The particular version of the Library with which the Combined Work was made is also called the “Linked Version”.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">The “Minimal Corresponding Source” for a Combined Work means the Corresponding Source for the Combined Work, excluding any source code for portions of the Combined Work that, considered in isolation, are based on the Application, and not on the Linked Version.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">The “Corresponding Application Code” for a Combined Work means the object code and/or source code for the Application, including any data and utility programs needed for reproducing the Combined Work from the Application, but excluding the System Libraries of the Combined Work.</span></p>"
                "<h4 style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">1. Exception to Section 3 of the GNU GPL.</span></h4>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">You may convey a covered work under sections 3 and 4 of this License without being bound by section 3 of the GNU GPL.</span></p>"
                "<h4 style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">2. Conveying Modified Versions.</span></h4>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">If you modify a copy of the Library, and, in your modifications, a facility refers to a function or data to be supplied by an Application that uses the facility (other than as an argument passed when the facility is invoked), then you may convey a copy of the modified version:</span></p>"
                "<ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 1;\"><li style=\" font-size:8pt;\" style=\" margin-top:12px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">a) under this License, provided that you make a good faith effort to ensure that, in the event an Application does not supply the function or data, the facility still operates, and performs whatever part of its purpose remains meaningful, or</li>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">b) under the GNU GPL, with none of the additional permissions of this License applicable to that copy.</li></ul>"
                "<h4 style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">3. Object Code Incorporating Material from Library Header Files.</span></h4>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">The object code form of an Application may incorporate material from a header file that is part of the Library. You may convey such object code under terms of your choice, provided that, if the incorporated material is not limited to numerical parameters, data structure layouts and accessors, or small macros, inline functions and templates (ten or fewer lines in length), you do both of the following:</span></p>"
                "<ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 1;\"><li style=\" font-size:8pt;\" style=\" margin-top:12px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">a) Give prominent notice with each copy of the object code that the Library is used in it and that the Library and its use are covered by this License.</li>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">b) Accompany the object code with a copy of the GNU GPL and this license document.</li></ul>"
                "<h4 style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">4. Combined Works.</span></h4>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">You may convey a Combined Work under terms of your choice that, taken together, effectively do not restrict modification of the portions of the Library contained in the Combined Work and reverse engineering for debugging such modifications, if you also do each of the following:</span></p>"
                "<ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 1;\"><li style=\" font-size:8pt;\" style=\" margin-top:12px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">a) Give prominent notice with each copy of the Combined Work that the Library is used in it and that the Library and its use are covered by this License.</li>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">b) Accompany the Combined Work with a copy of the GNU GPL and this license document.</li>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">c) For a Combined Work that displays copyright notices during execution, include the copyright notice for the Library among these notices, as well as a reference directing the user to the copies of the GNU GPL and this license document.</li>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">d) Do one of the following: </li>"
                "<ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 2;\"><li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">0) Convey the Minimal Corresponding Source under the terms of this License, and the Corresponding Application Code in a form suitable for, and under terms that permit, the user to recombine or relink the Application with a modified version of the Linked Version to produce a modified Combined Work, in the manner specified by section 6 of the GNU GPL for conveying Corresponding Source.</li>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">1) Use a suitable shared library mechanism for linking with the Library. A suitable mechanism is one that (a) uses at run time a copy of the Library already present on the user's computer system, and (b) will operate properly with a modified version of the Library that is interface-compatible with the Linked Version.</li></ul>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">e) Provide Installation Information, but only if you would otherwise be required to provide such information under section 6 of the GNU GPL, and only to the extent that such information is necessary to install and execute a modified version of the Combined Work produced by recombining or relinking the Application with a modified version of the Linked Version. (If you use option 4d0, the Installation Information must accompany the Minimal Corresponding Source and Corresponding Application Code. If you use option 4d1, you must provide the Installation Information in the manner specified by section 6 of the GNU GPL for conveying Corresponding Source.)</li></ul>"
                "<h4 style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">5. Combined Libraries.</span></h4>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">You may place library facilities that are a work based on the Library side by side in a single library together with other library facilities that are not Applications and are not covered by this License, and convey such a combined library under terms of your choice, if you do both of the following:</span></p>"
                "<ul style=\"margin-top: 0px; margin-bottom: 0px; margin-left: 0px; margin-right: 0px; -qt-list-indent: 1;\"><li style=\" font-size:8pt;\" style=\" margin-top:12px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">a) Accompany the combined library with a copy of the same work based on the Library, uncombined with any other library facilities, conveyed under the terms of this License.</li>"
                "<li style=\" font-size:8pt;\" style=\" margin-top:0px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">b) Give prominent notice with the combined library that part of it is a work based on the Library, and explaining where to find the accompanying uncombined form of the same work.</li></ul>"
                "<h4 style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt; font-weight:600;\">6. Revised Versions of the GNU Lesser General Public License.</span></h4>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">The Free Software Foundation may publish revised and/or new versions of the GNU Lesser General Public License from time to time. Such new versions will be similar in spirit to the present version, but may differ in detail to address new problems or concerns.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">Each version is given a distinguishing version number. If the Library as you received it specifies that a certain numbered version of the GNU Lesser General Public License “or any later version” applies to it, you have the option of following the terms and conditions either of that published version or of any later version published by the Free Software Foundation. If the Library as you received it does not specify a version number of the GNU Lesser General Public License, you may choose any version of the GNU Lesser General Public License ever published by the Free Software Foundation.</span></p>"
                "<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-size:8pt;\">If the Library as you received it specifies that a proxy can decide whether future versions of the GNU Lesser General Public License shall apply, that proxy's public statement of acceptance of any version is permanent authorization for you to choose that version for the Library. </span></p></body>"
                "<p>Additional Credits<BR>Greg Taillacq, David Bidelspach, Michael Geenen, Ken Lai, Mason Angel, Brandon Rosgen, Laura Burckhardt, and probably many others that I have forgotten to list. Thank you for your help in developing this software.</html>"
                );
    scroll->setWidget(licenseText);
    QHBoxLayout* layout = new QHBoxLayout();
    layout->addWidget(scroll);
    aboutBox->setLayout(layout);
    aboutBox->show();
}

void MainWindow::mobileView()
{
    ui->mMainTabWidget->addTab(ui->planViewFrame,"Plan View");
    ui->mMainTabWidget->addTab(ui->pointsContents, "Points");
    ui->mMainTabWidget->addTab(ui->alignmentContents,"Alignments");

    ui->pointsDockWidget->hide();
    ui->alignmentsDockWidget->hide();

    mMobileMode = true;
}

void MainWindow::alignmentContextMenu(QPoint pos)
{
    mContextIndex = ui->alignmentsTable->indexAt(pos);

    QMenu *menu = new QMenu(this);
    QAction *actionShowPoints = new QAction("show points", this);
    QAction *actionShowMeasurements = new QAction("show measurements", this);
    QObject::connect(actionShowPoints, &QAction::triggered, this, &MainWindow::openAlignmentPointsDock);
    QObject::connect(actionShowMeasurements, &QAction::triggered, this, &MainWindow::openAlignmentMeasurementDock);
    menu->addAction(actionShowPoints);
    menu->addAction(actionShowMeasurements);
    menu->popup(ui->alignmentsTable->viewport()->mapToGlobal(pos));
}

void MainWindow::openAlignmentPointsDock()
{

        mAlignmentPointsDockWidget = new QDockWidget("Alignment Points",this);
        //mAlignmentPointsDockLayout = new QHBoxLayout();
        //mAlignmentPointsDockWidget->setLayout(mAlignmentPointsDockLayout);

        alignmentPointsTable = new GATTableView(mAlignmentPointsDockWidget);
        //alignmentPointsTable->setMinimumSize(200,200);
        //alignmentPointsTable->setSizeAdjustPolicy(QAbstractScrollArea::AdjustToContents);
        alignmentPointsTable->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
        mAlignmentPointsDockWidget->setWidget(alignmentPointsTable);

        alignmentPointsTable->setModel(mAlignmentPointsModel);
        //sets the root index so the correct points are shown.
        //points are associated with parent column 0
        mAlignmentPointsModel->setRootIndex(mCrossSectionModel->index(mContextIndex.row(),0));

        mAlignmentPointsDockWidget->show();
}

void MainWindow::openAlignmentMeasurementDock()
{
    mAlignmentMeasurementDockWidget = new QDockWidget("Alignment Measurements", this);
    alignmentMeasurementTable = new GATTableView(mAlignmentMeasurementDockWidget);
    alignmentMeasurementTable->setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding);
    alignmentMeasurementTable->setModel(mCrossSectionMeasurementsModel);
    alignmentMeasurementTypeDelegate = new MeasurementItemDelegate();
    alignmentMeasurementTable->setItemDelegateForColumn(5, alignmentMeasurementTypeDelegate);
    mAlignmentMeasurementDockWidget->setWidget(alignmentMeasurementTable);
    mAlignmentMeasurementModel->setRootIndex(mAlignmentModel->index(mContextIndex.row(),1));
    mAlignmentMeasurementDockWidget->show();
}

void MainWindow::showEvent(QShowEvent *event)
{
    if(!event->spontaneous()){
        loadFileDialog();
    }
}

void MainWindow::setBackgroundImage()
{
    const QString DEFAULT_DIR_KEY("default_dir");
    QSettings settings;

    QUrl fileName = QFileDialog::getOpenFileUrl(this,
                                                tr("Open NEZ File"),
                                                QUrl::fromLocalFile(settings.value(DEFAULT_DIR_KEY).toString()),
                                                tr("tif (*.tif)"));
    //mGATPlanViewScene->setBackgroundImage(fileName.toLocalFile());
}

void MainWindow::exportTableData()
{
    const QString DEFAULT_DIR_KEY("default_dir");
    QSettings settings;
    //Need an option for what to export for now this is just csv of the NESZ
    QString fileName = QFileDialog::getSaveFileName(this,"Export CSV File",settings.value(DEFAULT_DIR_KEY).toString());
    qDebug() << "MainWindow::exportTableData before data export initialization";
    GATDataExport dataExport(fileName);
    qDebug() << "MainWindow::exportTableData file name set ready to set model";
    dataExport.setModel(mCrossSectionModel);
    qDebug() << "MainWindow::exportTableData model set ready to export";
    dataExport.exportFile(GATDataExport::CSV);

    dataExport.setModel(mLongProfileModel);
    dataExport.exportFile(GATDataExport::CSV);

    dataExport.closeFile();
}



