#include "pointlayer.h"
#include <QGraphicsScene>
#include <QDebug>
#include <iterator>

PointLayer::PointLayer(QGraphicsItem* parent) :QObject(),
    QGraphicsItemGroup (parent),
    mIndexPointHash()
{

}

void PointLayer::setModel(PointsModel *model)
{
    mPointsModel = model;
    QObject::connect(mPointsModel,&PointsModel::modelReset, this, &PointLayer::modelReset);
    QObject::connect(mPointsModel, &PointsModel::rowsAboutToBeRemoved, this, &PointLayer::rowsAboutToBeRemoved);
    QObject::connect(mPointsModel, &PointsModel::modelAboutToBeReset, this, &PointLayer::modelAboutToBeReset);
    QObject::connect(mPointsModel, &PointsModel::dataChanged, this, &PointLayer::dataChanged);
    QObject::connect(mPointsModel, &PointsModel::rowsInserted, this, &PointLayer::rowsInserted);
    loadPoints();
}

void PointLayer::loadPoints()
{
    qDebug() << "PointLayer::loadPoints";
    //TODO use the model to build the graphics items.
    for(int i = 0; i < mPointsModel->rowCount(); i++){
            PointGraphicsItem* newPoint = new PointGraphicsItem();
            QModelIndex eastingIndex = mPointsModel->index(i,2,QModelIndex());
            QModelIndex northingIndex = mPointsModel->index(i,1,QModelIndex());
            newPoint->setPosition(mPointsModel->data(eastingIndex).toDouble(),
                                  mPointsModel->data(northingIndex).toDouble());
            addToGroup(newPoint);

            QPersistentModelIndex pIdx(northingIndex);
            mIndexPointHash.insert(newPoint,pIdx);
    }
}

void PointLayer::clearPoints()
{
    if(mIndexPointHash.size() >0){
        QHash<PointGraphicsItem*, QPersistentModelIndex>::const_iterator i;
        for(i = mIndexPointHash.begin(); i!= mIndexPointHash.end();++i){
            PointGraphicsItem *point = i.key();
            removePoint(point);
            delete point;
        }
        mIndexPointHash.clear();
    }
}

QRectF PointLayer::boundingRect() const
{
    if(mIndexPointHash.isEmpty())
    {
        return QRectF();
    }
    QRectF rect;
    QHash<PointGraphicsItem*, QPersistentModelIndex>::const_iterator i;
    //Grab the first item.
    rect = mIndexPointHash.begin().key()->mapToScene(mIndexPointHash.begin().key()->boundingRect()).boundingRect();
    for(i =mIndexPointHash.begin()+1; i != mIndexPointHash.end(); ++i){
        PointGraphicsItem* item = i.key();
            //qDebug() << "PointLayer::boundingRect " << item->mapToScene(item->boundingRect()).boundingRect();
            rect = rect.united(item->mapToScene(item->boundingRect()).boundingRect());
    }

    return rect;

}

void PointLayer::modelReset()
{
    //clearPoints();
    loadPoints();
}

void PointLayer::modelAboutToBeReset()
{
    clearPoints();
}

void PointLayer::dataChanged(const QModelIndex &topLeft, const QModelIndex &bottomRight, const QVector<int> &roles)
{

    if(topLeft.column() >3 || bottomRight.column() < 1){
        //The position of the point did not change.
        return;
    }
    for(int i = topLeft.row(); i <= bottomRight.row(); i++){
        QPersistentModelIndex pIdx = QPersistentModelIndex(mPointsModel->index(i,0,topLeft.parent()));
        PointGraphicsItem* point = mIndexPointHash.key(pIdx);
        double easting = mPointsModel->data(mPointsModel->index(i,2,topLeft.parent())).toDouble();
        double northing = mPointsModel->data(mPointsModel->index(i,1,topLeft.parent())).toDouble();
        point->setPosition(easting,northing);
    }
}

void PointLayer::rowsInserted(const QModelIndex &parent, int first, int last)
{
    for(int i = first; i <=  last; i++){
        PointGraphicsItem* newPoint = new PointGraphicsItem();
        QModelIndex eastingIndex = mPointsModel->index(i, 2, parent);
        QModelIndex northingIndex = mPointsModel->index(i, 1, parent);
        double easting = mPointsModel->data(eastingIndex).toDouble();
        double northing= mPointsModel->data(northingIndex).toDouble();
        newPoint->setPosition(easting,northing);
        addToGroup(newPoint);
        QPersistentModelIndex pIdx(northingIndex);
        mIndexPointHash.insert(newPoint,pIdx);
    }
}

void PointLayer::rowsAboutToBeRemoved(const QModelIndex &parent, int first, int last)
{
    qDebug() << "PointLayer::rowsAboutToBeRemoved";
    for(int i = first; i <= last; i++){
        QPersistentModelIndex pIdx(mPointsModel->index(i,1,parent));
        PointGraphicsItem* point = mIndexPointHash.key(pIdx);
        removePoint(point);
        mIndexPointHash.remove(point);
        delete point;
    }
    prepareGeometryChange();

}

void PointLayer::removePoint(PointGraphicsItem *point)
{
    this->removeFromGroup(point);
}
