//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "profilechartview.h"
#include <QDebug>
#include <QtGui/QMouseEvent>
#include <QtCharts/QChart>
#include "gatrubberband.h"

ProfileChartView::ProfileChartView(QChart *chart, QWidget *parent) :
    QGraphicsView(new QGraphicsScene, parent),
    mIsTouching(false),
    mChart(chart),
    mMeasuring(false)
{
    setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    mMode = ProfileChartView::pan;
    scene()->addItem(mChart);
    mRubberBand = nullptr;


}

bool ProfileChartView::viewportEvent(QEvent *event)
{
    if (event->type() == QEvent::TouchBegin) {
        // By default touch events are converted to mouse events. So
        // after this event we will get a mouse event also but we want
        // to handle touch events as gestures only. So we need this safeguard
        // to block mouse events that are actually generated from touch.
        mIsTouching = true;

        // Turn off animations when handling gestures they
        // will only slow us down.
        mChart->setAnimationOptions(QChart::NoAnimation);
    }
    return QGraphicsView::viewportEvent(event);
}

void ProfileChartView::mousePressEvent(QMouseEvent *event)
{
    if (mIsTouching)
        return;
    if(event->button()== Qt::MidButton || (event->button()==Qt::LeftButton && mMode == ProfileChartView::pan)){
        mMouseDownX = event->x();
        mMouseDownY = event->y();
        setCursor(Qt::ClosedHandCursor);
        event->accept();
        mPan = true;
        return;
    }
    if(mMode == ProfileChartView::addStage && event->button()==Qt::LeftButton)
    {
        //get the point is Scene Coordinates

        QPointF point = this->mapToScene(event->pos());
        //This get the chart time coordinates
        point = mChart->mapFromScene(point);
        qDebug() << "CrossSectionChartView::mousePressEvent map to value " << mChart->mapToValue(point);
        emit addHydraulicStage(mChart->mapToValue(point).y());
        event->accept();
        return;
    }
    else if(mMode == measure && event->button()==Qt::LeftButton){
        if(mRubberBand == nullptr){
            mRubberBand = new GATRubberBand(this);
        }
        mMeasurePointOne = mChart->mapToValue(this->mapToScene(event->pos()));
        origin = event->pos();
        mRubberBand->setStartPoint(origin);
        mRubberBand->show();
        mMeasuring = true;
        event->accept();
        return;
    }

    QGraphicsView::mousePressEvent(event);
}

void ProfileChartView::mouseMoveEvent(QMouseEvent *event)
{
    if (mIsTouching)
        return;

    if(mMeasuring){
        mRubberBand->setEndPoint(event->pos());
        mMeasurePointTwo = mChart->mapToValue(this->mapToScene(event->pos()));
        QPointF displayPoint = mMeasurePointTwo - mMeasurePointOne;
        QString message = "dx: " + QString::number(displayPoint.x(), 'f',2) + " dy: " + QString::number(displayPoint.y(),'f',2) +
                " slope " + QString::number(displayPoint.y()/displayPoint.x(),'f',2);
        emit statusMessage(message);
    }
    else {
        QPointF point = this->mapToScene(event->pos());
        if(mChart->plotArea().contains(point)){
            point = mChart->mapFromScene(point);
            QString message = QString::number(mChart->mapToValue(point).x(),'f',2) + ", " + QString::number(mChart->mapToValue(point).y(),'f',2);
            emit statusMessage(message);
        }
    }

    QGraphicsView::mouseMoveEvent(event);
}

void ProfileChartView::mouseReleaseEvent(QMouseEvent *event)
{
    if (mIsTouching){
        mIsTouching = false;
    // Because we disabled animations when touch event was detected
    // we must put them back on.
        mChart->setAnimationOptions(QChart::SeriesAnimations);
    }
    if(((event->button()==Qt::MidButton || (event->button()==Qt::LeftButton && mMode == ProfileChartView::pan))&& mPan)){
        setCursor(Qt::ArrowCursor);
        mChart->scroll(mMouseDownX-event->x(), event->y()-mMouseDownY);
        event->accept();
        mPan = false;
        return;
    }
    else if(event->button() == Qt::LeftButton && mMode==ProfileChartView::measure && mMeasuring){
        qDebug() << "CrossSectionChartView::mouseReleaseEvent ";
        mRubberBand->hide();
        mMeasurePointTwo = mChart->mapToValue(this->mapToScene(event->pos()));
        emit addMeasurement(mMeasurePointOne, mMeasurePointTwo);
        mMeasuring = false;
        return;
    }
    mMeasuring = false;
    mPan = false;
    QGraphicsView::mouseReleaseEvent(event);
}

/*! \brief Key Commands for cross section chart +,-,left arrow, right arrow, up arrow, down arrow.
 *
 * Use the + key to zoom the chart in
 * Use the - key to zoom the chart out
 * Use the arrow keys to scroll right, left, up, or down.
 */
void ProfileChartView::keyPressEvent(QKeyEvent *event)
{
    switch (event->key()) {
    case Qt::Key_Plus:
        mChart->zoomIn();
        break;
    case Qt::Key_Minus:
        mChart->zoomOut();
        break;
    case Qt::Key_Left:
        mChart->scroll(-10, 0);
        break;
    case Qt::Key_Right:
        mChart->scroll(10, 0);
        break;
    case Qt::Key_Up:
        mChart->scroll(0, 10);
        break;
    case Qt::Key_Down:
        mChart->scroll(0, -10);
        break;
    default:
        QGraphicsView::keyPressEvent(event);
        break;
    }
}

void ProfileChartView::wheelEvent(QWheelEvent *wheelEvent)
{
    if(wheelEvent->delta() > 0){
        mChart->zoomIn();
        wheelEvent->accept();
    }
    else if(wheelEvent->delta() < 0){
        mChart->zoomOut();
        wheelEvent->accept();
    }
    return;
}

void ProfileChartView::setMode(int mode)
{
    mMode = mode;
    //reset all the flags.
    mPan = false;
}

void ProfileChartView::resizeEvent(QResizeEvent *event)
{
    Q_UNUSED(event)
    this->fitInView(mChart,Qt::KeepAspectRatio);
}
