//Geomorphic Assessment Tools (GAT) for processing geomorphic survey data
//Copyright (C) 2019 Tailwater Limited, Andrea Taillacq

//This program is free software: you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation, either version 3 of the License, or
//(at your option) any later version.

//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.

//You should have received a copy of the GNU General Public License
//along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include <QLineSeries>
#include "profilechartwidget.h"
#include <QDebug>
#include "profileeditor.h"

//used https://doc.qt.io/qt-5/qtcharts-zoomlinechart-example.html for assistance with
//interaction.

ProfileChartWidget::ProfileChartWidget(bool XS, QWidget* parent):QWidget(parent),
    mPointMapper(new QVXYModelMapper),
    mSelectedProfileLineSeries(new QLineSeries),
    mSelectedProfileSeriesModel(new ProfilePointsModel),
    mXAxis(new QValueAxis),
    mYAxis(new QValueAxis),
    mProfileSelectList(new QComboBox()),
    mHydraulicParametersModel(new HydraulicParametersModel),
    mMeasurementPointModel(new MeasurementPointModel),
    mProfilePointsTable(new GATTableView),
    mHydraulicParametersTable(new GATTableView),
    mMeasurementsTable(new GATTableView)
{
    hasHydraulicParameters = XS;
    //initialize objects
    mChart = new QChart();
    mChart->setTitle("Cross Sections");
    mChart->legend()->hide();
    mChartView = new ProfileChartView(mChart);

    mPointMapper->setXColumn(3);
    mPointMapper->setYColumn(2);
    mSelectedProfileLineSeries->setPointsVisible(true);
    mPointMapper->setSeries(mSelectedProfileLineSeries);
    mPointMapper->setModel(mSelectedProfileSeriesModel);

    mSelectedProfileLineSeries->attachAxis(mXAxis);
    mSelectedProfileLineSeries->attachAxis(mYAxis);
    mChart->addSeries(mSelectedProfileLineSeries);
    mChart->addAxis(mXAxis, Qt::AlignBottom);
    mChart->addAxis(mYAxis, Qt::AlignLeft);

    mXAxis->setTitleText("Stationing");
    mYAxis->setTitleText("Elevation");

    mActionAddStage = new QAction("Stage Mode");
    mActionPan = new QAction("Pan");
    mActionMeasure = new QAction("Measure");
    mActionEditProfile = new QAction("Edit Profile");
    mActionCenterChart = new QAction ("Reset Chart");
    mActionChartTools = new QActionGroup(this);

    mStatusBar = new QStatusBar(this);
    mStatusBar->showMessage("GAT");
    if(hasHydraulicParameters){
        //If it doesn't have hydraulic parameters don't show this button.
        mActionChartTools->addAction(mActionAddStage);
    }
    mActionChartTools->addAction(mActionPan);
    mActionChartTools->addAction(mActionMeasure);
    mActionChartTools->addAction(mActionCenterChart);
    mActionChartTools->addAction(mActionEditProfile);

    mActionPan->setChecked(true);

    setupUi();
}

void ProfileChartWidget::setModel(ProfileModel * crossSectionModel)
{
    mModel = crossSectionModel;

    mSelectedProfileSeriesModel->setModel(mModel);
    mProfileSelectList->setModel(mModel);
    mProfileSelectList->setModelColumn(1);
    mSelectedProfileSeriesModel->setColumn(0);

    if(hasHydraulicParameters){
        mHydraulicParametersModel->setModel(mModel);
        mHydraulicParametersModel->setColumn(1);
        mHydraulicParametersTable->setModel(mHydraulicParametersModel);
    }

    mMeasurementPointModel->setModel(mModel);
    mMeasurementPointModel->setColumn(2);
    mMeasurementsTable->setModel(mMeasurementPointModel);

    mProfilePointsTable->setModel(mSelectedProfileSeriesModel);

    QObject::connect(mChartView, &ProfileChartView::addHydraulicStage, this, &ProfileChartWidget::addHydraulicStage);
    QObject::connect(mChartView, &ProfileChartView::addMeasurement, this, &ProfileChartWidget::addMeasurement);
    QObject::connect(mActionAddStage, &QAction::triggered, this, &ProfileChartWidget::setAddStageMode);
    QObject::connect(mActionPan, &QAction::triggered, this, &ProfileChartWidget::setPanMode);
    QObject::connect(mChartView, &ProfileChartView::statusMessage, this, &ProfileChartWidget::updateStatusBar);
    QObject::connect(mActionMeasure, &QAction::triggered, this, &ProfileChartWidget::setMeasureMode);
    QObject::connect(mProfileSelectList, QOverload<int>::of(&QComboBox::currentIndexChanged), this, &ProfileChartWidget::setRow);
    QObject::connect(mActionCenterChart, &QAction::triggered, this, &ProfileChartWidget::resetChart);
    QObject::connect(mActionEditProfile, &QAction::triggered, this, &ProfileChartWidget::editProfile);
    QObject::connect(mModel, &ProfileModel::rowsInserted, this, &ProfileChartWidget::rowsInserted);
    QObject::connect(mModel, &ProfileModel::rowsRemoved, this, &ProfileChartWidget::rowsRemoved);
}


void ProfileChartWidget::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
    mChart->resize(mChartView->size());
}

void ProfileChartWidget::addHydraulicStage(double stage)
{
    mModel->calculateHydraulicParameter(mSelectedProfileIndex, stage);
}

void ProfileChartWidget::addMeasurement(const QPointF pointOne, const QPointF pointTwo)
{
    mModel->addMeasurement(mSelectedProfileIndex, pointOne, pointTwo);
}

void ProfileChartWidget::setPanMode()
{
    mChartView->setMode(ProfileChartView::pan);
}

void ProfileChartWidget::setAddStageMode()
{
    mChartView->setMode(ProfileChartView::addStage);
    mTableTabs->setCurrentIndex(2);
}

void ProfileChartWidget::setMeasureMode()
{
    mChartView->setMode(ProfileChartView::measure);
    mTableTabs->setCurrentIndex(1);
}

void ProfileChartWidget::updateStatusBar(const QString message)
{
    mStatusBar->showMessage(message);
}

void ProfileChartWidget::setRow(int row)
{
    if(row == -1)
    {
        mSelectedProfileIndex = QModelIndex();
        return;
    }
    mSelectedProfileIndex = mModel->index(row, 0);
    mSelectedProfileSeriesModel->setRow(row);
    if(hasHydraulicParameters){
        mHydraulicParametersModel->setRow(row);
    }
    mMeasurementPointModel->setRow(row);
    loadSelectedSeries();
}

void ProfileChartWidget::resetChart()
{
    QModelIndex boundingRectIndex = mModel->index(mSelectedProfileIndex.row(),6);
    QRectF boundingRect = mModel->data(boundingRectIndex, Qt::DisplayRole).toRectF();
    qreal marginsSide = boundingRect.width()*.05;
    qreal marginsVert = boundingRect.height()*.05;
    boundingRect = boundingRect.marginsAdded(QMarginsF(marginsSide, marginsVert, marginsSide, marginsVert));
    mXAxis->setRange(boundingRect.left(), boundingRect.right());
    mYAxis->setRange(boundingRect.top(), boundingRect.bottom());
    mChartView->fitInView(mChart);
}

void ProfileChartWidget::editProfile()
{
    if(mSelectedProfileIndex.isValid()){
    ProfileEditor* pEditor = new ProfileEditor();
    pEditor->setModel(mModel);
    pEditor->setIndex(mSelectedProfileIndex);
    pEditor->show();
}
    else {
        QMessageBox errorMessage;
        errorMessage.setText("Please select a valid profile");
        errorMessage.exec();
    }


}

void ProfileChartWidget::rowsInserted(const QModelIndex &parent, int first, int last)
{
    if(mModel->isIndexValid(parent)){
        //Then it is one of the child tables that was modified.
        //This could be optimized to be more selective.
        loadMeasurementSeries();
        return;
    }

    //Selects the newly added profile
    mProfileSelectList->setCurrentIndex(first);
    return;
}

void ProfileChartWidget::rowsRemoved(const QModelIndex &parent, int first, int last)
{
    if(mModel->isIndexValid(parent))
    {
        //This could be optimized a bit to be more selective.
        loadMeasurementSeries();
        return;
    }
    //The profile select list will automatcally change so I don't think anything needs to be
    //implamented here.
    return;
}

void ProfileChartWidget::setupUi()
{
    mChartTableSplitter = new QSplitter(Qt::Vertical);
    mTableTabs = new QTabWidget();
    mLayout = new QVBoxLayout();
    mLayout->setContentsMargins(0,0,0,0);
    mLayout->setMargin(0);

    mProfileSelectList->setEditable(false);
    mToolBar = new QToolBar("Chart Tools");
    mToolBar->addActions(mActionChartTools->actions());
    mToolBar->addWidget(mProfileSelectList);

    mTableTabs->addTab(mProfilePointsTable,"Points");
    mTableTabs->addTab(mMeasurementsTable,"Measurements");
    mMeasurementTypeDelegate = new MeasurementItemDelegate();
    mMeasurementsTable->setItemDelegateForColumn(5, mMeasurementTypeDelegate);
    if(hasHydraulicParameters){
        mTableTabs->addTab(mHydraulicParametersTable,"Hydraulic Parameters");
    }
    mChartTableSplitter->addWidget(mChartView);
    mChartTableSplitter->addWidget(mTableTabs);
    mLayout->addWidget(mToolBar,0);
    mLayout->addWidget(mStatusBar,0);
    mLayout->addWidget(mChartTableSplitter,1);


    setLayout(mLayout);
}

void ProfileChartWidget::loadSelectedSeries()
{
    QModelIndex nameIndex = mModel->index(mSelectedProfileIndex.row(),1);
    qDebug() << "CrossSectionChartWidget::loadSelectedSeries " << mModel->data(nameIndex).toString();
    mSelectedProfileLineSeries->setName(mModel->data(nameIndex).toString());

    QModelIndex boundingRectIndex = mModel->index(mSelectedProfileIndex.row(),6);
    QRectF boundingRect = mModel->data(boundingRectIndex, Qt::DisplayRole).toRectF();
    qDebug() << "BoundingRect " << boundingRect;

    mXAxis->setRange(boundingRect.left(), boundingRect.right());
    mYAxis->setRange(boundingRect.top(), boundingRect.bottom());
    mSelectedProfileLineSeries->attachAxis(mXAxis);
    mSelectedProfileLineSeries->attachAxis(mYAxis);

    loadMeasurementSeries();
}

void ProfileChartWidget::loadMeasurementSeries()
{
    QHashIterator<QPersistentModelIndex, QLineSeries*> m(mHydraulicParameterSeries);
    while(m.hasNext()){
            m.next();
            mChart->removeSeries(m.value());
        }
    mHydraulicParameterSeries.clear();
    if(!mMeasurementSeries.isEmpty()){
        for(int i = mMeasurementSeries.length()-1; i >=0; i--){
            mChart->removeSeries(mMeasurementSeries.at(i));
        }
        mMeasurementSeries.erase(mMeasurementSeries.begin(), mMeasurementSeries.end());
    }
    //Hydraulic parameters are under root index column 1
    QModelIndex parent = mModel->index(mSelectedProfileIndex.row(),1);
    int rowCount = mModel->rowCount(parent);
    for(int i = 0; i < rowCount; i++){
        QLineSeries *hydParmSeries = new QLineSeries();
        QModelIndex stageIndex = mModel->index(i,1,parent);
        double stage = mModel->data(stageIndex,Qt::DisplayRole).toDouble();
        QModelIndex descriptionIndex = mModel->index(i, 0, parent);
        QString description = mModel->data(descriptionIndex,Qt::DisplayRole).toString();
        hydParmSeries->setName(description);

        //add the points
        double left = mChart->mapToValue(mChart->plotArea().topLeft()).x();
        double right = mChart->mapToValue(mChart->plotArea().bottomRight()).x();
        hydParmSeries->append(left,stage);
        hydParmSeries->append(right, stage);
        hydParmSeries->setPen(QPen(Qt::DashLine));
        //qDebug() << "CrossSectionChartWidget::loadHydraulicParameterSeries " << right << " " << left << " " << stage;
        mChart->addSeries(hydParmSeries);
        hydParmSeries->attachAxis(mXAxis);
        hydParmSeries->attachAxis(mYAxis);
        QPersistentModelIndex pIdx(stageIndex);
        mHydraulicParameterSeries.insert(pIdx,hydParmSeries);
    }
    //Measurement prameters are under root index column 2
    parent = mModel->index(mSelectedProfileIndex.row(),2);
    rowCount = mModel->rowCount(parent);
    for(int i = 0; i < rowCount; i++){
        QLineSeries *measSeries = new QLineSeries();

        //TODO: check columns
        qreal x1 = mModel->data(mModel->index(i,1,parent)).toReal();
        qreal y1 = mModel->data(mModel->index(i,2,parent)).toReal();
        qreal x2 = mModel->data(mModel->index(i,3,parent)).toReal();
        qreal y2 = mModel->data(mModel->index(i,4,parent)).toReal();
        measSeries->append(x1,y1);
        measSeries->append(x2,y2);
        measSeries->setPen(QPen(Qt::red));
        mChart->addSeries(measSeries);
        measSeries->attachAxis(mXAxis);
        measSeries->attachAxis(mYAxis);
        mMeasurementSeries.append(measSeries);
    }

}


